package de.marstra.dvo.evaluation;

import com.fasterxml.jackson.databind.annotation.JsonDeserialize;
import com.fasterxml.jackson.databind.annotation.JsonPOJOBuilder;
import de.marstra.dvo.obfuscator.util.CommandExecutor;
import lombok.Builder;
import lombok.Data;

import java.io.IOException;
import java.nio.file.Path;
import java.nio.file.Paths;
import java.util.regex.Matcher;
import java.util.regex.Pattern;
import java.util.stream.Stream;

public class AdbConnection {
    public static final String EMULATOR = "-e";
    public static final String USB_DEVICE = "-d";
    private final String device;

    public AdbConnection(String device) {
        this.device = " " + device.trim() + " ";
    }

    public AdbConnection() {
        this(" ");
    }

    public static class AdbException extends Exception {
        public AdbException(String s) {
            super(s);
        }
    }

    public MemUsage getMemUsage(String pck) throws IOException, AdbException {
        String output = shellExcute("dumpsys meminfo " + pck);
        return new MemUsage(
                parseNumber(output, "Native Heap\\s+\\d+\\s+(\\d+)"),
                parseNumber(output, "Dalvik Heap\\s+\\d+\\s+(\\d+)"),
                parseNumber(output, "Java Heap:\\s+(\\d+)"),
                parseNumber(output, "Native Heap:\\s+(\\d+)"),
                parseNumber(output, "Code:\\s+(\\d+)"),
                parseNumber(output, "Stack:\\s+(\\d+)"),
                parseNumber(output, "Graphics:\\s+(\\d+)"),
                parseNumber(output, "Private Other:\\s+(\\d+)"),
                parseNumber(output, "System:\\s+(\\d+)"),
                parseNumber(output, "TOTAL:\\s+(\\d+)")
        );
    }

    /**
     * get cpu usage using proc, see man proc
     * the usage is quotient of the cpu time over the total time the process is running
     */
    public double getMeanCpuUsage(int pid) throws IOException {
        String output = shellExcute("ps -p " + pid + " -o %cpu");
        return Double.parseDouble(output.split("\n")[1].trim());
    }

    @Data
    @Builder
    @JsonDeserialize(builder = MemUsage.MemUsageBuilder.class)
    public static class MemUsage {
        private final int privDirtyNativeHeap;
        private final int privDirtyDalvikHeap;
        private final int javaHeap;
        private final int nativeHeap;
        private final int code;
        private final int stack;
        private final int graphics;
        private final int privateOther;
        private final int system;
        private final int total;

        @JsonPOJOBuilder(withPrefix = "")
        public static class MemUsageBuilder {
        }
    }

    public boolean install(Path path) {
        Path remoteFile = Paths.get("/data/local/tmp/").resolve(path.getFileName());
        try {
            shellExcute("rm -f " + remoteFile);
            pushFile(path, remoteFile);
            shellExcute("pm install " + remoteFile, 120);
            return true;
        } catch (IOException e) {
            return false;
        } finally {
            try {
                shellExcute("rm -f " + remoteFile);
            } catch (IOException e) {
                throw new RuntimeException(e);
            }
        }
    }

    public boolean uninstall(String pck) {
        try {
            excute("uninstall " + pck);
            return true;
        } catch (IOException e) {
            return false;
        }
    }

    public void startApp(String pck) throws IOException {
        shellExcute("monkey -p " + pck + " -c android.intent.category.LAUNCHER 1");
    }

    public boolean isInstalled(String pck) throws IOException {
        String processList = null;
        processList = shellExcute("pm list packages");
        return Stream.of(processList.split("\n")).anyMatch((line) -> line.endsWith(pck));
    }

    public void clearLogs() throws IOException {
        excute("logcat -c");
    }

    public String getCrashLogs(int pid) throws IOException {
        return excute("logcat --pid=" + pid + " -d -b crash");
    }

    public int getPid(String pck) {
        try {
            String pidString = shellExcute("pidof " + pck);
            return Integer.parseInt(pidString.trim());
        } catch (Exception e) {
            return -1;
        }
    }

    private static int parseNumber(String output, String regex) throws AdbException {
        Matcher m = Pattern.compile(regex).matcher(output);
        if (!m.find()) throw new AdbException("parsing meminfo failed, output: " + output);
        return Integer.parseInt(m.group(1));
    }

    public void input(String cmd, int timeout) throws IOException {
        shellExcute("input " + cmd, timeout);
    }

    public void input(String cmd) throws IOException {
        shellExcute("input " + cmd, 30);
    }

    public void tap(int x, int y) throws IOException {
        input("tap " + x + " " + y);
    }

    public void swipe(int x1, int y1, int x2, int y2, int duration) throws IOException {
        input("swipe " + x1 + " " + y1 + " " + x2 + " " + y2 + " " + duration, duration + 10);
    }

    public void longPress(int x1, int y1, int duration) throws IOException {
        swipe(x1, y1, x1, y1, duration);
    }

    public void text(String text) throws IOException {
        input("text \"" + text + "\"");
    }

    public void keyEvent(String id) throws IOException {
        input("keyevent " + id);
    }

    public void pushFile(Path file, Path target) throws IOException {
        excute("push " + file + " " + target);
    }

    public String excute(String command) throws IOException {
        return excute(command, 120);
    }

    public String excute(String command, int timeout) throws IOException {
        return CommandExecutor.executeCommand("adb" + device + command, null, timeout).getStdout();
    }

    public String shellExcute(String command, int timeout) throws IOException {
        return CommandExecutor.executeCommand("adb" + device + "shell " + command, null, timeout).getStdout();
    }

    public String shellExcute(String command) throws IOException {
        return shellExcute(command, 30);
    }
}
