package de.marstra.dvo.evaluation;

import com.fasterxml.jackson.databind.ObjectMapper;
import de.marstra.dvo.obfuscator.*;
import de.marstra.dvo.obfuscator.analysis.MethodAnalysis;
import de.marstra.dvo.obfuscator.analysis.ParsedDex;
import de.marstra.dvo.obfuscator.util.CompileUtil;
import interpreter.VMODex;
import lombok.Data;
import lombok.extern.slf4j.Slf4j;
import org.apache.commons.io.FilenameUtils;
import org.jf.dexlib2.DexFileFactory;
import org.jf.dexlib2.iface.DexFile;

import java.io.File;
import java.io.IOException;
import java.nio.file.Files;
import java.nio.file.Path;
import java.nio.file.Paths;
import java.util.Arrays;
import java.util.LinkedList;
import java.util.List;
import java.util.Optional;
import java.util.stream.Collectors;

import static de.marstra.dvo.obfuscator.util.Utils.doInTmpDir;
import static interpreter.Utils.getRandomBytes;

@Slf4j
public class AlgoBenchmark {
    @Data
    public static class BenchmarkAlgo {
        private String name;
        private String sourceFile;
        private int inputSize;
        private List<Long> origRes;
        private List<Long> obfuscatedRes;
    }

    @Data
    private static class ObfuscationResult {
        private final DexFile dexFile;
        private final VMODex vmoDexFile;
    }

    private static DexFile compile(Path javaSource) {
        return doInTmpDir(tmpDir -> {
            Files.copy(javaSource, tmpDir.resolve(javaSource.getFileName()));
            CompileUtil.compileJava(tmpDir);
            return CompileUtil.dxClasses(tmpDir);
        });
    }

    private static ObfuscationResult obfuscate(DexFile origDex, byte[] key) {
        ByteCodeMapping mapping = ByteCodeMapping.getRandomMapping();
        ParsedDex parsedDex = new ParsedDex(origDex);

        Config config = Utils.generateConfig(parsedDex, (dm, pd) -> {
            if (!MethodAnalysis.analyseMethod(dm.getMethod(), pd).isSupported()) return true;
            if (dm.getMethod().getName().equals("main")) return true;
            return false;
        });

        VMODexCompiler.CompilationResult compilationResult = VMODexCompiler.compile(parsedDex, config.getMethodsToObfuscate(), mapping);
        DexFile interpreter = InterpreterGenerator.generateInterpreterDex(mapping, compilationResult.getUsedInstructionTypes(), config.getInterpreterVerbosityLevels(), key, false);
        StubInjector.injectInterpreter(parsedDex, compilationResult.getConvertedMethodReferences(), interpreter, config.isRenameInterpreter());
        List<DexFile> patchedDexFiles = parsedDex.toDexFiles();
        if (patchedDexFiles.size() != 1) throw new IllegalStateException();
        return new ObfuscationResult(
                patchedDexFiles.get(0),
                compilationResult.getVmoDex()
        );
    }

    public static void benchmarkAlgos(File index, List<String> only) throws IOException {
        Path benchFolder = index.toPath().getParent();
        ObjectMapper mapper = new ObjectMapper();
        Path indexFile = index.toPath();

        if (!Files.exists(indexFile))
            throw new IllegalArgumentException(indexFile + " does not exist");

        List<BenchmarkAlgo> allAlgos = Arrays.asList(mapper.readValue(indexFile.toFile(), BenchmarkAlgo[].class));

        List<BenchmarkAlgo> target = allAlgos;
        if(only != null && only.size() != 0) {
            target = only.stream().map(name -> {
                Optional<BenchmarkAlgo> bApp = allAlgos.stream().filter(algo -> algo.getName().equals(name)).findFirst();
                if (!bApp.isPresent()) throw new IllegalArgumentException("algorithm " + name + "not found");
                return bApp.get();
            }).collect(Collectors.toList());
        }

        AdbConnection adbConn = new AdbConnection(AdbConnection.USB_DEVICE);
        Path deviceDir = Paths.get("/data/local/tmp/");

        for (BenchmarkAlgo algo : target) {
            String name = algo.getName();
            String fileNameBase = FilenameUtils.removeExtension(algo.getSourceFile());
            Path sourcePath = benchFolder.resolve(algo.getSourceFile());
            Path origDexPath = benchFolder.resolve(fileNameBase + ".dex");
            Path patchedDexPath = benchFolder.resolve(fileNameBase + "_patched.dex");
            Path vmodexPath = benchFolder.resolve(fileNameBase + ".vmodex");
            Path origDexPathDevice = deviceDir.resolve(fileNameBase + ".dex");
            Path patchedDexPathDevice = deviceDir.resolve(fileNameBase + "_patched.dex");
            Path vmodexPathDevice = deviceDir.resolve("classes.vmodex");

            log.info("Benchmark Algorithm: " + name);

            for (Path p : new Path[]{origDexPath, patchedDexPath, vmodexPath})
                if (Files.exists(p)) Files.delete(p);

            for (Path p : new Path[]{origDexPathDevice, patchedDexPathDevice, vmodexPathDevice})
                adbConn.shellExcute("rm -f " + p);

            DexFile origDex = compile(sourcePath);
            byte[] key = getRandomBytes(16);
            ObfuscationResult obfRes = obfuscate(origDex, key);

            DexFileFactory.writeDexFile(origDexPath.toString(), origDex);
            DexFileFactory.writeDexFile(patchedDexPath.toString(), obfRes.getDexFile());
            obfRes.getVmoDexFile().writeVMODexFile(vmodexPath.toString(), key);

            adbConn.pushFile(origDexPath, origDexPathDevice);
            adbConn.pushFile(patchedDexPath, patchedDexPathDevice);
            adbConn.pushFile(vmodexPath, vmodexPathDevice);

            Path[] dexFiles = new Path[]{origDexPathDevice, patchedDexPathDevice};
            List<Long>[] measureMents = new List[]{new LinkedList(), new LinkedList()};

            for (int variant = 0; variant <= 1; variant++) {
                for (int run = 1; run <= 10; run++) {
                    long inputSize = algo.getInputSize();
                    String output = adbConn.shellExcute("dalvikvm -cp " + dexFiles[variant] + ":" + deviceDir + " " + fileNameBase + " " + inputSize, 600);
                    long duration = Long.parseLong(output.trim());
                    log.info("Algorithm " + algo.getName() + (variant == 0 ? "_orig" : "_patched")
                            + "(" + inputSize + ") run " + run + " lasted " + duration / 1000000 + "ms");
                    measureMents[variant].add(duration);
                }
            }
            algo.setOrigRes(measureMents[0]);
            algo.setObfuscatedRes(measureMents[1]);

            mapper.writer(new MyJsonPrettyPrinter(new MyJsonPrettyPrinter.MyArrayIndenter("durations")))
                    .writeValue(indexFile.toFile(), allAlgos);
        }
    }
}