package de.marstra.dvo.evaluation;

import com.fasterxml.jackson.annotation.JsonIgnore;
import com.fasterxml.jackson.annotation.JsonInclude;
import com.fasterxml.jackson.annotation.JsonProperty;
import com.fasterxml.jackson.core.JsonGenerator;
import com.fasterxml.jackson.core.JsonParser;
import com.fasterxml.jackson.core.JsonProcessingException;
import com.fasterxml.jackson.core.JsonToken;
import com.fasterxml.jackson.databind.DeserializationContext;
import com.fasterxml.jackson.databind.ObjectMapper;
import com.fasterxml.jackson.databind.SerializerProvider;
import com.fasterxml.jackson.databind.annotation.JsonDeserialize;
import com.fasterxml.jackson.databind.annotation.JsonSerialize;
import com.fasterxml.jackson.databind.deser.std.StdDeserializer;
import com.fasterxml.jackson.databind.ser.std.StdSerializer;
import de.marstra.dvo.obfuscator.Config;
import de.marstra.dvo.obfuscator.ObfuscatorService;
import de.marstra.dvo.obfuscator.analysis.ApkFile;
import de.marstra.dvo.obfuscator.analysis.DexMethod;
import de.marstra.dvo.obfuscator.analysis.MethodAnalysis;
import de.marstra.dvo.obfuscator.analysis.ParsedDex;
import joptsimple.internal.Strings;
import lombok.Data;
import lombok.Getter;
import lombok.RequiredArgsConstructor;
import lombok.extern.slf4j.Slf4j;

import java.io.File;
import java.io.IOException;
import java.nio.file.Files;
import java.nio.file.Path;
import java.nio.file.Paths;
import java.util.*;
import java.util.function.BiFunction;
import java.util.function.Function;
import java.util.stream.Collectors;
import java.util.stream.Stream;


@Slf4j
public class AppBenchmark {
    private static final List<String> EXCLUDED_PACKAGES = Arrays.asList(
            "android",                 // android libraries
            "androidx",
            "java",                    // java standard library
            "javax",                   // other java libraries
            "kotlin",                  // kotlin standard library
            "com/google",              // google libararies
            "okhttp3",                 // http library
            "okio",                    // io lib for java https://github.com/square/okio
            "com/fasterxml/jackson",   // json library
            "net/glxn/qrgen"           // qr code library
    );

    @Data
    static class BenchmarkApp {
        private String appName;
        private String apkUrl;
        private String category;
        @JsonProperty("package")
        private String pck;
        private String mainActivity;
        private int usability;
        private String notes;
        private List<TestRun> runs;
        private List<String> inputSequence;
        @JsonInclude(JsonInclude.Include.NON_NULL)
        private Map<String, List<String>> otherConfigs;

        @JsonIgnore
        private String getFileName(String suffix) {
            return getPck() + suffix + ".apk";
        }
    }

    public static void benchmarkApps(File appIndexFile, File apkFolder, List<String> apps, boolean record) throws IOException {
        ObjectMapper mapper = new ObjectMapper();
        List<BenchmarkApp> benchmarkApps = Arrays.asList(mapper.readValue(appIndexFile, BenchmarkApp[].class));
        AdbConnection adbConn = new AdbConnection(AdbConnection.USB_DEVICE);

        List<BenchmarkApp> target = benchmarkApps;
        if (apps != null && apps.size() != 0) {
            target = apps.stream().map(pck -> {
                Optional<BenchmarkApp> bApp = benchmarkApps.stream().filter(app -> app.getPck().equals(pck)).findFirst();
                if (!bApp.isPresent()) throw new IllegalArgumentException("package " + pck + "not found");
                return bApp.get();
            }).collect(Collectors.toList());
        }

        target.stream().forEach(app -> {
            if (record)
                recordKeys(adbConn, apkFolder, app);
            else
                benchmarkApp(adbConn, apkFolder, app);
            Utils.serializeJson(mapper, appIndexFile, benchmarkApps, Arrays.asList("cpuUsage", "memSamples"));
        });
    }

    public static void recordKeys(AdbConnection adbConn, File apkFolder, BenchmarkApp app) {
        Path origFile = Paths.get(apkFolder.getAbsolutePath(), app.getFileName(""));
        Path obfuscatedFile = Paths.get(apkFolder.getAbsolutePath(), app.getFileName("_obfuscated"));

        try {
            if (!Files.exists(origFile)) {
                log.info("download " + app.getPck());
                Utils.download(app.getApkUrl(), origFile);
            }

            log.info("obfuscate " + origFile);
            ApkFile apkFile = new ApkFile(origFile.toFile());
            Config config = Utils.generateConfig(apkFile.getParsedDex(), isMethodExcluded(EXCLUDED_PACKAGES),
//                        new HashSet<>(Arrays.asList(InterpreterGenerator.VerbosityLevel.INTERPRET_START, InterpreterGenerator.VerbosityLevel.EVERY_INSTRUCTION)),
//                        new HashSet<>(Arrays.asList(InterpreterGenerator.VerbosityLevel.INTERPRET_START)),
                    new HashSet<>(),
                    true, false, false, true);
            byte[] obfApkBytes = ObfuscatorService.obfuscate(apkFile, config);
            Files.write(obfuscatedFile, obfApkBytes);

            Scanner scanner = new Scanner(System.in);
            boolean retry = true;
            while (retry) {
                log.info("record keys for " + obfuscatedFile);

                if (adbConn.isInstalled(app.getPck()))
                    adbConn.uninstall(app.getPck());

                adbConn.clearLogs();
                adbConn.install(obfuscatedFile);
                adbConn.startApp(app.getPck());

                log.info("enter command");
                while (true) {
                    String line = scanner.nextLine().trim();
                    if (line.equals("next")) {
                        retry = false;
                        break;
                    } else if (line.equals("retry")) {
                        retry = true;
                        break;
                    } else if (line.equals("done")) {
                        adbConn.uninstall(app.getPck());
                        System.exit(0);
                    } else {
                        log.warn("unsupported command " + line);
                    }
                }
            }

            adbConn.uninstall(app.getPck());
        } catch (Exception e) {
            log.error("something unexpected happened: " + app.getPck(), e);
        }
    }

    public static void benchmarkApp(AdbConnection adbConn, File apkFolder, BenchmarkApp app) {
        Function<String, Path> getPath = (suffix) -> Paths.get(apkFolder.getAbsolutePath(), app.getFileName(suffix));

        Path origFile = getPath.apply("");
        Path obfuscatedFile = getPath.apply("_obfuscated");

        List<BenchConfig> benchConfigs = new LinkedList<>();

        try {
            if (!Files.exists(origFile)) {
                log.info("download " + app.getPck());
                Utils.download(app.getApkUrl(), origFile);
            }

            log.info("obfuscate " + origFile);
            {
                ApkFile apkFile = new ApkFile(origFile.toFile());
                Config config = Utils.generateConfig(apkFile.getParsedDex(), isMethodExcluded(EXCLUDED_PACKAGES));
                log.info("obfuscate " + config.getMethodsToObfuscate().size() + " methods");
                config.getMethodsToObfuscate().size();
                Files.write(obfuscatedFile, ObfuscatorService.obfuscate(new ApkFile(origFile.toFile()), config));
            }

            benchConfigs.add(new BenchConfig("orig", origFile));
            benchConfigs.add(new BenchConfig("obfuscated", obfuscatedFile));

            if (app.otherConfigs != null) {
                for (Map.Entry<String, List<String>> entry : app.otherConfigs.entrySet()) {
                    String type = entry.getKey();
                    Path outFile = getPath.apply("_" + type);
                    List<String> selection = entry.getValue();
                    log.info("obfuscate " + type + ": " + origFile);
                    ApkFile apkFile = new ApkFile(origFile.toFile());
                    Config config = Utils.generateConfig(apkFile.getParsedDex(), isMethodIncluded(selection));
                    log.info("obfuscate " + config.getMethodsToObfuscate().size() + " methods");
                    Files.write(outFile, ObfuscatorService.obfuscate(apkFile, config));
                    benchConfigs.add(new BenchConfig(type, outFile));
                }
            }

            if (adbConn.isInstalled(app.getPck()))
                adbConn.uninstall(app.getPck());

            for (int run = 0; run < 10; run++) {
                for (BenchConfig benchConfig : benchConfigs) {
                    log.info("evaluate " + benchConfig.apkFile);
                    if (!Files.exists(benchConfig.apkFile)) {
                        log.warn("skip non existent file: " + benchConfig.apkFile);
                        continue;
                    }
                    adbConn.clearLogs();
                    adbConn.install(benchConfig.apkFile);
                    adbConn.startApp(app.getPck());

                    Utils.sleep(1000);

                    int pid = adbConn.getPid(app.getPck());
                    if (pid < 0)
                        throw new RuntimeException("pid cannot be retrieved for " + app.getPck());

                    MemSamplerThread sampler = new MemSamplerThread(app.getPck(), 500, adbConn);
                    sampler.start();

                    for (String action : app.getInputSequence()) {
                        log.info("DO: " + action);
                        executeAction(action, adbConn);
                    }

                    sampler.interrupt();
                    try {
                        sampler.join();
                    } catch (InterruptedException e) {
                        throw new RuntimeException(e);
                    }

                    benchConfig.memSamples = sampler.getMemSamples();
                    benchConfig.cpuUsage = adbConn.getMeanCpuUsage(pid);

                    adbConn.uninstall(app.getPck());

                    // let phone cool done
                    Utils.sleep(10000);
                }

                if (app.getRuns() == null) app.setRuns(new LinkedList<>());
                app.getRuns().add(new TestRun(benchConfigs));
            }
        } catch (Exception e) {
            log.error("something unexpected happened: " + app.getPck(), e);
        }
    }

    private static BiFunction<DexMethod, ParsedDex, Boolean> isMethodExcluded(List<String> excluded) {
        return (DexMethod dexMethod, ParsedDex parsedDex) -> {
            if (!MethodAnalysis.analyseMethod(dexMethod.getMethod(), parsedDex).isSupported()) return true;
            String packageName = dexMethod.getDexClass().getDexPackage().getFullName();
            Optional<String> excludedPck = excluded.stream().filter(packageName::startsWith).findFirst();
            return excludedPck.isPresent();
        };
    }

    private static BiFunction<DexMethod, ParsedDex, Boolean> isMethodIncluded(List<String> included) {
        return (DexMethod dexMethod, ParsedDex parsedDex) -> {
            if (!MethodAnalysis.analyseMethod(dexMethod.getMethod(), parsedDex).isSupported()) return true;
            String packageName = dexMethod.getDexClass().getDexPackage().getFullName();
            Optional<String> includedPck = included.stream().filter(packageName::startsWith).findFirst();
            return !includedPck.isPresent();
        };
    }

    @RequiredArgsConstructor
    private static class BenchConfig {
        private final String name;
        private final Path apkFile;
        private double cpuUsage;
        private List<Integer> memSamples;
    }

    @JsonSerialize(using = TestRun.TestRunSerializer.class)
    @JsonDeserialize(using = TestRun.TestRunDeserializer.class)
    static class TestRun {
        Map<String, TestRunType> runTypes;

        public TestRun(Map<String, TestRunType> runTypes) {
            this.runTypes = runTypes;
        }

        private static class TestRunType {
            @JsonProperty
            double cpuUsage;
            @JsonProperty
            List<Integer> memSamples;

            public TestRunType() {
            }

            public TestRunType(double cpuUsage, List<Integer> memSamples) {
                this.cpuUsage = cpuUsage;
                this.memSamples = memSamples;
            }
        }

        public TestRun(List<BenchConfig> benchConfigs) {
            runTypes = new HashMap<>();
            benchConfigs.forEach(benchConfig -> runTypes.put(benchConfig.name, new TestRunType(benchConfig.cpuUsage, benchConfig.memSamples)));
        }


        static class TestRunSerializer extends StdSerializer<TestRun> {
            public TestRunSerializer() {
                this(null);
            }

            public TestRunSerializer(Class<TestRun> t) {
                super(t);
            }

            @Override
            public void serialize(TestRun run, JsonGenerator jgen, SerializerProvider provider) throws IOException, JsonProcessingException {
                jgen.writeStartObject();
                for (Map.Entry<String, TestRunType> entry : run.runTypes.entrySet()) {
                    jgen.writeFieldName(entry.getKey());
                    TestRunType runType = entry.getValue();
                    jgen.writeObject(runType);
                }
                jgen.writeEndObject();
            }
        }

        static class TestRunDeserializer extends StdDeserializer<TestRun> {
            public TestRunDeserializer() {
                this(null);
            }

            protected TestRunDeserializer(Class<?> vc) {
                super(vc);
            }

            @Override
            public TestRun deserialize(JsonParser jp, DeserializationContext deserializationContext) throws IOException, JsonProcessingException {
                Map<String, TestRunType> runTypes = new HashMap<>();
                ObjectMapper mapper = new ObjectMapper();
                while (true) {
                    JsonToken nextToken = jp.nextToken();
                    if (nextToken == JsonToken.FIELD_NAME) {
                        String key = jp.getCurrentName();
                        if(jp.nextToken() != JsonToken.START_OBJECT) throw new IllegalArgumentException();
                        TestRunType type = mapper.readValue(jp, TestRunType.class);
                        runTypes.put(key, type);
                    }
                    if (nextToken == JsonToken.END_OBJECT)
                        break;
                }
                return new TestRun(runTypes);
            }
        }
    }

    @RequiredArgsConstructor
    private static class MemSamplerThread extends Thread {
        private final String pck;
        private final int samplingTimeMs;
        private final AdbConnection adbConn;
        @Getter
        private final List<Integer> memSamples = new LinkedList<>();

        @Override
        public void run() {
            while (!isInterrupted()) {
                try {
                    AdbConnection.MemUsage memUsage = adbConn.getMemUsage(pck);
                    memSamples.add(memUsage.getTotal());
                } catch (IOException e) {
                    throw new RuntimeException(e);
                } catch (RuntimeException e) {
                    if (e.getCause() instanceof InterruptedException)
                        break;
                    throw e;
                } catch (AdbConnection.AdbException e) {
                    log.warn("AdbExpcetion occurred: ", e);
                    continue;
                }
                try {
                    Thread.sleep(samplingTimeMs);
                } catch (InterruptedException e) {
                    break;
                }
            }
        }
    }

    private static void executeAction(String action, AdbConnection adbConn) throws IOException {
        if (action.contains("#"))
            action = action.substring(0, action.indexOf("#")).trim();

        String[] parts = Stream.of(action.split(" "))
                .map(String::trim)
                .filter((s) -> !s.isEmpty())
                .toArray(String[]::new);

        switch (parts[0]) {
            case "tap": {
                if (parts.length != 3) throw new IllegalArgumentException("unsupported command " + action);
                adbConn.tap(Integer.parseInt(parts[1]), Integer.parseInt(parts[2]));
                break;
            }
            case "swipe": {
                if (parts.length != 6) throw new IllegalArgumentException("unsupported command " + action);
                adbConn.swipe(Integer.parseInt(parts[1]), Integer.parseInt(parts[2]), Integer.parseInt(parts[3]), Integer.parseInt(parts[4]), Integer.parseInt(parts[5]));
                break;
            }
            case "longPress": {
                if (parts.length != 4) throw new IllegalArgumentException("unsupported command " + action);
                adbConn.longPress(Integer.parseInt(parts[1]), Integer.parseInt(parts[2]), Integer.parseInt(parts[3]));
                break;
            }
            case "text": {
                if (parts.length < 2) throw new IllegalArgumentException("unsupported command " + action);
                String text = Strings.join(parts, " ").substring("text".length()).trim();
                adbConn.text(text);
                break;
            }
            case "keyevent": {
                if (parts.length != 2) throw new IllegalArgumentException("unsupported command " + action);
                adbConn.keyEvent(parts[1]);
                break;
            }
            case "sleep": {
                if (parts.length != 2) throw new IllegalArgumentException("unsupported command " + action);
                try {
                    Thread.sleep(Integer.parseInt(parts[1]));
                } catch (InterruptedException e) {
                    throw new RuntimeException(e);
                }
                break;
            }
            default: {
                throw new UnsupportedOperationException("unsupported command " + action);
            }
        }
    }
}
