package de.marstra.dvo.evaluation;

import com.fasterxml.jackson.annotation.JsonIgnore;
import com.fasterxml.jackson.annotation.JsonProperty;
import com.fasterxml.jackson.databind.ObjectMapper;
import de.marstra.dvo.obfuscator.Config;
import de.marstra.dvo.obfuscator.ObfuscatorService;
import de.marstra.dvo.obfuscator.analysis.*;
import de.marstra.dvo.obfuscator.util.CommandExecutor;
import lombok.Data;
import lombok.extern.slf4j.Slf4j;
import org.apache.commons.io.FileUtils;

import java.io.File;
import java.io.IOException;
import java.nio.file.Files;
import java.nio.file.Path;
import java.nio.file.Paths;
import java.util.Arrays;
import java.util.HashMap;
import java.util.List;
import java.util.Optional;
import java.util.stream.Collectors;

import static de.marstra.dvo.evaluation.Utils.sleep;

@Slf4j
public class PlayTop100 {
    @Data
    static class PlayApp {
        private Integer rank;
        @JsonProperty("package")
        private String pck;
        // applicability evaluation
        private Integer numMethods;
        private Integer numObfuscatableMethods;
        private HashMap<String, Integer> unsupportedReasons;
        private Integer numInstructions;
        private Integer numObfuscatableInstructions;
        private Integer numBytes;
        private Integer numObfuscatableBytes;
        // code size evaluation
        private Long dexSize;
        private Long dexSizeObfuscated;
        private Long vmodexSize;
        // runability evaluation
        private Boolean runsNormal;
        private Boolean runsRezipped;

        private Boolean canObfuscate;
        private Boolean runsObfuscated;


        private Boolean canProcessApktool;
        private Boolean runsApktool;

        @JsonIgnore
        public String getApkFileName(String suffix) {
            return String.format("%03d", rank) + "_" + pck + suffix + ".apk";
        }
    }

    private static long getFilesSize(ApkFile apkFile, String filePattern) {
        return apkFile.getZipEntries().stream()
                .filter(entry -> entry.getName().matches(filePattern))
                .mapToLong((entry) -> entry.getContent().length)
                .reduce(0, (sum, cur) -> sum + cur);
    }

    private static void staticEvaluation(PlayApp app, ApkFile apkFile) {
        app.setNumMethods(0);
        app.setNumObfuscatableMethods(0);
        app.setNumInstructions(0);
        app.setNumObfuscatableInstructions(0);
        app.setNumBytes(0);
        app.setNumObfuscatableBytes(0);

        HashMap<String, Integer> unsupportedReasons = new HashMap<>();
        ParsedDex parsedDex = apkFile.getParsedDex();

        parsedDex.accept(new AnalysisVisitorAdapter<Void, Void>() {
            @Override
            public Void visit(DexMethod dexMethod, Void p) {
                // TODO exclude methods of blacklisted packages
                MethodAnalysis.MethodAnalysisResult analysisResult = MethodAnalysis.analyseMethod(dexMethod.getMethod(), parsedDex);
                app.setNumMethods(app.getNumMethods() + 1);
                app.setNumInstructions(app.getNumInstructions() + analysisResult.getNumInstructions());
                app.setNumBytes(app.getNumBytes() + analysisResult.getNumBytes());

                if (analysisResult.isSupported()) {
                    app.setNumObfuscatableMethods(app.getNumObfuscatableMethods() + 1);
                    app.setNumObfuscatableInstructions(app.getNumObfuscatableInstructions() + analysisResult.getNumInstructions());
                    app.setNumObfuscatableBytes(app.getNumObfuscatableBytes() + analysisResult.getNumBytes());
                } else {
                    for (String reason : analysisResult.getUnsupportedReasons()) {
                        unsupportedReasons.putIfAbsent(reason, 0);
                        unsupportedReasons.put(reason, unsupportedReasons.get(reason) + 1);
                    }
                }
                return null;
            }
        });
        app.setUnsupportedReasons(unsupportedReasons);
    }

    private static void repackage(Path in, Path out, Path unzipped) {
        try {
            CommandExecutor.executeCommand("apktool d " + in, unzipped.getParent().toFile(), 120);
            CommandExecutor.executeCommand("apktool b " + unzipped);
            Path repackedPath = Paths.get(unzipped.toString(), "dist", in.getFileName().toString());
            Files.move(repackedPath, out);
            ApkFile.signApk(out.toFile());
        } catch (Exception e) {
            log.error("repackaging failed " + in, e);
        } finally {
            try {
                FileUtils.deleteDirectory(unzipped.toFile());
            } catch (IOException e) {
                throw new RuntimeException(e);
            }
        }
    }

    private static void rezip(Path in, Path out, Path unzipped) throws IOException {
        try {
            CommandExecutor.executeCommand("unzip -d " + unzipped + " " + in);
            CommandExecutor.executeCommand("zip -0 -r " + out + " .", unzipped.toFile(), 120);
            ApkFile.signApk(out.toFile());
        } catch (Exception e) {
            log.error("rezipping failed " + in, e);
        } finally {
            try {
                FileUtils.deleteDirectory(unzipped.toFile());
            } catch (IOException e) {
                throw new RuntimeException(e);
            }
        }
    }

    public static void evaluateApks(File appIndexFile, File apkFolder, List<String> apps) throws IOException {
        ObjectMapper mapper = new ObjectMapper();
        List<PlayApp> playApps = Arrays.asList(mapper.readValue(appIndexFile, PlayApp[].class));
        AdbConnection adbConn = new AdbConnection(AdbConnection.USB_DEVICE);

        List<PlayApp> target = playApps;
        if (apps != null && apps.size() != 0) {
            target = apps.stream().map(id -> {
                Optional<PlayApp> pApp = playApps.stream().filter(app -> id.equals(app.getPck()) || id.equals(app.getRank() + "")).findFirst();
                if (!pApp.isPresent()) throw new IllegalArgumentException("app " + id + "not found");
                return pApp.get();
            }).collect(Collectors.toList());
        }

        target.stream().forEach(app -> {
            evaluateApk(adbConn, apkFolder, app);
            Utils.serializeJson(mapper, appIndexFile, playApps);
        });
    }

    public static void evaluateApk(AdbConnection adbConn, File apkFolder, PlayApp app) {
        long start = System.currentTimeMillis();
        String pck = app.getPck();
        String appName = app.getRank() + " " + pck;
        log.info("analyse " + appName);

        try {
            Path origFile = Paths.get(apkFolder.getAbsolutePath(), app.getApkFileName(""));
            Path patchedFile = Paths.get(apkFolder.getAbsolutePath(), app.getApkFileName("_patched"));
            Path repackagedFile = Paths.get(apkFolder.getAbsolutePath(), app.getApkFileName("_repackaged"));
            Path rezippedFile = Paths.get(apkFolder.getAbsolutePath(), app.getApkFileName("_rezipped"));

            if (!Files.exists(origFile)) {
                log.warn("skip missing apk: " + appName);
                return;
            }

            log.info("static analyse " + origFile);
            ApkFile apkFile = null;
            try {
                apkFile = new ApkFile(origFile.toFile());
            } catch (Exception e) {
                log.error("parsing apk failed " + appName, e);
            }

            if (apkFile != null) {
                staticEvaluation(app, apkFile);
                // obfuscate apk
                Config config = Utils.generateConfig(apkFile.getParsedDex(), null);
                byte[] obfApkBytes = null;
                if (Files.exists(patchedFile)) {
                    obfApkBytes = Files.readAllBytes(patchedFile);
                } else {
                    log.info("obfuscate " + origFile);
                    try {
                        obfApkBytes = ObfuscatorService.obfuscate(apkFile, config);
                        Files.write(patchedFile, obfApkBytes);
                    } catch (Exception e) {
                        log.error("obfuscating apk failed " + appName, e);
                    }
                }
                if (obfApkBytes != null) {
                    ApkFile obfApk = new ApkFile(obfApkBytes);
                    log.info("analyze dex sizes for " + appName);
                    app.setDexSize(getFilesSize(apkFile, "^classes\\d*\\.dex$"));
                    app.setDexSizeObfuscated(getFilesSize(obfApk, "^classes\\d*\\.dex$"));
                    app.setVmodexSize(getFilesSize(obfApk, "^classes\\.vmodex$"));
                }
            }

            Path unzipped = Paths.get(origFile.toString().substring(0, origFile.toString().length() - 4));
            if (Files.exists(unzipped))
                FileUtils.deleteDirectory(unzipped.toFile());

            // check runability for patched file
            app.setCanObfuscate(Files.exists(patchedFile));
            if (Files.exists(patchedFile)) {
                app.setRunsObfuscated(checkRunability(pck, patchedFile, adbConn));
            }
            // check runability for repackaged file
            if (!Files.exists(repackagedFile)) {
                log.info("repackage app " + appName);
                repackage(origFile, repackagedFile, unzipped);
            }
            app.setCanProcessApktool(Files.exists(repackagedFile));
            if (Files.exists(repackagedFile)) {
                app.setRunsApktool(checkRunability(pck, repackagedFile, adbConn));
            }
            // check runability for rezipped file
            if (!Files.exists(rezippedFile)) {
                log.info("rezip app " + appName);
                rezip(origFile, rezippedFile, unzipped);
            }
            if (Files.exists(rezippedFile)) {
                app.setRunsRezipped(checkRunability(pck, rezippedFile, adbConn));
            }
            // check runability for original apk
            app.setRunsNormal(checkRunability(pck, origFile, adbConn));
        } catch (Exception e) {
            log.error("something unexpected happened: " + appName, e);
        }
        long duration = System.currentTimeMillis() - start;
        log.info("analyzing " + appName + " lasted " + duration + "ms");
        log.info(app.toString());
    }

    private static boolean checkRunability(String pck, Path apkPath, AdbConnection adbConn) throws IOException {
        log.info("checkRunability for " + apkPath);
        String path = apkPath.toString();
        String fileName = apkPath.getFileName().toString();

        if (adbConn.isInstalled(pck)) {
            adbConn.uninstall(pck);
        }
        adbConn.clearLogs();
        if (!adbConn.install(apkPath)) {
            log.warn("app '" + apkPath + "' could not be installed");
            return false;
        }

        adbConn.startApp(pck);
        sleep(1000);

        boolean runs = true;
        for (int i = 0; i < 30 && runs; i++) {
            sleep(1000);
            int pid = adbConn.getPid(pck);
            if (pid < 0) {
                log.warn("getPid for " + pck + " failed");
                runs = false;
                break;
            }

            String crashLogs = adbConn.getCrashLogs(pid);
            // TODO is empty to restrictive, or use something like contains("FATAL EXCEPTION: main")
            runs = crashLogs.isEmpty();
        }
        adbConn.uninstall(pck);
        if (!runs)
            log.warn("app '" + apkPath + "' does not run");
        return runs;
    }
}
