package de.marstra.dvo.obfuscator;

import com.squareup.javapoet.*;
import de.marstra.dvo.obfuscator.insn.InstructionType;
import de.marstra.dvo.obfuscator.util.CompileUtil;
import de.marstra.dvo.obfuscator.util.GenericMethodReference;
import de.marstra.dvo.obfuscator.util.Utils;
import interpreter.VMODex;
import org.apache.commons.io.FileUtils;
import org.apache.commons.io.IOUtils;
import org.jf.dexlib2.iface.DexFile;

import javax.lang.model.element.Modifier;
import java.io.*;
import java.net.URL;
import java.net.URLDecoder;
import java.nio.file.Path;
import java.util.Collections;
import java.util.Enumeration;
import java.util.Set;
import java.util.jar.JarEntry;
import java.util.jar.JarFile;
import java.util.stream.Collectors;
import java.util.stream.IntStream;

public class InterpreterGenerator {
    public enum VerbosityLevel {
        INTERPRET_START,
        EVERY_INSTRUCTION
    }

    public static final String INTERPRETER_PACKAGE_NAME = "interpreter";
    public static final String INTERPRETER_CLASS_NAME = "Interpreter";
    public static final String INTERPRETER_METHOD_NAME = "interpret";
    public static final GenericMethodReference INTERPRETER_METHOD = new GenericMethodReference(
            "L" + INTERPRETER_PACKAGE_NAME + "/" + INTERPRETER_CLASS_NAME + ";",
            INTERPRETER_METHOD_NAME,
            "Ljava/lang/Object;",
            "I", "[Ljava/lang/Object;"
    );
    public static final String VMODEX_FILE = "classes.vmodex";

    private final ByteCodeMapping byteCodeMapping;
    private final Set<InstructionType> usedInstructionTypes;
    private final Set<VerbosityLevel> verbosityLevels;
    private final byte[] key;
    private final boolean testMode;

    public InterpreterGenerator(ByteCodeMapping byteCodeMapping, Set<InstructionType> usedInstructionTypes, Set<VerbosityLevel> verbosityLevels, byte[] key, boolean testMode) {
        this.byteCodeMapping = byteCodeMapping;
        this.usedInstructionTypes = usedInstructionTypes;
        this.verbosityLevels = verbosityLevels == null ? Collections.emptySet() : verbosityLevels;
        this.key = key;
        this.testMode = testMode;
    }

    private CodeBlock generateInterpretBody(MethodSpec getVMODexMethod, ParameterSpec methodIdSpec, ParameterSpec argsSpec) {
        CodeBlock.Builder builder = CodeBlock.builder();
        // init code
        builder.addStatement("VMODex vmoDex = $N()", getVMODexMethod)
                .addStatement("$T method = vmoDex.methods[$N]", interpreter.Method.class, methodIdSpec)
                .addStatement("Object lastResult = null, lastException = null");

        if (verbosityLevels.contains(VerbosityLevel.INTERPRET_START)) {
            // NOTE printing args right here might lead to endless recursion when an object is passed whose
            // toString() is patched by the interpreter and whose implementation calls the caller
            builder.addStatement("System.out.println(\"interpret method: \" + method.definingClass + \"->\" + method.name)");
        }

        builder.addStatement("String[] paramTypes = method.paramTypes")
                .addStatement("Object regs[] = new Object[method.numRegs]")
                .add("\n")
                // move args to regs
                .beginControlFlow("for (int ai = $N.length - 1, ri = regs.length - 1; ai >= 0; ai--, ri--)", argsSpec)
                // Long and Double use register pairs
                .beginControlFlow("if (paramTypes[ai].charAt(0) == 'D' || paramTypes[ai].charAt(0) == 'J')")
                .addStatement("ri--")
                .endControlFlow()
                .addStatement("regs[ri] = Utils.externalTypeToVmType(paramTypes[ai], $N[ai])", argsSpec)
                .endControlFlow()
                .addStatement("byte[] insn = method.instructions")
                .addStatement("int pc = 0")
                .add("\n");

        // surround interpret loop with try and catch for class not found exception and so on, ...
        // TODO do not wrap execptions coming from interpreted code!
        // TODO what about exections coming from called code? Should be original!

        // actual loop
        builder.add("interpreterLoop:\n")
                .beginControlFlow("while(pc < insn.length)")
                .beginControlFlow("try")
                .addStatement("byte opcode = insn[pc++]")
                .beginControlFlow("switch (opcode)");

        for (InstructionType instructionType : this.usedInstructionTypes) {
            builder.beginControlFlow("case $L:", byteCodeMapping.toVmOpcode(instructionType));
            builder.add("// " + instructionType + "\n");
            if (verbosityLevels.contains(VerbosityLevel.EVERY_INSTRUCTION)) {
                builder.addStatement("System.out.println(\"" + instructionType + "\")");
            }
            instructionType.generateCase(builder);
            builder.endControlFlow();
        }

        builder.beginControlFlow("default:")
                .addStatement("throw new $T(\"Unsupported Instruction op: \" + opcode)", IllegalStateException.class)
                .endControlFlow()
                .endControlFlow() // close switch

                .nextControlFlow("catch (Throwable e)")
                .addStatement("lastException = e")
                .addStatement("int target = -1")
                .beginControlFlow("for (TryBlock tryBlock : method.tryBlocks)")
                .beginControlFlow("if (tryBlock.start <= pc && pc <= tryBlock.end)")
                // when using try-with resource exceptions of type <any> are stored in bytecode
                // dexlib2 parses their type to null
                .beginControlFlow("for (ExceptionHandler handler : tryBlock.exceptionHandlers)")
                .addStatement("$T exceptionType = handler.exceptionType == null ? null : Utils.loadClassByName(handler.exceptionType, classLoader)", Class.class)
                .beginControlFlow("if(exceptionType == null || exceptionType.isAssignableFrom(e.getClass()))")
                .addStatement("target = handler.address")
                .addStatement("break")
                .endControlFlow()
                .endControlFlow()
                .endControlFlow()
                .addStatement("if (target >= 0) break")
                .endControlFlow()

                .beginControlFlow("if (target >= 0)")
                .addStatement("pc = target")
                .addStatement("continue")
                .endControlFlow()

                .addStatement("throw e")
                .endControlFlow()

                .endControlFlow() // close while
                .addStatement("throw new IllegalStateException(\"interpreter left loop\")");
        return builder.build();
    }

    public JavaFile generateInterpreterSource() {
        TypeSpec.Builder classBuilder = TypeSpec.classBuilder(INTERPRETER_CLASS_NAME)
                .addModifiers(Modifier.PUBLIC, Modifier.FINAL);

        FieldSpec vmoDexField = FieldSpec.builder(VMODex.class, "vmoDex", Modifier.PRIVATE, Modifier.STATIC).build();
        classBuilder.addField(vmoDexField);

        // use classloader of interpreter for loading classes, because this is the one the classes are loaded in tests
        FieldSpec classLoaderField = FieldSpec.builder(ClassLoader.class, "classLoader", Modifier.PRIVATE, Modifier.STATIC, Modifier.FINAL)
                .initializer("Interpreter.class.getClassLoader()").build();
        classBuilder.addField(classLoaderField);

        String keyArg = "new byte[] {" + IntStream.range(0, key.length).mapToObj(idx -> Byte.toString(key[idx])).collect(Collectors.joining(", ")) + "}";
        MethodSpec getVMODexMethod = MethodSpec.methodBuilder("getVMODex")
                .addModifiers(Modifier.PRIVATE, Modifier.STATIC)
                .returns(VMODex.class)
                .beginControlFlow("if ($N == null)", vmoDexField)
                .addStatement("$N = $T.loadVMODexFile(classLoader.getResourceAsStream(\"" + VMODEX_FILE + "\"), " + keyArg + ")", vmoDexField, VMODex.class)
                .endControlFlow()
                .addStatement("return $N", vmoDexField)
                .build();
        classBuilder.addMethod(getVMODexMethod);

        ParameterSpec methodIdSpec = ParameterSpec.builder(int.class, "methodId").build();
        ParameterSpec argsSpec = ParameterSpec.builder(Object[].class, "args").build();
        MethodSpec interpretMethod = MethodSpec.methodBuilder(INTERPRETER_METHOD_NAME)
                .addModifiers(Modifier.PUBLIC, Modifier.STATIC)
                .returns(Object.class)
                .addParameter(methodIdSpec)
                .addException(Throwable.class)
                .addParameter(argsSpec)
                .addCode(generateInterpretBody(getVMODexMethod, methodIdSpec, argsSpec))
                .build();
        classBuilder.addMethod(interpretMethod);

        if (testMode) {
            ParameterSpec newVmoDexSpec = ParameterSpec.builder(VMODex.class, "newVmoDex").build();
            MethodSpec setVmoDexMethod = MethodSpec.methodBuilder("setVmoDex")
                    .addModifiers(Modifier.PUBLIC, Modifier.STATIC)
                    .returns(void.class)
                    .addParameter(newVmoDexSpec)
                    .addStatement("$N = $N", vmoDexField, newVmoDexSpec)
                    .build();
            classBuilder.addMethod(setVmoDexMethod);
        }

        TypeSpec interpreterType = classBuilder.build();

        return JavaFile.builder(INTERPRETER_PACKAGE_NAME, interpreterType).build();
    }

    public DexFile generateDex() {
        return Utils.doInTmpDir(tmpDir -> {
            compile(tmpDir);
            return CompileUtil.dxClasses(tmpDir);
        });
    }

    private static void copyResourceDir(String resourcePath, Path targetDir) {
        assert resourcePath.startsWith("/") && resourcePath.endsWith("/");
        URL uri = InterpreterGenerator.class.getResource("/interpreterSources");

        try {
            if (uri.getProtocol().equals("jar")) {
                String jarPath = uri.getPath().substring(5, uri.getPath().indexOf("!"));
                JarFile jar = new JarFile(URLDecoder.decode(jarPath, "UTF-8"));
                Enumeration<JarEntry> entries = jar.entries();
                while (entries.hasMoreElements()) {
                    String entryPath = "/" + entries.nextElement().getName();
                    if (entryPath.startsWith(resourcePath) && !entryPath.equals(resourcePath)) {
                        InputStream in = InterpreterGenerator.class.getResourceAsStream(entryPath);
                        Path target = targetDir.resolve(entryPath.substring(resourcePath.length()));
                        target.getParent().toFile().mkdirs();
                        OutputStream out = new FileOutputStream(target.toFile());
                        IOUtils.copy(in, out);
                    }
                }
            } else {
                FileUtils.copyDirectory(new File(uri.getFile()), targetDir.toFile());
            }
        } catch (IOException e) {
            throw new RuntimeException(e);
        }
    }

    /**
     * creates folder INTERPRETER_PACKAGE_NAME in targetDir, moves all interpreter sources there
     * and compiles them to class files using javac
     *
     * @param targetDir Directory where interpreter should be compiled
     */
    public void compile(Path targetDir) throws IOException {
        final Path interpreterDir = targetDir.resolve(INTERPRETER_PACKAGE_NAME);
        final Path interpreterSourceFile = interpreterDir.resolve(INTERPRETER_CLASS_NAME + ".java");

        copyResourceDir("/interpreterSources/", interpreterDir);

        JavaFile interpreterSrc = generateInterpreterSource();
        FileWriter fileWriter = new FileWriter(interpreterSourceFile.toFile());
        interpreterSrc.writeTo(fileWriter);
        fileWriter.close();

        CompileUtil.compileJava(interpreterDir, Collections.singletonList("-g"));
    }

    public static DexFile generateInterpreterDex(ByteCodeMapping byteCodeMapping, Set<InstructionType> usedInstructionTypes, Set<VerbosityLevel> verbosityLevels, byte[] key, boolean testMode) {
        return new InterpreterGenerator(byteCodeMapping, usedInstructionTypes, verbosityLevels, key, testMode).generateDex();
    }
}
