package de.marstra.dvo.obfuscator;

import de.marstra.dvo.obfuscator.analysis.ParsedDex;
import de.marstra.dvo.obfuscator.util.GenericMethodReference;
import de.marstra.dvo.obfuscator.util.Utils;
import lombok.Data;
import lombok.extern.slf4j.Slf4j;
import org.jf.dexlib2.AccessFlags;
import org.jf.dexlib2.Opcodes;
import org.jf.dexlib2.builder.BuilderInstruction;
import org.jf.dexlib2.builder.instruction.*;
import org.jf.dexlib2.iface.*;
import org.jf.dexlib2.iface.instruction.Instruction;
import org.jf.dexlib2.iface.instruction.ReferenceInstruction;
import org.jf.dexlib2.iface.instruction.formats.*;
import org.jf.dexlib2.iface.reference.MethodReference;
import org.jf.dexlib2.iface.reference.Reference;
import org.jf.dexlib2.iface.reference.TypeReference;
import org.jf.dexlib2.immutable.*;
import org.jf.dexlib2.immutable.reference.ImmutableTypeReference;

import java.util.*;
import java.util.stream.Collectors;


@Slf4j
public class InterpreterRenamer {
    @Data
    public static class RenameResult {
        private final GenericMethodReference interpret;
        private final DexFile interpreterDex;
    }

    private final ParsedDex parsedDex;
    private final DexFile interpreterDex;
    private Map<Reference, Reference> refs = new HashMap<>();
    private Set<String> usedNames = new HashSet<>();

    private InterpreterRenamer(ParsedDex parsedDex, DexFile interpreterDex) {
        this.parsedDex = parsedDex;
        this.interpreterDex = interpreterDex;
    }

    private String getUniqueName() {
        while (true) {
            String name = Utils.randomName();
            if (!usedNames.contains(name)) {
                usedNames.add(name);
                return name;
            }
        }
    }

    private String getRandomClassName(ParsedDex parsedDex) {
        while (true) {
            String name = "L" + parsedDex.getRandomPackagePath() + "/" + getUniqueName() + ";";
            if (parsedDex.findClass(name) == null)
                return name;
        }
    }

    public static RenameResult renameInterpreter(ParsedDex parsedDex, DexFile interpreterDex) {
        log.info("rename interpreter package");
        InterpreterRenamer ir = new InterpreterRenamer(parsedDex, interpreterDex);
        ir.chooseNames();
        DexFile renamed = ir.doRename();
        MethodReference newInterpret = (MethodReference) ir.refs.get(InterpreterGenerator.INTERPRETER_METHOD.toDexLibMethodReference());
        if (newInterpret == null) throw new IllegalStateException();
        return new RenameResult(new GenericMethodReference(newInterpret), renamed);
    }

    private DexFile doRename() {
        Set<ClassDef> classes = new HashSet<>();

        for (ClassDef origCd : interpreterDex.getClasses()) {
            List<Field> fields = new LinkedList<>();
            List<Method> methods = new LinkedList<>();

            TypeReference newCd = (TypeReference) refs.get(origCd);
            if (newCd == null) throw new IllegalStateException();

            for (Field f : origCd.getFields()) {
                Field field = (Field) refs.get(f);
                if (field == null) throw new IllegalStateException();
                if (!field.getDefiningClass().equals(newCd.getType())) throw new IllegalStateException();
                fields.add(field);
            }

            for (Method m : origCd.getMethods()) {
                Method method = (Method) refs.get(m);
                log.debug("renamed " + m + " to " + method);
                if (method == null) throw new IllegalStateException();
                if (!method.getDefiningClass().equals(newCd.getType())) throw new IllegalStateException();
                methods.add(new ImmutableMethod(
                        method.getDefiningClass(),
                        method.getName(),
                        method.getParameters(),
                        method.getReturnType(),
                        method.getAccessFlags(),
                        method.getAnnotations(),
                        patchMethodImpl(method.getImplementation())
                ));
            }

            classes.add(new ImmutableClassDef(
                    newCd.getType(),
                    origCd.getAccessFlags(),
                    origCd.getSuperclass(),
                    origCd.getInterfaces(),
                    null,
                    origCd.getAnnotations(),
                    fields,
                    methods
            ));
        }

        return new DexFile() {
            public Set<? extends ClassDef> getClasses() {
                return classes;
            }

            public Opcodes getOpcodes() {
                return interpreterDex.getOpcodes();
            }
        };
    }

    private MethodImplementation patchMethodImpl(MethodImplementation impl) {
        List<Instruction> insn = new LinkedList<>();

        for (Instruction ins : impl.getInstructions()) {
            BuilderInstruction newIns = replaceRefs(ins);
            if (newIns != null)
                insn.add(newIns);
            else
                insn.add(ins);
        }
        return new ImmutableMethodImplementation(
                impl.getRegisterCount(),
                insn,
                impl.getTryBlocks().stream().map(t ->
                        new ImmutableTryBlock(
                                t.getStartCodeAddress(),
                                t.getCodeUnitCount(),
                                t.getExceptionHandlers().stream().map(eh ->
                                        new ImmutableExceptionHandler(getPatchedType(eh.getExceptionType()), eh.getHandlerCodeAddress())
                                ).collect(Collectors.toList())
                        )).collect(Collectors.toList()),
                null
        );
    }


    private BuilderInstruction replaceRefs(Instruction ins) {
        if (!(ins instanceof ReferenceInstruction)) return null;
        ReferenceInstruction refIns = (ReferenceInstruction) ins;
        Reference ref = refIns.getReference();
        Reference newRef;
        if (ref instanceof TypeReference) {
            TypeReference typeRef = (TypeReference) ref;
            newRef = new ImmutableTypeReference(getPatchedType(typeRef.getType()));
        } else {
            newRef = refs.get(ref);
        }
        // reference instruction refering code outside interpreter package
        if (newRef == null) return null;
        switch (ins.getOpcode().format) {
            case Format21c: {
                Instruction21c typed = (Instruction21c) ins;
                return new BuilderInstruction21c(ins.getOpcode(), typed.getRegisterA(), newRef);
            }
            case Format22c: {
                Instruction22c typed = (Instruction22c) ins;
                return new BuilderInstruction22c(ins.getOpcode(), typed.getRegisterA(), typed.getRegisterB(), newRef);
            }
            case Format3rc: {
                Instruction3rc typed = (Instruction3rc) ins;
                return new BuilderInstruction3rc(ins.getOpcode(), typed.getStartRegister(), typed.getRegisterCount(), newRef);
            }
            case Format31c: {
                Instruction31c typed = (Instruction31c) ins;
                return new BuilderInstruction31c(ins.getOpcode(), typed.getRegisterA(), newRef);
            }
            case Format35c: {
                Instruction35c typed = (Instruction35c) ins;
                return new BuilderInstruction35c(ins.getOpcode(), typed.getRegisterCount(), typed.getRegisterC(),
                        typed.getRegisterD(), typed.getRegisterE(), typed.getRegisterF(), typed.getRegisterG(), newRef);
            }
            case Format4rcc: {
                Instruction4rcc typed = (Instruction4rcc) ins;
                Reference newRef2 = refs.get(typed.getReference2());
                return new BuilderInstruction4rcc(ins.getOpcode(), typed.getStartRegister(), typed.getRegisterCount(), newRef, newRef2);
            }
            case Format45cc: {
                Instruction45cc typed = (Instruction45cc) ins;
                Reference newRef2 = refs.get(typed.getReference2());
                return new BuilderInstruction45cc(typed.getOpcode(), typed.getRegisterCount(), typed.getRegisterC(),
                        typed.getRegisterD(), typed.getRegisterE(), typed.getRegisterF(), typed.getRegisterG(),
                        newRef, newRef2);
            }
            default:
                throw new IllegalStateException();
        }
    }

    private String getPatchedType(String orig) {
        int arrayEnd = orig.lastIndexOf('[') + 1;
        String arrayPart = orig.substring(0, arrayEnd);
        String typePart = orig.substring(arrayEnd);

        ImmutableTypeReference origType = new ImmutableTypeReference(typePart);
        CharSequence newType = (TypeReference) refs.get(origType);
        if (newType != null) return arrayPart + newType.toString();
        return orig;
    }

    private void chooseNames() {
        // add classes first so afterwards fields, parameters and return types can be adjusted
        for (ClassDef cd : interpreterDex.getClasses()) {
            String className = getRandomClassName(parsedDex);
            refs.put(cd, new ImmutableTypeReference(className));
        }

        for (ClassDef cd : interpreterDex.getClasses()) {
            TypeReference classType = (TypeReference) refs.get(cd);
            if (classType == null) throw new IllegalStateException();

            for (Method m : cd.getMethods()) {
                boolean isConstructor = (m.getAccessFlags() & AccessFlags.CONSTRUCTOR.getValue()) != 0;
                // constructors always have name <init> or <clinit>
                String newName = isConstructor ? m.getName() : getUniqueName();

                refs.put(m, new ImmutableMethod(
                        classType.getType(),
                        newName,
                        m.getParameters().stream().map(p ->
                                new ImmutableMethodParameter(
                                        getPatchedType(p.getType()),
                                        p.getAnnotations(),
                                        getUniqueName() // TODO does this work, where is method parameter even used??
                                )).collect(Collectors.toList()),
                        getPatchedType(m.getReturnType()),
                        m.getAccessFlags(),
                        m.getAnnotations(),
                        m.getImplementation()
                ));
            }
            for (Field f : cd.getFields()) {
                refs.put(f, new ImmutableField(
                        classType.getType(),
                        getUniqueName(),
                        getPatchedType(f.getType()),
                        f.getAccessFlags(),
                        f.getInitialValue(),
                        f.getAnnotations()
                ));
            }
        }
    }
}
