package de.marstra.dvo.obfuscator;

import com.squareup.javapoet.*;
import de.marstra.dvo.obfuscator.analysis.ApkFile;
import de.marstra.dvo.obfuscator.analysis.DexMethod;
import de.marstra.dvo.obfuscator.analysis.ParsedDex;
import de.marstra.dvo.obfuscator.util.CompileUtil;
import de.marstra.dvo.obfuscator.util.GenericMethodReference;
import de.marstra.dvo.obfuscator.util.Utils;
import lombok.extern.slf4j.Slf4j;
import org.jf.dexlib2.Opcode;
import org.jf.dexlib2.builder.MutableMethodImplementation;
import org.jf.dexlib2.builder.instruction.BuilderInstruction35c;
import org.jf.dexlib2.iface.ClassDef;
import org.jf.dexlib2.iface.DexFile;
import org.jf.dexlib2.iface.Method;
import org.jf.dexlib2.immutable.ImmutableMethod;

import javax.lang.model.element.Modifier;
import java.io.File;
import java.io.FileInputStream;
import java.io.IOException;
import java.io.InputStream;
import java.net.URI;
import java.nio.charset.StandardCharsets;
import java.nio.file.*;
import java.util.*;
import java.util.concurrent.Semaphore;
import java.util.concurrent.atomic.AtomicInteger;
import java.util.stream.Collectors;
import java.util.zip.CRC32;
import java.util.zip.ZipFile;

import static de.marstra.dvo.obfuscator.util.Utils.randomName;
import static interpreter.Utils.intToBytes;

@Slf4j
public class ProtectionInjector {
    // TODO atomic integer ranges variabel
    private static class ProtectionGenerator {
        private static final ClassName ANDROID_CONTEXT = ClassName.bestGuess("android.content.Context");
        private static final int ENSURE_INTEGRITY = 0;
        private static final int RUN = 1;
        private static final int START_PROTECTION_THREAD = 2;

        private final String typeDescriptor;
        private final List<String> checksumLocations;
        private final String classString;
        private final String packageString;
        private final ClassName className;
        private final boolean debug;
        private final GenericMethodReference[] methods;

        private ProtectionGenerator(String typeDescriptor, List<String> checksumLocations, boolean debug) {
            this.typeDescriptor = typeDescriptor;
            this.debug = debug;
            this.checksumLocations = checksumLocations;
            String[] parsedType = Utils.parseType(typeDescriptor);
            this.packageString = String.join(".", Arrays.copyOf(parsedType, parsedType.length - 1));
            this.classString = parsedType[parsedType.length - 1];
            this.className = ClassName.get(packageString, classString);
            this.methods = new GenericMethodReference[3];
            this.methods[ENSURE_INTEGRITY] = new GenericMethodReference(typeDescriptor, getName("ensureIntegrity"), "V", new String[]{});
            this.methods[RUN] = new GenericMethodReference(typeDescriptor, "run", "V", new String[]{});
            this.methods[START_PROTECTION_THREAD] = new GenericMethodReference(typeDescriptor, getName("startProtectionThread"), "V", "Landroid/content/Context;");
        }

        private MethodSpec genEnsureIntegrity(FieldSpec instanceField) {
            CodeBlock.Builder cb = CodeBlock.builder();
            cb.beginControlFlow("if ($N == null)", instanceField);
            if (debug) cb.addStatement("System.out.println(\"protection measures not present\")");
            cb.addStatement("System.exit((int) (Math.random() * 255))");
            cb.endControlFlow();
            cb.beginControlFlow("if (!$N.isAlive())", instanceField);
            if (debug) cb.addStatement("System.out.println(\"integrity check failed, shutdown\")");
            cb.addStatement("System.exit((int) (Math.random() * 255))");
            cb.endControlFlow();

            return MethodSpec.methodBuilder(this.methods[ENSURE_INTEGRITY].getName())
                    .addModifiers(Modifier.PUBLIC, Modifier.STATIC)
                    .returns(void.class)
                    .addCode(cb.build())
                    .build();
        }

        private MethodSpec genThreadRun(FieldSpec semaphoreField, FieldSpec contextField) {
            CodeBlock.Builder cb = CodeBlock.builder();
            cb.addStatement("boolean crash = false");
            cb.addStatement("$N.release()", semaphoreField);

            cb.beginControlFlow("try");

            cb.beginControlFlow("while (!crash)");

            // debug check
            cb.beginControlFlow("if (($N.getApplicationInfo().flags & 2) != 0)", contextField);
            cb.addStatement("crash = true");
            if (debug) cb.addStatement("System.out.println(\"debug positive\")");
            cb.addStatement("break");
            cb.endControlFlow();

            // root check 1
            cb.addStatement("String[] path = System.getenv(\"PATH\").split(\":\")");
            cb.beginControlFlow("for (String part : path)");
            cb.beginControlFlow("if (new $T(part, \"su\").exists())", File.class);
            cb.addStatement("crash = true");
            if (debug) cb.addStatement("System.out.println(\"root1 positive\")");
            cb.addStatement("break");
            cb.endControlFlow();
            cb.endControlFlow();

            // root check 2
            cb.addStatement("String tags = android.os.Build.TAGS");
            cb.beginControlFlow("if (tags != null && tags.contains(\"test-keys\"))");
            cb.addStatement("crash = true");
            if (debug) cb.addStatement("System.out.println(\"root2 positive\")");
            cb.endControlFlow();

            // root check 3
            cb.addStatement("String[] badFiles = new String[]{\"/system/app/Superuser.apk\", \"/system/xbin/daemonsu\", \"/system/etc/init.d/99SuperSUDaemon\", \"/system/bin/.ext/.su\", \"/system/etc/.has_su_daemon\", \"/system/etc/.installed_su_daemon\", \"/dev/com.koushikdutta.superuser.daemon/\"}");
            cb.beginControlFlow("for (String badFile : badFiles)");
            cb.beginControlFlow("if (new $T(badFile).exists())", File.class);
            cb.addStatement("crash = true");
            if (debug) cb.addStatement("System.out.println(\"root3 positive\")");
            cb.addStatement("break");
            cb.endControlFlow();
            cb.endControlFlow();

            // tamper check crc
            cb.addStatement("$1T zipFile = new $1T($2N.getPackageCodePath())", ZipFile.class, contextField);
            cb.add("$1T[] checksumFileStreams = new $1T[]{\n", InputStream.class);
            cb.indent();
            cb.add(checksumLocations.stream().map(loc -> classString + ".class.getResourceAsStream(\"" + loc + "\")")
                    .collect(Collectors.joining(",\n")));
            cb.unindent();
            cb.add("\n};\n");
            // TODO classes1 dex?
            cb.addStatement("String[] prooveFiles = new String[]{\"classes.dex\", \"classes.vmodex\"}");
            cb.beginControlFlow("for (String prooveFile : prooveFiles)");
            cb.addStatement("long actual = zipFile.getEntry(prooveFile).getCrc()");
            cb.addStatement("long expected = 0");
            cb.addStatement("byte[] buf = new byte[4]");
            cb.beginControlFlow("for (InputStream in : checksumFileStreams)");
            cb.addStatement("in.read(buf)");
            cb.addStatement("expected ^= (((buf[0] & 0xFF) << 24) | ((buf[1] & 0xFF) << 16) | ((buf[2] & 0xFF) << 8) | (buf[3] & 0xFF)) & 0xFFFFFFFFL");
            cb.endControlFlow();
            cb.beginControlFlow("if (actual != expected)");
            if (debug)
                cb.addStatement("System.out.println(\"tampering detected of file: \" + prooveFile + \" expected: \" + expected + \", actual: \" + actual)");
            cb.addStatement("crash = true");
            cb.addStatement("break");
            cb.endControlFlow();
            cb.endControlFlow();
            cb.beginControlFlow("for (InputStream in : checksumFileStreams)");
            cb.addStatement("in.close()");
            cb.endControlFlow();

            // check for frida and xposed is attached
            cb.addStatement("$1T fin = new $1T(\"/proc/self/maps\")", FileInputStream.class);
            cb.addStatement("$1T scanner = new $1T(fin)", Scanner.class);
            cb.beginControlFlow("while (scanner.hasNextLine())");
            cb.addStatement("String line = scanner.nextLine()");
            cb.beginControlFlow("if (line.contains(\"frida\") || line.contains(\"xposed\"))");
            if (debug) cb.addStatement("System.out.println(\"hooking framework detected\")");
            cb.addStatement("crash = true");
            cb.addStatement("break");
            cb.endControlFlow();
            cb.endControlFlow();
            cb.addStatement("fin.close()");

            cb.addStatement("$T.sleep(1000)", Thread.class);

            cb.endControlFlow();
            cb.nextControlFlow("catch ($T t)", Throwable.class);
            if (debug) cb.addStatement("System.out.println(\"protection thread failed\")");
            if (debug) cb.addStatement("t.printStackTrace()");
            cb.endControlFlow();

            return MethodSpec.methodBuilder(this.methods[RUN].getName())
                    .addModifiers(Modifier.PUBLIC)
                    .returns(void.class)
                    .addCode(cb.build())
                    .build();
        }

        private MethodSpec genStartThread(FieldSpec instanceField, FieldSpec semaphoreField) {
            ParameterSpec contextParam = ParameterSpec.builder(ANDROID_CONTEXT, "ctx").build();
            CodeBlock.Builder cb = CodeBlock.builder();
            cb.beginControlFlow("if ($N == null)", instanceField);
            cb.addStatement("$N = new $T(0)", semaphoreField, Semaphore.class);
            cb.addStatement("$N = new $T($N)", instanceField, className, contextParam);
            cb.addStatement("$N.start()", instanceField);
            cb.beginControlFlow("try");
            cb.addStatement("$N.acquire()", semaphoreField);
            cb.nextControlFlow("catch ($T e)", InterruptedException.class);
            if (debug) cb.addStatement("System.out.println(\"failed to acquire semaphore\")");
            cb.endControlFlow();
            cb.endControlFlow();

            return MethodSpec.methodBuilder(this.methods[START_PROTECTION_THREAD].getName())
                    .addModifiers(Modifier.PUBLIC, Modifier.STATIC)
                    .addParameter(contextParam)
                    .returns(void.class)
                    .addCode(cb.build())
                    .build();
        }

        private MethodSpec genConstructor(FieldSpec contextField) {
            ParameterSpec contextParam = ParameterSpec.builder(ANDROID_CONTEXT, "ctx").build();
            return MethodSpec.constructorBuilder()
                    .addModifiers(Modifier.PRIVATE)
                    .addParameter(contextParam)
                    .addCode(CodeBlock.builder().addStatement("$N = $N", contextField, contextParam).build())
                    .build();
        }

        private String getName(String name) {
            if (debug) return name;
            return randomName();
        }

        private JavaFile genJavaSource() {
            FieldSpec instanceField = FieldSpec.builder(className, getName("instance"), Modifier.PRIVATE, Modifier.STATIC).build();
            FieldSpec semaphoreField = FieldSpec.builder(Semaphore.class, getName("semaphore"), Modifier.PRIVATE, Modifier.STATIC).build();
            FieldSpec contextField = FieldSpec.builder(ANDROID_CONTEXT, getName("context"), Modifier.PRIVATE).build();

            TypeSpec.Builder tb = TypeSpec.classBuilder(className)
                    .addModifiers(Modifier.PUBLIC, Modifier.FINAL)
                    .superclass(Thread.class);
            tb.addField(instanceField);
            tb.addField(semaphoreField);
            tb.addField(contextField);
            tb.addMethod(genConstructor(contextField));
            tb.addMethod(genThreadRun(semaphoreField, contextField));
            tb.addMethod(genEnsureIntegrity(instanceField));
            tb.addMethod(genStartThread(instanceField, semaphoreField));

            TypeSpec typeSpec = tb.build();
            return JavaFile.builder(packageString, typeSpec).build();
        }

        ClassDef getDexClass() {
            Set<? extends ClassDef> classes = generateDex().getClasses();
            if (classes.size() != 1) throw new IllegalStateException("more than one protection class");
            return classes.iterator().next();
        }

        private String generateSource() {
            StringBuilder builder = new StringBuilder();
            try {
                genJavaSource().writeTo(builder);
            } catch (IOException e) {
                throw new RuntimeException(e);
            }
            return builder.toString();
        }

        private DexFile generateDex() {
            return Utils.doInTmpDir(tmpDir -> {
                Path srcFile = tmpDir.resolve(this.classString + ".java");
                Files.write(srcFile, generateSource().getBytes(StandardCharsets.UTF_8));
                CompileUtil.compileJava(tmpDir, Arrays.asList("-cp", CompileUtil.pathToAndroidJar().toString()));
                return CompileUtil.dxClasses(tmpDir);
            });
        }

        public GenericMethodReference getGenericMethodReference(int id) {
            if (id < 0 || id > START_PROTECTION_THREAD) throw new IllegalArgumentException();
            return methods[id];
        }

        public List<GenericMethodReference> getMethodReferences() {
            return Arrays.asList(methods);
        }
    }

    public static List<GenericMethodReference> addProtectionMeasures(ParsedDex parsedDex, List<GenericMethodReference> entryPoints,
                                                                     Set<GenericMethodReference> methodsToObfuscate, List<String> checksumLocations, boolean debug) {
        String type = debug ? "LProtectionMeasures;" : "L" + parsedDex.getRandomPackagePath() + "/" + randomName() + ";";
        ProtectionGenerator protGen = new ProtectionGenerator(type, checksumLocations, debug);

        for (GenericMethodReference ep : entryPoints) {
            DexMethod dexMethod = parsedDex.findMethod(ep);
            if (dexMethod == null)
                throw new IllegalArgumentException("Entrypoint to be patched not found: " + ep);
            dexMethod.setMethod(addProtectionThreadCreation(
                    dexMethod.getMethod(),
                    protGen.getGenericMethodReference(ProtectionGenerator.START_PROTECTION_THREAD)
            ));
        }

        for (GenericMethodReference gmr : methodsToObfuscate) {
            DexMethod dexMethod = parsedDex.findMethod(gmr);
            if (dexMethod == null)
                throw new IllegalArgumentException("Method to be obufscated not found: " + gmr);
            dexMethod.setMethod(addEnsureIntegrityCall(
                    dexMethod.getMethod(),
                    protGen.getGenericMethodReference(ProtectionGenerator.ENSURE_INTEGRITY)
            ));
        }
        log.info("added protection methods: " + protGen.getMethodReferences().stream()
                .map(GenericMethodReference::toString)
                .collect(Collectors.joining(" ")));
        parsedDex.addClass(protGen.getDexClass());
        return protGen.getMethodReferences();
    }

    private static Method addEnsureIntegrityCall(Method toPatch, GenericMethodReference ensureIntegrity) {
        MutableMethodImplementation mimpl = new MutableMethodImplementation(toPatch.getImplementation());
        mimpl.addInstruction(0, new BuilderInstruction35c(Opcode.INVOKE_STATIC,
                0, 0, 0, 0, 0, 0, ensureIntegrity.toDexLibMethodReference()));
        return new ImmutableMethod(
                toPatch.getDefiningClass(),
                toPatch.getName(),
                toPatch.getParameters(),
                toPatch.getReturnType(),
                toPatch.getAccessFlags(),
                toPatch.getAnnotations(),
                mimpl
        );
    }

    private static Method addProtectionThreadCreation(Method toPatch, GenericMethodReference startProtection) {
        MutableMethodImplementation mimpl = new MutableMethodImplementation(toPatch.getImplementation());
        int thisReg = mimpl.getRegisterCount() - 2;
        mimpl.addInstruction(0, new BuilderInstruction35c(Opcode.INVOKE_STATIC,
                1, thisReg, 0, 0, 0, 0, startProtection.toDexLibMethodReference()));
        return new ImmutableMethod(
                toPatch.getDefiningClass(),
                toPatch.getName(),
                toPatch.getParameters(),
                toPatch.getReturnType(),
                toPatch.getAccessFlags(),
                toPatch.getAnnotations(),
                mimpl
        );
    }

    public static List<String> getChecksumLocations(ApkFile apk, int num) {
        String[] folderCandidates = new String[]{"res/animator", "res/anim", "res/animator", "res/drawable", "res/layout", "res/color", "META-INF"};
        Random random = new Random();

        List<String> checkSumCandidates = new LinkedList<>();
        for (int i = 0; i < num; i++) {
            String folder = folderCandidates[random.nextInt(folderCandidates.length)];
            String entry;
            do {
                entry = folder + "/" + randomName() + ".xml";
            } while (apk.exits(entry));
            checkSumCandidates.add(entry);

        }
        return checkSumCandidates;
    }

    public static byte[] crc32(byte[] data) {
        CRC32 crc32 = new CRC32();
        crc32.update(data);
        long value = crc32.getValue();
        return intToBytes((int) value);
    }

    public static byte[][] split(byte[] data, int num) {
        Random random = new Random();
        byte[][] split = new byte[num][];
        for (int i = 1; i < num; i++) {
            split[i] = new byte[data.length];
            random.nextBytes(split[i]);
        }

        split[0] = new byte[data.length];
        for (int i = 0; i < data.length; i++) {
            byte cur = 0;
            for (int j = 1; j < num; j++) {
                cur ^= split[j][i];
            }
            cur ^= data[i];
            split[0][i] = cur;
        }
        return split;
    }

    public static byte[] addChecksumFiles(byte[] apk, List<String> checksumLocations) {
        log.info("add checksum files: " + checksumLocations.stream().collect(Collectors.joining(" ")));
        return Utils.doWithTmpFile("obfapk", f -> {
            Files.write(f.toPath(), apk);
            FileSystem fs = FileSystems.newFileSystem(URI.create("jar:" + f.toURI()), Collections.singletonMap("create", "false"));
            byte[] dexContent = Files.readAllBytes(fs.getPath("classes.dex"));
            byte[] vmodexContent = Files.readAllBytes(fs.getPath("classes.vmodex"));

            for (byte[] content : new byte[][]{dexContent, vmodexContent}) {
                byte[] crc = crc32(content);
                byte[][] split = split(crc, checksumLocations.size());
                for (int i = 0; i < checksumLocations.size(); i++) {
                    Path location = fs.getPath(checksumLocations.get(i));
                    Files.write(location, split[i], StandardOpenOption.CREATE, StandardOpenOption.APPEND);
                }
            }

            fs.close();
            ApkFile.signApk(f);
            return Files.readAllBytes(f.toPath());
        });
    }
}



