package de.marstra.dvo.obfuscator;

import de.marstra.dvo.obfuscator.analysis.DexMethod;
import de.marstra.dvo.obfuscator.analysis.ParsedDex;
import de.marstra.dvo.obfuscator.util.GenericMethodReference;
import de.marstra.dvo.obfuscator.util.PrimitiveWrapperUtils;
import interpreter.Utils;
import lombok.extern.slf4j.Slf4j;
import org.jf.dexlib2.AccessFlags;
import org.jf.dexlib2.Opcode;
import org.jf.dexlib2.builder.MethodImplementationBuilder;
import org.jf.dexlib2.builder.instruction.*;
import org.jf.dexlib2.iface.DexFile;
import org.jf.dexlib2.iface.Method;
import org.jf.dexlib2.iface.MethodParameter;
import org.jf.dexlib2.immutable.ImmutableMethod;
import org.jf.dexlib2.immutable.reference.ImmutableTypeReference;
import org.jf.dexlib2.util.MethodUtil;
import org.objectweb.asm.ClassReader;
import org.objectweb.asm.ClassVisitor;
import org.objectweb.asm.ClassWriter;
import org.objectweb.asm.MethodVisitor;

import java.io.FileInputStream;
import java.io.IOException;
import java.nio.file.Files;
import java.nio.file.Path;
import java.util.LinkedList;
import java.util.List;
import java.util.Optional;
import java.util.stream.Stream;

import static de.marstra.dvo.obfuscator.VMODexCompiler.ConvertedMethodReference;
import static de.marstra.dvo.obfuscator.util.GenericMethodReference.convertDexTypeToJVMType;
import static org.objectweb.asm.Opcodes.*;

@Slf4j
public class StubInjector {
    private static Method patchMethod(Method toPatch, GenericMethodReference interpret, int methodId) {
        List<String> parameterTypes = new LinkedList<>();
        if ((toPatch.getAccessFlags() & AccessFlags.STATIC.getValue()) == 0) {
            parameterTypes.add(toPatch.getDefiningClass());
        }
        for (MethodParameter p : toPatch.getParameters()) {
            parameterTypes.add(p.getType());
        }

        final int numArgs = parameterTypes.size();
        final int regsUsed = 4;
        final int totalRegs = regsUsed + MethodUtil.getParameterRegisterCount(toPatch);
        MethodImplementationBuilder builder = new MethodImplementationBuilder(totalRegs);

        // create array for arguments with reference in register 0
        builder.addInstruction(new BuilderInstruction31i(Opcode.CONST, 1, numArgs));
        builder.addInstruction(new BuilderInstruction22c(Opcode.NEW_ARRAY, 0, 1, new ImmutableTypeReference("[Ljava/lang/Object;")));

        // move parameters into array, call wrapper methods if necessary
        int curParReg = regsUsed;
        int curPar = 0;
        for (String ptype : parameterTypes) {
            // load array index in register 3
            builder.addInstruction(new BuilderInstruction31i(Opcode.CONST, 3, curPar));

            GenericMethodReference wrap = PrimitiveWrapperUtils.getWrapMethod(ptype);

            if (wrap != null) {
                if (Utils.isWideType(ptype)) {
                    // long and double use two registers
                    // move parameter register to low register number, because in case of large parameter numbers
                    // the register containing the parameter is to large to be addressed by INVOKE_STATIC
                    builder.addInstruction(new BuilderInstruction22x(Opcode.MOVE_WIDE_FROM16, 1, curParReg));
                    builder.addInstruction(new BuilderInstruction35c(Opcode.INVOKE_STATIC,
                            2, 1, 2, 0, 0, 0, wrap.toDexLibMethodReference()));
                    curParReg += 1;
                } else {
                    builder.addInstruction(new BuilderInstruction22x(Opcode.MOVE_FROM16, 1, curParReg));
                    builder.addInstruction(new BuilderInstruction35c(Opcode.INVOKE_STATIC,
                            1, 1, 0, 0, 0, 0, wrap.toDexLibMethodReference()));
                }
                builder.addInstruction(new BuilderInstruction11x(Opcode.MOVE_RESULT_OBJECT, 2));
                builder.addInstruction(new BuilderInstruction23x(Opcode.APUT_OBJECT, 2, 0, 3));
            } else {
                // move parameter register to low register number, because in case of large parameter numbers
                // the register containing the parameter is to large to be addressed by APUT_OBJECT
                builder.addInstruction(new BuilderInstruction22x(Opcode.MOVE_OBJECT_FROM16, 1, curParReg));
                builder.addInstruction(new BuilderInstruction23x(Opcode.APUT_OBJECT, 1, 0, 3));
            }

            curPar += 1;
            curParReg += 1;
        }

        builder.addInstruction(new BuilderInstruction31i(Opcode.CONST, 3, methodId));
        builder.addInstruction(new BuilderInstruction35c(Opcode.INVOKE_STATIC,
                2, 3, 0, 0, 0, 0, interpret.toDexLibMethodReference()));

        builder.addInstruction(new BuilderInstruction11x(Opcode.MOVE_RESULT_OBJECT, 0));

        String returnType = toPatch.getReturnType();
        GenericMethodReference unwrapResult = PrimitiveWrapperUtils.getUnwrapMethod(returnType);

        if (returnType.charAt(0) == 'V') {
            builder.addInstruction(new BuilderInstruction10x(Opcode.RETURN_VOID));
        } else if (unwrapResult != null) {
            builder.addInstruction(new BuilderInstruction21c(Opcode.CHECK_CAST,
                    0, new ImmutableTypeReference(unwrapResult.getDefiningClass())));
            builder.addInstruction(new BuilderInstruction35c(Opcode.INVOKE_VIRTUAL,
                    1, 0, 0, 0, 0, 0, unwrapResult.toDexLibMethodReference()));
            if (Utils.isWideType(returnType)) {
                builder.addInstruction(new BuilderInstruction11x(Opcode.MOVE_RESULT_WIDE, 0));
                builder.addInstruction(new BuilderInstruction11x(Opcode.RETURN_WIDE, 0));
            } else {
                builder.addInstruction(new BuilderInstruction11x(Opcode.MOVE_RESULT, 0));
                builder.addInstruction(new BuilderInstruction11x(Opcode.RETURN, 0));
            }
        } else {
            builder.addInstruction(new BuilderInstruction21c(Opcode.CHECK_CAST,
                    0, new ImmutableTypeReference(toPatch.getReturnType())));
            builder.addInstruction(new BuilderInstruction11x(Opcode.RETURN_OBJECT, 0));
        }

        return new ImmutableMethod(
                toPatch.getDefiningClass(),
                toPatch.getName(),
                toPatch.getParameters(),
                toPatch.getReturnType(),
                toPatch.getAccessFlags(),
                toPatch.getAnnotations(),
                builder.getMethodImplementation()
        );
    }

    /**
     * Patches Methods toPatch in origDex by inserting calls to interpreter
     *
     * @param parsedDex      dex files as parsed data structure, MODIFIED during execution
     * @param toPatch        methods that should be patched
     * @param interpreterDex Dexfile containing the Intrepreter and all the classes needed
     */
    public static void injectInterpreter(ParsedDex parsedDex, List<ConvertedMethodReference> toPatch, DexFile interpreterDex, boolean rename) {
        GenericMethodReference interpret;
        if (rename) {
            InterpreterRenamer.RenameResult res = InterpreterRenamer.renameInterpreter(parsedDex, interpreterDex);
            interpret = res.getInterpret();
            res.getInterpreterDex().getClasses().forEach(parsedDex::addClass);
        } else {
            interpret = InterpreterGenerator.INTERPRETER_METHOD;
            interpreterDex.getClasses().forEach(parsedDex::addClass);
        }

        for (ConvertedMethodReference cmr : toPatch) {
            GenericMethodReference gmr = cmr.getGenericMethodReference();
            DexMethod dexMethod = parsedDex.findMethod(gmr);
            if (dexMethod == null)
                throw new IllegalArgumentException("Method to be patched not found: " + gmr);

            log.debug("Patch method " + gmr);
            Method patched = patchMethod(dexMethod.getMethod(), interpret, cmr.getVmoDexMethodId());
            dexMethod.setMethod(patched);
        }
    }

    public static void injectInterpreterClassFiles(List<ConvertedMethodReference> toPatch, Path[] classFilesToPatch) throws IOException {
        final List<ConvertedMethodReference> remaining = new LinkedList<>(toPatch);

        for (int i = 0; i < classFilesToPatch.length; i++) {
            ClassReader cr = new ClassReader(new FileInputStream(classFilesToPatch[i].toFile()));
            ClassWriter cw = new ClassWriter(ClassWriter.COMPUTE_FRAMES);

            cr.accept(new ClassVisitor(ASM5, cw) {
                private String currentClass;

                @Override
                public void visit(int version, int access, String name, String signature, String superName, String[] interfaces) {
                    cw.visit(version, access, name, signature, superName, interfaces);
                    this.currentClass = name;
                }

                @Override
                public MethodVisitor visitMethod(int access, String name, String descriptor, String signature, String[] exceptions) {
                    final MethodVisitor origMv = cv.visitMethod(access, name, descriptor, signature, exceptions);

                    GenericMethodReference gmr = new GenericMethodReference("L" + currentClass + ";", name, descriptor);

                    Optional<ConvertedMethodReference> cmrOp = toPatch.stream()
                            .filter((cmr) -> cmr.getGenericMethodReference().equals(gmr))
                            .findFirst();

                    if (!cmrOp.isPresent())
                        return origMv;

                    final ConvertedMethodReference cmr = cmrOp.get();
                    final int methodId = cmr.getVmoDexMethodId();

                    return new MethodVisitor(ASM5, null) {
                        @Override
                        public void visitCode() {
                            if (!remaining.contains(cmr))
                                throw new IllegalStateException("Method " + gmr + " was already patched");
                            log.info("Patch method " + gmr);
                            origMv.visitCode();

                            String[] parameterTypes = gmr.getParameterTypes();

                            // add this as first parameter
                            if ((access & ACC_STATIC) == 0) {
                                String[] newParameterTypes = new String[parameterTypes.length + 1];
                                System.arraycopy(parameterTypes, 0, newParameterTypes, 1, parameterTypes.length);
                                newParameterTypes[0] = gmr.getDefiningClass();
                                parameterTypes = newParameterTypes;
                            }

                            final int arrayVar = parameterTypes.length +
                                    (int) Stream.of(parameterTypes).filter(Utils::isWideType).count();

                            // create object array for parameters
                            origMv.visitLdcInsn(parameterTypes.length);
                            origMv.visitTypeInsn(ANEWARRAY, "java/lang/Object");
                            origMv.visitVarInsn(ASTORE, arrayVar);

                            int curParReg = 0;
                            for (int i = 0; i < parameterTypes.length; i++) {
                                GenericMethodReference wrap = PrimitiveWrapperUtils.getWrapMethod(parameterTypes[i]);

                                // load array variable
                                origMv.visitVarInsn(ALOAD, arrayVar);
                                // load array index
                                origMv.visitLdcInsn(i);

                                if (wrap != null) {
                                    // load parameter
                                    switch (parameterTypes[i].charAt(0)) {
                                        case 'Z':
                                        case 'B':
                                        case 'S':
                                        case 'C':
                                        case 'I':
                                            origMv.visitVarInsn(ILOAD, curParReg);
                                            break;
                                        case 'J':
                                            origMv.visitVarInsn(LLOAD, curParReg);
                                            // longs use double register with
                                            curParReg++;
                                            break;
                                        case 'F':
                                            origMv.visitVarInsn(FLOAD, curParReg);
                                            break;
                                        case 'D':
                                            // doubles use double register with
                                            origMv.visitVarInsn(DLOAD, curParReg);
                                            curParReg++;
                                            break;
                                    }
                                    origMv.visitMethodInsn(INVOKESTATIC, convertDexTypeToJVMType(wrap.getDefiningClass()), wrap.getName(), wrap.getSignature(), false);
                                } else {
                                    origMv.visitVarInsn(ALOAD, curParReg);
                                }
                                // store argument to array
                                origMv.visitInsn(AASTORE);
                                curParReg++;
                            }

                            origMv.visitIntInsn(BIPUSH, methodId);
                            origMv.visitVarInsn(ALOAD, arrayVar);
                            final GenericMethodReference imr = InterpreterGenerator.INTERPRETER_METHOD;

                            origMv.visitMethodInsn(
                                    INVOKESTATIC,
                                    convertDexTypeToJVMType(imr.getDefiningClass()),
                                    imr.getName(),
                                    imr.getSignature(),
                                    false
                            );

                            String returnType = gmr.getReturnType();
                            GenericMethodReference unwrapResult = PrimitiveWrapperUtils.getUnwrapMethod(returnType);

                            if (returnType.charAt(0) == 'V') {
                                origMv.visitInsn(RETURN);
                            } else if (unwrapResult != null) {
                                origMv.visitTypeInsn(CHECKCAST, convertDexTypeToJVMType(unwrapResult.getDefiningClass()));
                                origMv.visitMethodInsn(INVOKEVIRTUAL, convertDexTypeToJVMType(unwrapResult.getDefiningClass()), unwrapResult.getName(), unwrapResult.getSignature(), false);
                                switch (returnType.charAt(0)) {
                                    case 'Z':
                                    case 'B':
                                    case 'S':
                                    case 'C':
                                    case 'I':
                                        origMv.visitInsn(IRETURN);
                                        break;
                                    case 'J':
                                        origMv.visitInsn(LRETURN);
                                        break;
                                    case 'F':
                                        origMv.visitInsn(FRETURN);
                                        break;
                                    case 'D':
                                        origMv.visitInsn(DRETURN);
                                        break;
                                }

                            } else {
                                origMv.visitTypeInsn(CHECKCAST, convertDexTypeToJVMType(returnType));
                                origMv.visitInsn(ARETURN);
                            }

                            origMv.visitMaxs(3, arrayVar + 1);
                            origMv.visitEnd();
                            remaining.remove(cmr);
                        }
                    };
                }
            }, 0);

            Files.write(classFilesToPatch[i], cw.toByteArray());
        }

        for (ConvertedMethodReference cmr : remaining) {
            log.warn("Method to be patched not found: " + cmr.getGenericMethodReference());
        }

        if (remaining.size() > 0) {
            throw new IllegalArgumentException("Some methods to be patched were not found");
        }
    }
}