package de.marstra.dvo.obfuscator;

import com.google.common.collect.Sets;
import de.marstra.dvo.obfuscator.analysis.DexMethod;
import de.marstra.dvo.obfuscator.analysis.MethodAnalysis;
import de.marstra.dvo.obfuscator.analysis.ParsedDex;
import de.marstra.dvo.obfuscator.insn.*;
import de.marstra.dvo.obfuscator.insn.Instruction;
import de.marstra.dvo.obfuscator.util.GenericMethodReference;
import interpreter.ExceptionHandler;
import interpreter.TryBlock;
import interpreter.VMODex;
import lombok.Data;
import lombok.Getter;
import lombok.extern.slf4j.Slf4j;
import org.jf.dexlib2.AccessFlags;
import org.jf.dexlib2.Opcode;
import org.jf.dexlib2.ReferenceType;
import org.jf.dexlib2.iface.Method;
import org.jf.dexlib2.iface.instruction.*;
import org.jf.dexlib2.iface.instruction.formats.*;

import java.io.ByteArrayOutputStream;
import java.io.IOException;
import java.util.*;
import java.util.function.Function;
import java.util.stream.Stream;

import static org.jf.dexlib2.Opcode.*;

@Slf4j
public class VMODexCompiler {
    @Data
    static class ConvertedMethodReference {
        private final int vmoDexMethodId;
        private final GenericMethodReference genericMethodReference;
    }

    @Data
    public static class CompilationResult {
        private final VMODex vmoDex;
        private final List<ConvertedMethodReference> convertedMethodReferences;
        private final Set<InstructionType> usedInstructionTypes;
    }

    public static class VMODexBuilder {
        private final static int FIELD = 0;
        private final static int METHOD = 1;
        private final static int STRING = 2;
        private final static int TYPE = 3;
        private final List[] references = new List[]{
                new LinkedList<interpreter.FieldReference>(),
                new LinkedList<interpreter.MethodReference>(),
                new LinkedList<String>(),
                new LinkedList<String>()
        };
        private final Map<Object, Integer>[] referenceToId = new Map[]{
                new HashMap<interpreter.FieldReference, Integer>(),
                new HashMap<interpreter.MethodReference, Integer>(),
                new HashMap<String, Integer>(),
                new HashMap<String, Integer>()
        };

        private final List<interpreter.Method> methods = new LinkedList<>();
        private final Set<InstructionType> usedInstructionTypes = new HashSet<>();
        private final List<ConvertedMethodReference> convertedMethodReferences = new LinkedList<>();

        @Getter
        private final ByteCodeMapping byteCodeMapping;

        public VMODexBuilder(ByteCodeMapping byteCodeMapping) {
            this.byteCodeMapping = byteCodeMapping;
        }

        private <T> int addReference(T value, int type) {
            Integer newRef = referenceToId[type].get(value);
            if (newRef == null) {
                references[type].add(value);
                newRef = references[type].size() - 1;
                referenceToId[type].put(value, newRef);
            }
            return newRef;
        }

        public int addMethodReference(interpreter.MethodReference value) {
            int id = addReference(value, METHOD);
            if (id >= (1 << 20))
                throw new IllegalStateException("to much methods");
            return id;
        }

        public int addFieldReference(interpreter.FieldReference value) {
            int id = addReference(value, FIELD);
            if (id >= (1 << 24))
                throw new IllegalStateException("to much fields");
            return id;
        }

        public int addTypeReference(String value) {
            int id = addReference(value, TYPE);
            if (id >= (1 << 16))
                throw new IllegalStateException("to much types");
            return id;
        }

        public int addStringReference(String value) {
            return addReference(value, STRING);
        }

        public void addMethod(Method method) {
            interpreter.Method Method = new MethodBuilder(this, method).build();
            methods.add(Method);
            int methodId = methods.size() - 1;
            convertedMethodReferences.add(new ConvertedMethodReference(methodId, new GenericMethodReference(method)));
        }

        public VMODex build() {
            return new VMODex(
                    (interpreter.FieldReference[]) references[FIELD].toArray(new interpreter.FieldReference[0]),
                    (interpreter.MethodReference[]) references[METHOD].toArray(new interpreter.MethodReference[0]),
                    methods.toArray(new interpreter.Method[0]),
                    (String[]) references[STRING].toArray(new String[0]),
                    (String[]) references[TYPE].toArray(new String[0])
            );
        }

        public void useInstructionType(InstructionType type) {
            usedInstructionTypes.add(type);
        }

        public List<ConvertedMethodReference> getConvertedMethodReferences() {
            return convertedMethodReferences;
        }

        public Set<InstructionType> getUsedInstructionTypes() {
            return usedInstructionTypes;
        }
    }

    public static class DexInstructionList implements Iterable<org.jf.dexlib2.iface.instruction.Instruction> {
        private final ArrayList<org.jf.dexlib2.iface.instruction.Instruction> instructions;
        private final ArrayList<Integer> instructionOffsets;

        public DexInstructionList(Iterable<? extends org.jf.dexlib2.iface.instruction.Instruction> instructions) {
            this.instructions = new ArrayList<>();
            instructions.forEach(this.instructions::add);
            this.instructionOffsets = new ArrayList<>();

            int offset = 0;
            for (org.jf.dexlib2.iface.instruction.Instruction ins : instructions) {
                instructionOffsets.add(offset);
                offset += ins.getCodeUnits();
            }
        }

        public int getInstructionIndexAtOffset(int offset) {
            int idx = Collections.binarySearch(instructionOffsets, offset);
            if (idx < 0)
                throw new NoSuchElementException("No element at offset " + offset);
            return idx;
        }

        public org.jf.dexlib2.iface.instruction.Instruction getInstructionAtOffset(int offset) {
            return getInstruction(getInstructionIndexAtOffset(offset));
        }

        private org.jf.dexlib2.iface.instruction.Instruction getInstruction(int idx) {
            return instructions.get(idx);
        }

        private int getOffset(int idx) {
            return instructionOffsets.get(idx);
        }

        @Override
        public Iterator<org.jf.dexlib2.iface.instruction.Instruction> iterator() {
            return instructions.iterator();
        }
    }

    public static class MethodBuilder {
        @Getter
        private final VMODexBuilder vmoDexBuilder;
        private final Method origMethod;
        private final LinkedList<Instruction> newInstructions = new LinkedList<>();
        private final DexInstructionList instructionList;
        private List<byte[]> payloads = new LinkedList<>();
        private int sourceOffset = 0, targetOffset = 0, payloadOffset = 0, idx = 0;
        Map<Integer, Integer> jumpTargetMapping = new HashMap<>();
        Map<Instruction, Integer> insToSourcePos = new HashMap<>();

        public MethodBuilder(VMODexBuilder vmoDexBuilder, Method method) {
            this.vmoDexBuilder = vmoDexBuilder;
            this.origMethod = method;
            this.instructionList = new DexInstructionList(method.getImplementation().getInstructions());
        }

        public int addPayload(byte[] payload) {
            int offset = payloadOffset;
            payloads.add(payload);
            payloadOffset += payload.length;
            return offset;
        }

        public int translateCodeOffset(Instruction instruction, int offset) {
            int sourcePos = insToSourcePos.get(instruction);
            return jumpTargetMapping.get(sourcePos + offset);
        }

        private void addInstruction(org.jf.dexlib2.iface.instruction.Instruction ins) {
            Instruction newInstruction = translateInstruction(ins);
            insToSourcePos.put(newInstruction, sourceOffset);

            if (newInstruction != null) {
                vmoDexBuilder.useInstructionType(newInstruction.getInstructionType());
                newInstructions.add(newInstruction);
                targetOffset += newInstruction.getInstructionType().getSize();
            }
            sourceOffset += ins.getCodeUnits();
            jumpTargetMapping.put(sourceOffset, targetOffset);
            idx++;
        }

        public interpreter.Method build() {
            jumpTargetMapping.put(sourceOffset, targetOffset);
            instructionList.forEach(this::addInstruction);
            payloadOffset = targetOffset;

            Stream<String> thisParamType = ((origMethod.getAccessFlags() & AccessFlags.STATIC.getValue()) == 0) ? Stream.of(origMethod.getDefiningClass()) : Stream.empty();

            String[] paramTypes = Stream.concat(
                    thisParamType,
                    origMethod.getParameterTypes().stream().map(Object::toString)
            ).toArray(String[]::new);

            TryBlock[] tryBlocks = origMethod.getImplementation().getTryBlocks().stream().map((tb) -> {
                int start = jumpTargetMapping.get(tb.getStartCodeAddress());
                int end = jumpTargetMapping.get(tb.getStartCodeAddress() + tb.getCodeUnitCount());
                ExceptionHandler[] handlers = tb.getExceptionHandlers().stream()
                        .map((eh) -> new ExceptionHandler(jumpTargetMapping.get(eh.getHandlerCodeAddress()), eh.getExceptionType()))
                        .toArray(ExceptionHandler[]::new);
                return new TryBlock(start, end, handlers);
            }).toArray(TryBlock[]::new);

            // serialize instructions
            byte[] rawInstructions;
            try {
                ByteArrayOutputStream baout = new ByteArrayOutputStream();

                for (Instruction newInstruction : newInstructions)
                    baout.write(newInstruction.serialize(this));

                for (byte[] payload : payloads)
                    baout.write(payload);

                rawInstructions = baout.toByteArray();
            } catch (IOException e) {
                throw new RuntimeException(e);
            }

            return new interpreter.Method(
                    origMethod.getName(),
                    origMethod.getDefiningClass(),
                    origMethod.getImplementation().getRegisterCount(),
                    rawInstructions,
                    origMethod.getReturnType(),
                    paramTypes,
                    tryBlocks
            );
        }

        // TODO collapse invoke and move-result
        // TODO collapse new_instance and invoke_direct <init>
        private Instruction translateInstruction(org.jf.dexlib2.iface.instruction.Instruction origInstruction) {
            Opcode opcode = origInstruction.getOpcode();

            Function<Integer, Integer> addRef = (a) -> a + a;

            switch (opcode) {
                // no nops in interpreter language
                case NOP:
                    return null;

                // move ops
                case MOVE:
                case MOVE_FROM16:
                case MOVE_16:
                case MOVE_WIDE:
                case MOVE_WIDE_FROM16:
                case MOVE_WIDE_16: {
                    TwoRegisterInstruction ins = (TwoRegisterInstruction) origInstruction;
                    return new Move(InstructionType.MOVE, ins.getRegisterB(), ins.getRegisterA());
                }
                case MOVE_OBJECT:
                case MOVE_OBJECT_FROM16:
                case MOVE_OBJECT_16: {
                    TwoRegisterInstruction ins = (TwoRegisterInstruction) origInstruction;
                    return new Move(InstructionType.MOVE_OBJECT, ins.getRegisterB(), ins.getRegisterA());
                }
                case MOVE_EXCEPTION: {
                    OneRegisterInstruction ins = (OneRegisterInstruction) origInstruction;
                    return new Move(InstructionType.MOVE_EXCEPTION, ins.getRegisterA());
                }
                case MOVE_RESULT:
                case MOVE_RESULT_OBJECT:
                case MOVE_RESULT_WIDE: {
                    OneRegisterInstruction ins = (OneRegisterInstruction) origInstruction;
                    return new Move(InstructionType.MOVE_RESULT, ins.getRegisterA());
                }

                // return ops
                case RETURN_VOID: {
                    return new Return(InstructionType.RETURN_VOID);
                }
                case RETURN:
                case RETURN_WIDE:
                case RETURN_OBJECT: {
                    OneRegisterInstruction ins = (OneRegisterInstruction) origInstruction;
                    return new Return(InstructionType.RETURN, ins.getRegisterA());
                }

                // const ops
                case CONST_4:
                case CONST_16:
                case CONST:
                case CONST_HIGH16: {
                    return new Const(
                            InstructionType.CONST,
                            ((OneRegisterInstruction) origInstruction).getRegisterA(),
                            ((NarrowLiteralInstruction) origInstruction).getNarrowLiteral()
                    );
                }
                case CONST_WIDE_16:
                case CONST_WIDE_32:
                case CONST_WIDE:
                case CONST_WIDE_HIGH16: {
                    return new Const(
                            InstructionType.CONST_WIDE,
                            ((OneRegisterInstruction) origInstruction).getRegisterA(),
                            ((WideLiteralInstruction) origInstruction).getWideLiteral()
                    );
                }
                case CONST_STRING:
                case CONST_STRING_JUMBO: {
                    return new Const(
                            InstructionType.CONST_STRING,
                            ((OneRegisterInstruction) origInstruction).getRegisterA(),
                            convertReference((ReferenceInstruction) origInstruction)
                    );
                }
                case CONST_CLASS: {
                    return new Const(
                            InstructionType.CONST_CLASS,
                            ((OneRegisterInstruction) origInstruction).getRegisterA(),
                            convertReference((ReferenceInstruction) origInstruction)
                    );
                }

                case CHECK_CAST: {
                    Instruction21c ins = (Instruction21c) origInstruction;
                    return new CheckCast(ins.getRegisterA(), (String) convertReference(ins));
                }
                case INSTANCE_OF: {
                    Instruction22c ins = (Instruction22c) origInstruction;
                    return new InstanceOf(ins.getRegisterA(), ins.getRegisterB(), (String) convertReference(ins));
                }
                case ARRAY_LENGTH: {
                    Instruction12x ins = (Instruction12x) origInstruction;
                    return new ArrayLength(ins.getRegisterA(), ins.getRegisterB());
                }
                case NEW_INSTANCE: {
                    Instruction21c ins = (Instruction21c) origInstruction;
                    return new NewInstance(ins.getRegisterA(), (String) convertReference(ins));
                }
                case NEW_ARRAY: {
                    Instruction22c ins = (Instruction22c) origInstruction;
                    return new NewArray(ins.getRegisterA(), ins.getRegisterB(), (String) convertReference(ins));
                }
                case FILLED_NEW_ARRAY: {
                    Instruction35c ins = (Instruction35c) origInstruction;
                    return new FilledNewArray((String) convertReference(ins), ins.getRegisterCount(), ins.getRegisterC(), ins.getRegisterD(), ins.getRegisterE(), ins.getRegisterF(), ins.getRegisterG());
                }
                case FILLED_NEW_ARRAY_RANGE: {
                    Instruction3rc ins = (Instruction3rc) origInstruction;
                    return new FilledNewArrayRange((String) convertReference(ins), ins.getRegisterCount(), ins.getStartRegister());
                }
                case FILL_ARRAY_DATA: {
                    Instruction31t ins = (Instruction31t) origInstruction;
                    ArrayPayload payload = (ArrayPayload) instructionList.getInstructionAtOffset(sourceOffset + ins.getCodeOffset());
                    return new FillArrayData(ins.getRegisterA(), payload.getElementWidth(), payload.getArrayElements());
                }
                case ARRAY_PAYLOAD: {
                    return null;
                }
                case THROW: {
                    return new Throw(((OneRegisterInstruction) origInstruction).getRegisterA());
                }

                case GOTO:
                case GOTO_16:
                case GOTO_32: {
                    return new Goto(((org.jf.dexlib2.iface.instruction.OffsetInstruction) origInstruction).getCodeOffset());
                }

                case SPARSE_SWITCH:
                case PACKED_SWITCH: {
                    Instruction31t ins = (Instruction31t) origInstruction;
                    SwitchPayload payload = (SwitchPayload) instructionList.getInstructionAtOffset(sourceOffset + ins.getCodeOffset());
                    return new Switch(ins.getRegisterA(), payload.getSwitchElements());
                }
                case SPARSE_SWITCH_PAYLOAD:
                case PACKED_SWITCH_PAYLOAD: {
                    return null;
                }

                // compare operations
                case CMPL_FLOAT: {
                    return binOp(InstructionType.CMPL_FLOAT, (ThreeRegisterInstruction) origInstruction);
                }
                case CMPG_FLOAT: {
                    return binOp(InstructionType.CMPG_FLOAT, (ThreeRegisterInstruction) origInstruction);
                }
                case CMPL_DOUBLE: {
                    return binOp(InstructionType.CMPL_DOUBLE, (ThreeRegisterInstruction) origInstruction);
                }
                case CMPG_DOUBLE: {
                    return binOp(InstructionType.CMPG_DOUBLE, (ThreeRegisterInstruction) origInstruction);
                }
                case CMP_LONG: {
                    return binOp(InstructionType.CMP_LONG, (ThreeRegisterInstruction) origInstruction);
                }

                // branch operations
                case IF_EQ: {
                    return branch(InstructionType.IF_EQ, (Instruction22t) origInstruction);
                }
                case IF_NE: {
                    return branch(InstructionType.IF_NE, (Instruction22t) origInstruction);
                }
                case IF_LT: {
                    return branch(InstructionType.IF_LT, (Instruction22t) origInstruction);
                }
                case IF_GE: {
                    return branch(InstructionType.IF_GE, (Instruction22t) origInstruction);
                }
                case IF_GT: {
                    return branch(InstructionType.IF_GT, (Instruction22t) origInstruction);
                }
                case IF_LE: {
                    return branch(InstructionType.IF_LE, (Instruction22t) origInstruction);
                }
                case IF_EQZ: {
                    return branchZero(InstructionType.IF_EQZ, (Instruction21t) origInstruction);
                }
                case IF_NEZ: {
                    return branchZero(InstructionType.IF_NEZ, (Instruction21t) origInstruction);
                }
                case IF_LTZ: {
                    return branchZero(InstructionType.IF_LTZ, (Instruction21t) origInstruction);
                }
                case IF_GEZ: {
                    return branchZero(InstructionType.IF_GEZ, (Instruction21t) origInstruction);
                }
                case IF_GTZ: {
                    return branchZero(InstructionType.IF_GTZ, (Instruction21t) origInstruction);
                }
                case IF_LEZ: {
                    return branchZero(InstructionType.IF_LEZ, (Instruction21t) origInstruction);
                }

                // load store operations
                case AGET:
                case AGET_WIDE:
                case AGET_OBJECT:
                case AGET_BOOLEAN:
                case AGET_BYTE:
                case AGET_CHAR:
                case AGET_SHORT: {
                    Instruction23x ins = (Instruction23x) origInstruction;
                    return new ArrayLoadStore(InstructionType.AGET, ins.getRegisterA(), ins.getRegisterB(), ins.getRegisterC());
                }
                case APUT:
                case APUT_WIDE:
                case APUT_OBJECT:
                case APUT_BOOLEAN:
                case APUT_BYTE:
                case APUT_CHAR:
                case APUT_SHORT: {
                    Instruction23x ins = (Instruction23x) origInstruction;
                    return new ArrayLoadStore(InstructionType.APUT, ins.getRegisterA(), ins.getRegisterB(), ins.getRegisterC());
                }
                case IGET:
                case IGET_WIDE:
                case IGET_OBJECT:
                case IGET_BOOLEAN:
                case IGET_BYTE:
                case IGET_CHAR:
                case IGET_SHORT: {
                    Instruction22c ins = (Instruction22c) origInstruction;
                    return new InstanceLoadStore(InstructionType.IGET, ins.getRegisterA(), ins.getRegisterB(), (interpreter.FieldReference) convertReference(ins));
                }
                case IPUT:
                case IPUT_WIDE:
                case IPUT_OBJECT:
                case IPUT_BOOLEAN:
                case IPUT_BYTE:
                case IPUT_CHAR:
                case IPUT_SHORT: {
                    Instruction22c ins = (Instruction22c) origInstruction;
                    return new InstanceLoadStore(InstructionType.IPUT, ins.getRegisterA(), ins.getRegisterB(), (interpreter.FieldReference) convertReference(ins));
                }
                case SGET:
                case SGET_WIDE:
                case SGET_OBJECT:
                case SGET_BOOLEAN:
                case SGET_BYTE:
                case SGET_CHAR:
                case SGET_SHORT: {
                    Instruction21c ins = (Instruction21c) origInstruction;
                    return new StaticLoadStore(InstructionType.SGET, ins.getRegisterA(), (interpreter.FieldReference) convertReference(ins));
                }
                case SPUT:
                case SPUT_WIDE:
                case SPUT_OBJECT:
                case SPUT_BOOLEAN:
                case SPUT_BYTE:
                case SPUT_CHAR:
                case SPUT_SHORT: {
                    Instruction21c ins = (Instruction21c) origInstruction;
                    return new StaticLoadStore(InstructionType.SPUT, ins.getRegisterA(), (interpreter.FieldReference) convertReference(ins));
                }

                // invoke operations
                case INVOKE_VIRTUAL: {
                    return invokeOp(InstructionType.INVOKE_VIRTUAL, (Instruction35c) origInstruction);
                }
                case INVOKE_DIRECT: {
                    return invokeOp(InstructionType.INVOKE_DIRECT, (Instruction35c) origInstruction);
                }
                case INVOKE_STATIC: {
                    return invokeOp(InstructionType.INVOKE_STATIC, (Instruction35c) origInstruction);
                }
                case INVOKE_INTERFACE: {
                    return invokeOp(InstructionType.INVOKE_INTERFACE, (Instruction35c) origInstruction);
                }
                case INVOKE_VIRTUAL_RANGE: {
                    return invokeRangeOp(InstructionType.INVOKE_VIRTUAL_RANGE, (Instruction3rc) origInstruction);
                }
                case INVOKE_DIRECT_RANGE: {
                    return invokeRangeOp(InstructionType.INVOKE_DIRECT_RANGE, (Instruction3rc) origInstruction);
                }
                case INVOKE_STATIC_RANGE: {
                    return invokeRangeOp(InstructionType.INVOKE_STATIC_RANGE, (Instruction3rc) origInstruction);
                }
                case INVOKE_INTERFACE_RANGE: {
                    return invokeRangeOp(InstructionType.INVOKE_INTERFACE_RANGE, (Instruction3rc) origInstruction);
                }

                // unary operations
                case NEG_INT: {
                    return unOp(InstructionType.NEG_INT, (TwoRegisterInstruction) origInstruction);
                }
                case NOT_INT: {
                    return unOp(InstructionType.NOT_INT, (TwoRegisterInstruction) origInstruction);
                }
                case NEG_LONG: {
                    return unOp(InstructionType.NEG_LONG, (TwoRegisterInstruction) origInstruction);
                }
                case NOT_LONG: {
                    return unOp(InstructionType.NOT_LONG, (TwoRegisterInstruction) origInstruction);
                }
                case NEG_FLOAT: {
                    return unOp(InstructionType.NEG_FLOAT, (TwoRegisterInstruction) origInstruction);
                }
                case NEG_DOUBLE: {
                    return unOp(InstructionType.NEG_DOUBLE, (TwoRegisterInstruction) origInstruction);
                }
                case INT_TO_LONG: {
                    return unOp(InstructionType.INT_TO_LONG, (TwoRegisterInstruction) origInstruction);
                }
                case INT_TO_FLOAT: {
                    return unOp(InstructionType.INT_TO_FLOAT, (TwoRegisterInstruction) origInstruction);
                }
                case INT_TO_DOUBLE: {
                    return unOp(InstructionType.INT_TO_DOUBLE, (TwoRegisterInstruction) origInstruction);
                }
                case LONG_TO_INT: {
                    return unOp(InstructionType.LONG_TO_INT, (TwoRegisterInstruction) origInstruction);
                }
                case LONG_TO_FLOAT: {
                    return unOp(InstructionType.LONG_TO_FLOAT, (TwoRegisterInstruction) origInstruction);
                }
                case LONG_TO_DOUBLE: {
                    return unOp(InstructionType.LONG_TO_DOUBLE, (TwoRegisterInstruction) origInstruction);
                }
                case FLOAT_TO_INT: {
                    return unOp(InstructionType.FLOAT_TO_INT, (TwoRegisterInstruction) origInstruction);
                }
                case FLOAT_TO_LONG: {
                    return unOp(InstructionType.FLOAT_TO_LONG, (TwoRegisterInstruction) origInstruction);
                }
                case FLOAT_TO_DOUBLE: {
                    return unOp(InstructionType.FLOAT_TO_DOUBLE, (TwoRegisterInstruction) origInstruction);
                }
                case DOUBLE_TO_INT: {
                    return unOp(InstructionType.DOUBLE_TO_INT, (TwoRegisterInstruction) origInstruction);
                }
                case DOUBLE_TO_LONG: {
                    return unOp(InstructionType.DOUBLE_TO_LONG, (TwoRegisterInstruction) origInstruction);
                }
                case DOUBLE_TO_FLOAT: {
                    return unOp(InstructionType.DOUBLE_TO_FLOAT, (TwoRegisterInstruction) origInstruction);
                }
                case INT_TO_BYTE: {
                    return unOp(InstructionType.INT_TO_BYTE, (TwoRegisterInstruction) origInstruction);
                }
                case INT_TO_CHAR: {
                    return unOp(InstructionType.INT_TO_CHAR, (TwoRegisterInstruction) origInstruction);
                }
                case INT_TO_SHORT: {
                    return unOp(InstructionType.INT_TO_SHORT, (TwoRegisterInstruction) origInstruction);
                }

                // binary operations
                case ADD_INT: {
                    return binOp(InstructionType.ADD_INT, (ThreeRegisterInstruction) origInstruction);
                }
                case SUB_INT: {
                    return binOp(InstructionType.SUB_INT, (ThreeRegisterInstruction) origInstruction);
                }
                case MUL_INT: {
                    return binOp(InstructionType.MUL_INT, (ThreeRegisterInstruction) origInstruction);
                }
                case DIV_INT: {
                    return binOp(InstructionType.DIV_INT, (ThreeRegisterInstruction) origInstruction);
                }
                case REM_INT: {
                    return binOp(InstructionType.REM_INT, (ThreeRegisterInstruction) origInstruction);
                }
                case ADD_FLOAT: {
                    return binOp(InstructionType.ADD_FLOAT, (ThreeRegisterInstruction) origInstruction);
                }
                case SUB_FLOAT: {
                    return binOp(InstructionType.SUB_FLOAT, (ThreeRegisterInstruction) origInstruction);
                }
                case MUL_FLOAT: {
                    return binOp(InstructionType.MUL_FLOAT, (ThreeRegisterInstruction) origInstruction);
                }
                case DIV_FLOAT: {
                    return binOp(InstructionType.DIV_FLOAT, (ThreeRegisterInstruction) origInstruction);
                }
                case REM_FLOAT: {
                    return binOp(InstructionType.REM_FLOAT, (ThreeRegisterInstruction) origInstruction);
                }
                case AND_INT: {
                    return binOp(InstructionType.AND_INT, (ThreeRegisterInstruction) origInstruction);
                }
                case OR_INT: {
                    return binOp(InstructionType.OR_INT, (ThreeRegisterInstruction) origInstruction);
                }
                case XOR_INT: {
                    return binOp(InstructionType.XOR_INT, (ThreeRegisterInstruction) origInstruction);
                }
                case SHL_INT: {
                    return binOp(InstructionType.SHL_INT, (ThreeRegisterInstruction) origInstruction);
                }
                case SHR_INT: {
                    return binOp(InstructionType.SHR_INT, (ThreeRegisterInstruction) origInstruction);
                }
                case USHR_INT: {
                    return binOp(InstructionType.USHR_INT, (ThreeRegisterInstruction) origInstruction);
                }
                case ADD_LONG: {
                    return binOp(InstructionType.ADD_LONG, (ThreeRegisterInstruction) origInstruction);
                }
                case SUB_LONG: {
                    return binOp(InstructionType.SUB_LONG, (ThreeRegisterInstruction) origInstruction);
                }
                case MUL_LONG: {
                    return binOp(InstructionType.MUL_LONG, (ThreeRegisterInstruction) origInstruction);
                }
                case DIV_LONG: {
                    return binOp(InstructionType.DIV_LONG, (ThreeRegisterInstruction) origInstruction);
                }
                case REM_LONG: {
                    return binOp(InstructionType.REM_LONG, (ThreeRegisterInstruction) origInstruction);
                }
                case ADD_DOUBLE: {
                    return binOp(InstructionType.ADD_DOUBLE, (ThreeRegisterInstruction) origInstruction);
                }
                case SUB_DOUBLE: {
                    return binOp(InstructionType.SUB_DOUBLE, (ThreeRegisterInstruction) origInstruction);
                }
                case MUL_DOUBLE: {
                    return binOp(InstructionType.MUL_DOUBLE, (ThreeRegisterInstruction) origInstruction);
                }
                case DIV_DOUBLE: {
                    return binOp(InstructionType.DIV_DOUBLE, (ThreeRegisterInstruction) origInstruction);
                }
                case REM_DOUBLE: {
                    return binOp(InstructionType.REM_DOUBLE, (ThreeRegisterInstruction) origInstruction);
                }
                case AND_LONG: {
                    return binOp(InstructionType.AND_LONG, (ThreeRegisterInstruction) origInstruction);
                }
                case OR_LONG: {
                    return binOp(InstructionType.OR_LONG, (ThreeRegisterInstruction) origInstruction);
                }
                case XOR_LONG: {
                    return binOp(InstructionType.XOR_LONG, (ThreeRegisterInstruction) origInstruction);
                }
                case SHL_LONG: {
                    return binOp(InstructionType.SHL_LONG, (ThreeRegisterInstruction) origInstruction);
                }
                case SHR_LONG: {
                    return binOp(InstructionType.SHR_LONG, (ThreeRegisterInstruction) origInstruction);
                }
                case USHR_LONG: {
                    return binOp(InstructionType.USHR_LONG, (ThreeRegisterInstruction) origInstruction);
                }

                // 2addr binary operations
                case ADD_INT_2ADDR: {
                    return binOp(InstructionType.ADD_INT, (TwoRegisterInstruction) origInstruction);
                }
                case SUB_INT_2ADDR: {
                    return binOp(InstructionType.SUB_INT, (TwoRegisterInstruction) origInstruction);
                }
                case MUL_INT_2ADDR: {
                    return binOp(InstructionType.MUL_INT, (TwoRegisterInstruction) origInstruction);
                }
                case DIV_INT_2ADDR: {
                    return binOp(InstructionType.DIV_INT, (TwoRegisterInstruction) origInstruction);
                }
                case REM_INT_2ADDR: {
                    return binOp(InstructionType.REM_INT, (TwoRegisterInstruction) origInstruction);
                }
                case ADD_FLOAT_2ADDR: {
                    return binOp(InstructionType.ADD_FLOAT, (TwoRegisterInstruction) origInstruction);
                }
                case SUB_FLOAT_2ADDR: {
                    return binOp(InstructionType.SUB_FLOAT, (TwoRegisterInstruction) origInstruction);
                }
                case MUL_FLOAT_2ADDR: {
                    return binOp(InstructionType.MUL_FLOAT, (TwoRegisterInstruction) origInstruction);
                }
                case DIV_FLOAT_2ADDR: {
                    return binOp(InstructionType.DIV_FLOAT, (TwoRegisterInstruction) origInstruction);
                }
                case REM_FLOAT_2ADDR: {
                    return binOp(InstructionType.REM_FLOAT, (TwoRegisterInstruction) origInstruction);
                }
                case AND_INT_2ADDR: {
                    return binOp(InstructionType.AND_INT, (TwoRegisterInstruction) origInstruction);
                }
                case OR_INT_2ADDR: {
                    return binOp(InstructionType.OR_INT, (TwoRegisterInstruction) origInstruction);
                }
                case XOR_INT_2ADDR: {
                    return binOp(InstructionType.XOR_INT, (TwoRegisterInstruction) origInstruction);
                }
                case SHL_INT_2ADDR: {
                    return binOp(InstructionType.SHL_INT, (TwoRegisterInstruction) origInstruction);
                }
                case SHR_INT_2ADDR: {
                    return binOp(InstructionType.SHR_INT, (TwoRegisterInstruction) origInstruction);
                }
                case USHR_INT_2ADDR: {
                    return binOp(InstructionType.USHR_INT, (TwoRegisterInstruction) origInstruction);
                }
                case ADD_LONG_2ADDR: {
                    return binOp(InstructionType.ADD_LONG, (TwoRegisterInstruction) origInstruction);
                }
                case SUB_LONG_2ADDR: {
                    return binOp(InstructionType.SUB_LONG, (TwoRegisterInstruction) origInstruction);
                }
                case MUL_LONG_2ADDR: {
                    return binOp(InstructionType.MUL_LONG, (TwoRegisterInstruction) origInstruction);
                }
                case DIV_LONG_2ADDR: {
                    return binOp(InstructionType.DIV_LONG, (TwoRegisterInstruction) origInstruction);
                }
                case REM_LONG_2ADDR: {
                    return binOp(InstructionType.REM_LONG, (TwoRegisterInstruction) origInstruction);
                }
                case ADD_DOUBLE_2ADDR: {
                    return binOp(InstructionType.ADD_DOUBLE, (TwoRegisterInstruction) origInstruction);
                }
                case SUB_DOUBLE_2ADDR: {
                    return binOp(InstructionType.SUB_DOUBLE, (TwoRegisterInstruction) origInstruction);
                }
                case MUL_DOUBLE_2ADDR: {
                    return binOp(InstructionType.MUL_DOUBLE, (TwoRegisterInstruction) origInstruction);
                }
                case DIV_DOUBLE_2ADDR: {
                    return binOp(InstructionType.DIV_DOUBLE, (TwoRegisterInstruction) origInstruction);
                }
                case REM_DOUBLE_2ADDR: {
                    return binOp(InstructionType.REM_DOUBLE, (TwoRegisterInstruction) origInstruction);
                }
                case AND_LONG_2ADDR: {
                    return binOp(InstructionType.AND_LONG, (TwoRegisterInstruction) origInstruction);
                }
                case OR_LONG_2ADDR: {
                    return binOp(InstructionType.OR_LONG, (TwoRegisterInstruction) origInstruction);
                }
                case XOR_LONG_2ADDR: {
                    return binOp(InstructionType.XOR_LONG, (TwoRegisterInstruction) origInstruction);
                }
                case SHL_LONG_2ADDR: {
                    return binOp(InstructionType.SHL_LONG, (TwoRegisterInstruction) origInstruction);
                }
                case SHR_LONG_2ADDR: {
                    return binOp(InstructionType.SHR_LONG, (TwoRegisterInstruction) origInstruction);
                }
                case USHR_LONG_2ADDR: {
                    return binOp(InstructionType.USHR_LONG, (TwoRegisterInstruction) origInstruction);
                }

                // literal binary operations
                case ADD_INT_LIT8: {
                    return binOpLit(InstructionType.ADD_INT_LIT, (Instruction22b) origInstruction);
                }
                case RSUB_INT_LIT8: {
                    return binOpLit(InstructionType.RSUB_INT_LIT, (Instruction22b) origInstruction);
                }
                case MUL_INT_LIT8: {
                    return binOpLit(InstructionType.MUL_INT_LIT, (Instruction22b) origInstruction);
                }
                case DIV_INT_LIT8: {
                    return binOpLit(InstructionType.DIV_INT_LIT, (Instruction22b) origInstruction);
                }
                case REM_INT_LIT8: {
                    return binOpLit(InstructionType.REM_INT_LIT, (Instruction22b) origInstruction);
                }
                case AND_INT_LIT8: {
                    return binOpLit(InstructionType.AND_INT_LIT, (Instruction22b) origInstruction);
                }
                case OR_INT_LIT8: {
                    return binOpLit(InstructionType.OR_INT_LIT, (Instruction22b) origInstruction);
                }
                case XOR_INT_LIT8: {
                    return binOpLit(InstructionType.XOR_INT_LIT, (Instruction22b) origInstruction);
                }
                case SHL_INT_LIT8: {
                    return binOpLit(InstructionType.SHL_INT_LIT, (Instruction22b) origInstruction);
                }
                case SHR_INT_LIT8: {
                    return binOpLit(InstructionType.SHR_INT_LIT, (Instruction22b) origInstruction);
                }
                case USHR_INT_LIT8: {
                    return binOpLit(InstructionType.USHR_INT_LIT, (Instruction22b) origInstruction);
                }
                case ADD_INT_LIT16: {
                    return binOpLit(InstructionType.ADD_INT_LIT, (Instruction22s) origInstruction);
                }
                case RSUB_INT: {
                    return binOpLit(InstructionType.RSUB_INT_LIT, (Instruction22s) origInstruction);
                }
                case MUL_INT_LIT16: {
                    return binOpLit(InstructionType.MUL_INT_LIT, (Instruction22s) origInstruction);
                }
                case DIV_INT_LIT16: {
                    return binOpLit(InstructionType.DIV_INT_LIT, (Instruction22s) origInstruction);
                }
                case REM_INT_LIT16: {
                    return binOpLit(InstructionType.REM_INT_LIT, (Instruction22s) origInstruction);
                }
                case AND_INT_LIT16: {
                    return binOpLit(InstructionType.AND_INT_LIT, (Instruction22s) origInstruction);
                }
                case OR_INT_LIT16: {
                    return binOpLit(InstructionType.OR_INT_LIT, (Instruction22s) origInstruction);
                }
                case XOR_INT_LIT16: {
                    return binOpLit(InstructionType.XOR_INT_LIT, (Instruction22s) origInstruction);
                }
                default:
                    throw new IllegalStateException("Trying to compile unsupported operation: " + origInstruction.getOpcode());
            }
        }
    }

    public static CompilationResult compile(ParsedDex parsedDex, Set<GenericMethodReference> toConvert, ByteCodeMapping byteCodeMapping) {
        VMODexBuilder vmoDexBuilder = new VMODexBuilder(byteCodeMapping);

        for (GenericMethodReference gmr : toConvert) {
            DexMethod dexMethod = parsedDex.findMethod(gmr);
            if (dexMethod == null)
                throw new IllegalArgumentException("Method to be converted not found: " + gmr);

            Method m = dexMethod.getMethod();
            MethodAnalysis.MethodAnalysisResult analysisResult = MethodAnalysis.analyseMethod(m, parsedDex);
            if (analysisResult.isSupported()) {
                log.debug("Compile method '" + gmr + "' to vmodex");
                vmoDexBuilder.addMethod(m);
            } else {
                log.warn("Skipping unsupported method '" + gmr + "', reasons: [" +
                        String.join("; ", analysisResult.getUnsupportedReasons()) + "]");
            }
        }

        return new VMODexCompiler.CompilationResult(
                vmoDexBuilder.build(),
                vmoDexBuilder.getConvertedMethodReferences(),
                vmoDexBuilder.getUsedInstructionTypes()
        );
    }

    private static UnOp unOp(InstructionType type, TwoRegisterInstruction ins) {
        return new UnOp(type, ins.getRegisterA(), ins.getRegisterB());
    }

    private static BinOp binOp(InstructionType type, TwoRegisterInstruction ins) {
        return new BinOp(type, ins.getRegisterA(), ins.getRegisterA(), ins.getRegisterB());
    }

    private static BinOp binOp(InstructionType type, ThreeRegisterInstruction ins) {
        return new BinOp(type, ins.getRegisterA(), ins.getRegisterB(), ins.getRegisterC());
    }

    private static <I extends TwoRegisterInstruction & NarrowLiteralInstruction> BinOpLit binOpLit(InstructionType type, I ins) {
        return new BinOpLit(type, ins.getRegisterA(), ins.getRegisterB(), ins.getNarrowLiteral());
    }

    private static Branch branch(InstructionType type, Instruction22t ins) {
        return new Branch(type, ins.getRegisterA(), ins.getRegisterB(), ins.getCodeOffset());
    }

    private static BranchZero branchZero(InstructionType type, Instruction21t ins) {
        return new BranchZero(type, ins.getRegisterA(), ins.getCodeOffset());
    }

    private static Invoke invokeOp(InstructionType type, Instruction35c ins) {
        return new Invoke(type, (interpreter.MethodReference) convertReference(ins), ins.getRegisterC(), ins.getRegisterD(), ins.getRegisterE(), ins.getRegisterF(), ins.getRegisterG());
    }

    private static InvokeRange invokeRangeOp(InstructionType type, Instruction3rc ins) {
        return new InvokeRange(type, (interpreter.MethodReference) convertReference(ins), ins.getStartRegister());
    }

    private static Object convertReference(ReferenceInstruction ins) {
        switch (ins.getReferenceType()) {
            case ReferenceType.FIELD:
                org.jf.dexlib2.iface.reference.FieldReference fieldReference = (org.jf.dexlib2.iface.reference.FieldReference) ins.getReference();
                return new interpreter.FieldReference(
                        fieldReference.getDefiningClass(),
                        fieldReference.getName(),
                        fieldReference.getType()
                );
            case ReferenceType.METHOD:
                org.jf.dexlib2.iface.reference.MethodReference methodReference = (org.jf.dexlib2.iface.reference.MethodReference) ins.getReference();
                return new interpreter.MethodReference(
                        methodReference.getDefiningClass(),
                        methodReference.getName(),
                        methodReference.getParameterTypes().stream()
                                .map(CharSequence::toString)
                                .toArray(String[]::new),
                        methodReference.getReturnType()
                );
            case ReferenceType.STRING:
                return ((CharSequence) ins.getReference()).toString();
            case ReferenceType.TYPE:
                return ((CharSequence) ins.getReference()).toString();
            default:
                throw new IllegalArgumentException("Unsupported ReferenceType: " + ins.getReferenceType());
        }
    }

    public static Set<Opcode> supported = Sets.newHashSet(NOP, MOVE, MOVE_FROM16, MOVE_16, MOVE_WIDE, MOVE_WIDE_FROM16, MOVE_WIDE_16, MOVE_OBJECT, MOVE_OBJECT_FROM16, MOVE_OBJECT_16, MOVE_EXCEPTION, MOVE_RESULT, MOVE_RESULT_OBJECT, MOVE_RESULT_WIDE, RETURN_VOID, RETURN, RETURN_WIDE, RETURN_OBJECT, CONST_4, CONST_16, CONST, CONST_HIGH16, CONST_WIDE_16, CONST_WIDE_32, CONST_WIDE, CONST_WIDE_HIGH16, CONST_STRING, CONST_STRING_JUMBO, CONST_CLASS, CHECK_CAST, INSTANCE_OF, ARRAY_LENGTH, NEW_INSTANCE, NEW_ARRAY, FILLED_NEW_ARRAY, FILLED_NEW_ARRAY_RANGE, FILL_ARRAY_DATA, ARRAY_PAYLOAD, THROW, GOTO, GOTO_16, GOTO_32, SPARSE_SWITCH, PACKED_SWITCH, SPARSE_SWITCH_PAYLOAD, PACKED_SWITCH_PAYLOAD, CMPL_FLOAT, CMPG_FLOAT, CMPL_DOUBLE, CMPG_DOUBLE, CMP_LONG, IF_EQ, IF_NE, IF_LT, IF_GE, IF_GT, IF_LE, IF_EQZ, IF_NEZ, IF_LTZ, IF_GEZ, IF_GTZ, IF_LEZ, AGET, AGET_WIDE, AGET_OBJECT, AGET_BOOLEAN, AGET_BYTE, AGET_CHAR, AGET_SHORT, APUT, APUT_WIDE, APUT_OBJECT, APUT_BOOLEAN, APUT_BYTE, APUT_CHAR, APUT_SHORT, IGET, IGET_WIDE, IGET_OBJECT, IGET_BOOLEAN, IGET_BYTE, IGET_CHAR, IGET_SHORT, IPUT, IPUT_WIDE, IPUT_OBJECT, IPUT_BOOLEAN, IPUT_BYTE, IPUT_CHAR, IPUT_SHORT, SGET, SGET_WIDE, SGET_OBJECT, SGET_BOOLEAN, SGET_BYTE, SGET_CHAR, SGET_SHORT, SPUT, SPUT_WIDE, SPUT_OBJECT, SPUT_BOOLEAN, SPUT_BYTE, SPUT_CHAR, SPUT_SHORT, INVOKE_VIRTUAL, INVOKE_DIRECT, INVOKE_STATIC, INVOKE_INTERFACE, INVOKE_VIRTUAL_RANGE, INVOKE_DIRECT_RANGE, INVOKE_STATIC_RANGE, INVOKE_INTERFACE_RANGE, NEG_INT, NOT_INT, NEG_LONG, NOT_LONG, NEG_FLOAT, NEG_DOUBLE, INT_TO_LONG, INT_TO_FLOAT, INT_TO_DOUBLE, LONG_TO_INT, LONG_TO_FLOAT, LONG_TO_DOUBLE, FLOAT_TO_INT, FLOAT_TO_LONG, FLOAT_TO_DOUBLE, DOUBLE_TO_INT, DOUBLE_TO_LONG, DOUBLE_TO_FLOAT, INT_TO_BYTE, INT_TO_CHAR, INT_TO_SHORT, ADD_INT, SUB_INT, MUL_INT, DIV_INT, REM_INT, ADD_FLOAT, SUB_FLOAT, MUL_FLOAT, DIV_FLOAT, REM_FLOAT, AND_INT, OR_INT, XOR_INT, SHL_INT, SHR_INT, USHR_INT, ADD_LONG, SUB_LONG, MUL_LONG, DIV_LONG, REM_LONG, ADD_DOUBLE, SUB_DOUBLE, MUL_DOUBLE, DIV_DOUBLE, REM_DOUBLE, AND_LONG, OR_LONG, XOR_LONG, SHL_LONG, SHR_LONG, USHR_LONG, ADD_INT_2ADDR, SUB_INT_2ADDR, MUL_INT_2ADDR, DIV_INT_2ADDR, REM_INT_2ADDR, ADD_FLOAT_2ADDR, SUB_FLOAT_2ADDR, MUL_FLOAT_2ADDR, DIV_FLOAT_2ADDR, REM_FLOAT_2ADDR, AND_INT_2ADDR, OR_INT_2ADDR, XOR_INT_2ADDR, SHL_INT_2ADDR, SHR_INT_2ADDR, USHR_INT_2ADDR, ADD_LONG_2ADDR, SUB_LONG_2ADDR, MUL_LONG_2ADDR, DIV_LONG_2ADDR, REM_LONG_2ADDR, ADD_DOUBLE_2ADDR, SUB_DOUBLE_2ADDR, MUL_DOUBLE_2ADDR, DIV_DOUBLE_2ADDR, REM_DOUBLE_2ADDR, AND_LONG_2ADDR, OR_LONG_2ADDR, XOR_LONG_2ADDR, SHL_LONG_2ADDR, SHR_LONG_2ADDR, USHR_LONG_2ADDR, ADD_INT_LIT8, RSUB_INT_LIT8, MUL_INT_LIT8, DIV_INT_LIT8, REM_INT_LIT8, AND_INT_LIT8, OR_INT_LIT8, XOR_INT_LIT8, SHL_INT_LIT8, SHR_INT_LIT8, USHR_INT_LIT8, ADD_INT_LIT16, RSUB_INT, MUL_INT_LIT16, DIV_INT_LIT16, REM_INT_LIT16, AND_INT_LIT16, OR_INT_LIT16, XOR_INT_LIT16);

    public static boolean isInstructionSupported(org.jf.dexlib2.iface.instruction.Instruction ins) {
        // TODO find non redundant way
        return supported.contains(ins.getOpcode());
    }
}
