package de.marstra.dvo.obfuscator.analysis;

import com.google.common.collect.Streams;
import de.marstra.dvo.obfuscator.util.CommandExecutor;
import de.marstra.dvo.obfuscator.util.GenericMethodReference;
import fr.xgouchet.axml.CompressedXmlParser;
import lombok.Data;
import lombok.Getter;
import lombok.extern.slf4j.Slf4j;
import org.apache.commons.io.IOUtils;
import org.jf.dexlib2.dexbacked.DexBackedDexFile;
import org.jf.dexlib2.iface.DexFile;
import org.jf.dexlib2.writer.io.MemoryDataStore;
import org.jf.dexlib2.writer.pool.DexPool;
import org.w3c.dom.Document;
import org.w3c.dom.Element;
import org.w3c.dom.Node;
import org.w3c.dom.NodeList;

import javax.xml.parsers.ParserConfigurationException;
import javax.xml.transform.*;
import javax.xml.transform.dom.DOMSource;
import javax.xml.transform.stream.StreamResult;
import javax.xml.xpath.XPath;
import javax.xml.xpath.XPathConstants;
import javax.xml.xpath.XPathExpressionException;
import javax.xml.xpath.XPathFactory;
import java.io.*;
import java.nio.charset.StandardCharsets;
import java.nio.file.Files;
import java.nio.file.Path;
import java.util.HashMap;
import java.util.Iterator;
import java.util.LinkedList;
import java.util.List;
import java.util.regex.Matcher;
import java.util.regex.Pattern;
import java.util.stream.Collectors;
import java.util.zip.*;

import static de.marstra.dvo.obfuscator.util.Utils.doWithTmpFile;

@Slf4j
public class ApkFile {
    @Getter
    private final List<ZippedFile> zipEntries;
    private List<DexFile> dexFiles;
    private final HashMap<String, ZippedFile> otherEntries;
    private Document manifest;
    private ParsedDex parsedDex;

    @Data
    public static class ZippedFile {
        private final String name;
        private final byte[] content;
        private final boolean compressed;
    }

    public ApkFile(List<ZippedFile> zipEntries) {
        this.zipEntries = zipEntries;
        this.dexFiles = new LinkedList<>();
        this.otherEntries = new HashMap<>();

        final List<ZippedFile> zippedDexFiles = new LinkedList<>();

        zipEntries.forEach((zf) -> {
            if (zf.getName().matches("classes\\d*\\.dex")) {
                zippedDexFiles.add(zf);
            } else {
                otherEntries.put(zf.getName(), zf);
            }
        });

        zippedDexFiles.sort((a, b) -> {
            Pattern pattern = Pattern.compile("^classes(\\d*)\\.dex$");
            Matcher ma = pattern.matcher(a.getName());
            Matcher mb = pattern.matcher(b.getName());
            if (!ma.find() || !mb.find())
                throw new IllegalStateException();
            return Integer.compare(
                    ma.group(1).isEmpty() ? -1 : Integer.parseInt(ma.group(1)),
                    mb.group(1).isEmpty() ? -1 : Integer.parseInt(mb.group(1))
            );
        });

        for (int i = 0; i < zippedDexFiles.size(); i++) {
            ZippedFile zippedFile = zippedDexFiles.get(i);

            if (!zippedFile.getName().equals(getDexName(i)))
                throw new IllegalArgumentException("Found invalid named dex file in apk:" + zippedFile.getName());

            try {
                DexFile dexFile = DexBackedDexFile.fromInputStream(null, new ByteArrayInputStream(zippedFile.getContent()));
                dexFiles.add(dexFile);
            } catch (IOException e) {
                throw new RuntimeException(e);
            }
        }

        if (dexFiles.isEmpty())
            throw new IllegalArgumentException("ApkFile without any dex files");
    }

    public ApkFile(File file) throws IOException {
        this(zipFileToZippedFileList(new ZipFile(file)));
    }

    public ApkFile(byte[] data) {
        this(zipInputStreamToZippedFileList(new ZipInputStream(new ByteArrayInputStream(data))));
    }

    public ParsedDex getParsedDex() {
        if (parsedDex == null) {
            parsedDex = new ParsedDex(dexFiles);
        }
        return parsedDex;
    }

    private static List<ZippedFile> zipFileToZippedFileList(ZipFile zipFile) {
        return zipFile.stream().map((entry) -> {
            try {
                byte[] content = IOUtils.toByteArray(zipFile.getInputStream(entry));
                return new ZippedFile(entry.getName(), content, entry.getMethod() != ZipEntry.STORED);
            } catch (IOException e) {
                throw new RuntimeException(e);
            }
        }).collect(Collectors.toList());
    }

    private static List<ZippedFile> zipInputStreamToZippedFileList(ZipInputStream zis) {
        return Streams.stream(new Iterator<ZippedFile>() {
            private ZipEntry next;

            @Override
            public boolean hasNext() {
                try {
                    if (next != null)
                        return true;
                    next = zis.getNextEntry();
                    return next != null;
                } catch (IOException e) {
                    throw new RuntimeException(e);
                }
            }

            @Override
            public ZippedFile next() {
                try {
                    if (next == null)
                        next = zis.getNextEntry();

                    ByteArrayOutputStream baout = new ByteArrayOutputStream();
                    byte[] buf = new byte[1024];
                    int read;

                    while ((read = zis.read(buf, 0, buf.length)) != -1) {
                        baout.write(buf, 0, read);
                    }

                    ZippedFile result = new ZippedFile(next.getName(), baout.toByteArray(), next.getMethod() != ZipEntry.STORED);
                    next = null;
                    return result;
                } catch (IOException e) {
                    throw new RuntimeException(e);
                }
            }
        }).collect(Collectors.toList());
    }

    private static String getDexName(int i) {
        return "classes" + (i == 0 ? "" : (i + 1)) + ".dex";
    }

    public void setDexFiles(List<DexFile> dexFiles) {
        if (this.dexFiles.size() != dexFiles.size())
            log.warn("Number of dex files in apk changed from " + this.dexFiles.size() + " to " + dexFiles.size());
        this.dexFiles = dexFiles;
    }

    private static byte[] dexFileToByteArray(DexFile dexFile) {
        // TODO findout apropriate initial capacity for buffer to avoid lots of memcopies
        MemoryDataStore memoryDataStore = new MemoryDataStore();
        try {
            DexPool.writeTo(memoryDataStore, dexFile);
        } catch (IOException e) {
            throw new RuntimeException(e);
        }
        return memoryDataStore.getData();
    }

    private void writeTo(ZipOutputStream zos) {
        try {
            // write dex files
            for (int i = 0; i < dexFiles.size(); i++) {
                zos.putNextEntry(new ZipEntry(getDexName(i)));
                DexFile dexFile = dexFiles.get(i);
                zos.write(dexFileToByteArray(dexFile));
            }
        } catch (IOException e) {
            throw new RuntimeException(e);
        }

        // write all other files
        otherEntries.forEach((name, zippedFile) -> {
            ZipEntry entry = new ZipEntry(name);
            byte[] content = zippedFile.getContent();

            if (!zippedFile.isCompressed()) {
                // some apks won't install if some packages
                entry.setMethod(ZipEntry.STORED);
                entry.setSize(content.length);
                entry.setCompressedSize(content.length);
                CRC32 crc32 = new CRC32();
                crc32.update(content);
                entry.setCrc(crc32.getValue());
            }
            try {
                zos.putNextEntry(entry);
                zos.write(content);
            } catch (IOException e) {
                throw new RuntimeException(e);
            }
        });
    }

    public void setEntry(String path, byte[] content, boolean compressed) {
        otherEntries.put(path, new ZippedFile(path, content, compressed));
    }

    public byte[] getEntry(String path) {
        return otherEntries.get(path).getContent();
    }

    public boolean exits(String path) {
        return otherEntries.get(path) != null;
    }

    public String getPackage() {
        try {
            XPath xPath = XPathFactory.newInstance().newXPath();
            return (String) xPath.evaluate("/manifest/@package", getManifest(), XPathConstants.STRING);
        } catch (XPathExpressionException e) {
            throw new RuntimeException(e);
        }
    }

    public List<GenericMethodReference> getEntryPoints() {
        List<GenericMethodReference> entryPoints = new LinkedList<>();

        try {
            XPath xPath = XPathFactory.newInstance().newXPath();
            NodeList mainActivities = (NodeList) xPath.evaluate(
                    "/manifest/application/activity" +
                            "[" +
                            // workaround because "intent-filter/action[@android:name='android.intent.action.MAIN']"
                            // is not working (maybe due to : inside attribute name)
                            "(intent-filter/action[@*[name()='android:name' and .='android.intent.action.MAIN']])" +
                            " and " +
                            "(intent-filter/category[@*[name()='android:name' and .='android.intent.category.LAUNCHER']])" +
                            "]"
                    , getManifest(), XPathConstants.NODESET);

            for (int i = 0; i < mainActivities.getLength(); i++) {
                Element activity = (Element) mainActivities.item(i);
                String clazz = "L" + activity.getAttribute("android:name").replaceAll("\\.", "/") + ";";
                GenericMethodReference entry = new GenericMethodReference(
                        clazz,
                        "onCreate",
                        "V",
                        "Landroid/os/Bundle;"
                );
                entryPoints.add(entry);
            }
        } catch (XPathExpressionException e) {
            throw new RuntimeException(e);
        }
        return entryPoints;
    }

    public String manifestAsString() {
        try {
            Transformer transformer = TransformerFactory.newInstance().newTransformer();
            transformer.setOutputProperty(OutputKeys.OMIT_XML_DECLARATION, "no");
            transformer.setOutputProperty(OutputKeys.METHOD, "xml");
            transformer.setOutputProperty(OutputKeys.INDENT, "yes");
            transformer.setOutputProperty(OutputKeys.ENCODING, StandardCharsets.UTF_8.name());
            Source source = new DOMSource(getManifest());
            ByteArrayOutputStream baout = new ByteArrayOutputStream();
            Result output = new StreamResult(baout);
            transformer.transform(source, output);
            return baout.toString(StandardCharsets.UTF_8.name());
        } catch (TransformerException | UnsupportedEncodingException e) {
            throw new RuntimeException(e);
        }
    }

    private Node getManifest() {
        if (manifest == null) {
            ZippedFile zippedFile = otherEntries.get("AndroidManifest.xml");
            try {
                manifest = new CompressedXmlParser().parseDOM(new ByteArrayInputStream(zippedFile.getContent()));
            } catch (IOException | ParserConfigurationException e) {
                throw new RuntimeException(e);
            }
        }
        return manifest;
    }

    public void write(Path outPath) throws IOException {
        ZipOutputStream zos = new ZipOutputStream(Files.newOutputStream(outPath));
        writeTo(zos);
        zos.close();
    }

    public byte[] sign() {
        return doWithTmpFile("sigendApk", tempFile -> {
            this.write(tempFile.toPath());
            // sign apk
            signApk(tempFile);
            return Files.readAllBytes(tempFile.toPath());
        });
    }

    public static void signApk(File apkFile) {
        // TODO find a way to use apksig https://mvnrepository.com/artifact/com.android.tools.build/apksig?repo=google
        doWithTmpFile("keystore", keyFile -> {
            IOUtils.copy(ApkFile.class.getResourceAsStream("/debug.keystore"), new FileOutputStream(keyFile));
            CommandExecutor.executeCommand("apksigner sign --ks " + keyFile + " --ks-pass pass:password " + apkFile);
            return null;
        });
    }

    public byte[] serialize() {
        ByteArrayOutputStream baout = new ByteArrayOutputStream();
        ZipOutputStream zos = new ZipOutputStream(baout);
        writeTo(zos);
        try {
            zos.close();
            return baout.toByteArray();
        } catch (IOException e) {
            throw new RuntimeException(e);
        }
    }
}
