package de.marstra.dvo.obfuscator.analysis;

import com.google.common.collect.Streams;
import de.marstra.dvo.obfuscator.util.GenericMethodReference;
import lombok.EqualsAndHashCode;
import lombok.Getter;
import org.jf.dexlib2.iface.Annotation;
import org.jf.dexlib2.iface.ClassDef;
import org.jf.dexlib2.iface.Field;
import org.jf.dexlib2.immutable.ImmutableClassDef;

import java.util.Collection;
import java.util.HashMap;
import java.util.List;
import java.util.Set;
import java.util.stream.Collectors;

import static de.marstra.dvo.obfuscator.util.Utils.getClassName;
import static de.marstra.dvo.obfuscator.util.Utils.iteratorToList;

@EqualsAndHashCode
public class DexClass extends DexPackageItem {
    @Getter
    private final DexPackage dexPackage;
    @Getter
    private final String name;
    @Getter
    private final String superclass;
    @Getter
    private final List<String> interfaces;
    @Getter
    private final String type;
    @Getter
    private final int accessFlags;
    @Getter
    private final List<? extends Field> instanceFields;
    @Getter
    private final List<? extends Field> staticFields;
    @Getter
    private final Set<? extends Annotation> annotations;

    private final HashMap<GenericMethodReference, DexMethod> methods;

    public DexClass(DexPackage dexPackage, ClassDef classDef) {
        if (!classDef.getType().startsWith("L" + dexPackage.getFullName()))
            throw new IllegalArgumentException();

        this.dexPackage = dexPackage;
        this.instanceFields = iteratorToList(classDef.getInstanceFields());
        this.staticFields = iteratorToList(classDef.getStaticFields());
        this.type = classDef.getType();
        this.name = getClassName(type);
        this.accessFlags = classDef.getAccessFlags();
        this.superclass = classDef.getSuperclass();
        this.interfaces = classDef.getInterfaces();
        this.annotations = classDef.getAnnotations();
        this.methods = new HashMap<>();
        classDef.getMethods().forEach(m -> methods.put(
                new GenericMethodReference(m),
                new DexMethod(m, this)
        ));
    }

    public ClassDef toClassDef() {
        return new ImmutableClassDef(
                type,
                accessFlags,
                superclass,
                interfaces,
                null,
                annotations,
                Streams.concat(instanceFields.stream(), staticFields.stream()).collect(Collectors.toList()),
                methods.values().stream().map(DexMethod::getMethod).collect(Collectors.toList())
        );
    }

    @Override
    public final <RetType, Param> RetType accept(final AnalysisVisitor<RetType, Param> av, Param p) {
        return av.visit(this, p);
    }

    public DexMethod getMethod(GenericMethodReference gmr) {
        return methods.get(gmr);
    }

    public Collection<DexMethod> getMethods() {
        return methods.values();
    }
}
