package de.marstra.dvo.obfuscator.analysis;

import de.marstra.dvo.obfuscator.util.GenericMethodReference;
import de.marstra.dvo.obfuscator.util.Utils;
import lombok.EqualsAndHashCode;
import lombok.Getter;
import lombok.RequiredArgsConstructor;
import lombok.extern.slf4j.Slf4j;
import org.jf.dexlib2.iface.ClassDef;

import java.util.*;
import java.util.stream.Collectors;

@Slf4j
@RequiredArgsConstructor
@EqualsAndHashCode
public class DexPackage extends DexPackageItem {
    @Getter
    private final DexPackage parent;
    @Getter
    private final String name;

    private String fullName;

    private HashMap<String, DexPackageItem> children = new HashMap<>();

    public String getFullName() {
        if (fullName == null) {
            List<String> components = new LinkedList<>();
            DexPackage cur = this;
            while (cur != null && cur.getName() != null) {
                components.add(cur.getName());
                cur = cur.parent;
            }
            Collections.reverse(components);
            fullName = String.join("/", components);
        }
        return fullName;
    }

    public DexMethod findMethod(GenericMethodReference gmr) {
        DexClass dexClass = findClass(gmr.getDefiningClass());
        if (dexClass == null) return null;
        return dexClass.getMethod(gmr);
    }

    public DexClass findClass(String type) {
        String clazz = type.substring(1, type.length() - 1);
        String[] parts = clazz.split("/");

        HashMap<String, DexPackageItem> curChildren = this.children;
        DexPackageItem child = null;
        for (int i = 0; i < parts.length - 1; i++) {
            child = curChildren.get(parts[i]);
            if (child == null) return null;
            curChildren = ((DexPackage) child).children;
        }
        return (DexClass) curChildren.get(type);
    }

    @Override
    public final <RetType, Param> RetType accept(final AnalysisVisitor<RetType, Param> av, Param p) {
        return av.visit(this, p);
    }

    @Override
    public String toString() {
        return getFullName();
    }

    private DexClass insert(List<String> fullName, ClassDef cd) {
        if (fullName.size() == 1) {
            // in package where to insert class
            if (children.get(cd.getType()) != null) {
                // Games from playgendary.com are highly obfuscated containing ~100 dex files with some classes
                // occuring in multiple Dex files. In this case the first class is kept.
                log.warn("tried to insert class " + cd.getType() + " twice");
                return null;
            }

            DexClass newClass = new DexClass(this, cd);
            children.put(cd.getType(), newClass);
            return newClass;
        } else {
            // find or create package containing class
            String packageName = fullName.remove(0);
            DexPackage targetPackage = (DexPackage) children.get(packageName);

            if (targetPackage != null) {
                return targetPackage.insert(fullName, cd);
            } else {
                DexPackage newPckg = new DexPackage(this, packageName);
                children.put(packageName, newPckg);
                return newPckg.insert(fullName, cd);
            }
        }
    }

    public DexClass addClass(ClassDef cd) {
        String clazz = cd.getType();
        List<String> fullName = new LinkedList<>(Arrays.asList(Utils.parseType(clazz)));
        if (!clazz.substring(1, clazz.length() - 1).startsWith(getFullName()))
            throw new IllegalArgumentException("tried to add class " + clazz + " to package " + getFullName());
        return insert(fullName, cd);
    }

    public DexPackage getRandomChildPackage() {
        List<DexPackage> pckChildern = this.children.values().stream()
                .filter(c -> c instanceof DexPackage)
                .map(c -> (DexPackage) c)
                .collect(Collectors.toList());
        if (pckChildern.size() == 0) return null;
        return pckChildern.get(new Random().nextInt(pckChildern.size()));
    }

    public Iterable<DexPackageItem> getChildren() {
        return children.values();
    }
}
