package de.marstra.dvo.obfuscator.analysis;

import lombok.Data;
import org.jf.dexlib2.iface.Method;
import org.jf.dexlib2.iface.MethodImplementation;
import org.jf.dexlib2.iface.instruction.Instruction;

import java.util.HashSet;
import java.util.Set;

public class MethodAnalysis {
    public static final String NO_IMPL = "No implementation";
    public static final String INITIALIZER = "initializer";
    public static final String STATIC_INITIALIZER = "static initializer";
    public static final String MONITOR_INSTRUCTION = "monitor instruction";
    public static final String INVOKE_SUPER = "invoke super";
    public static final String CALLER_SENSITIVE = "caller sensitive invoke";
    public static final String UNSUPPORTED_INS = "unsupported instruction";

    @Data
    public static class MethodAnalysisResult {
        private final boolean supported;
        private final Set<String> unsupportedReasons;
        private final int numInstructions;
        private final int numBytes;
    }

    /**
     * Check wether method m is supported
     *
     * @param m Method to check
     * @return Reason why method is not supported, or null if it is supported
     */
    public static MethodAnalysisResult analyseMethod(Method m, ParsedDex parsedDex) {
        HashSet<String> unsupportedReasons = new HashSet<>();

        if (m.getName().equals("<init>")) {
            unsupportedReasons.add(INITIALIZER);
        }

        if (m.getName().equals("<clinit>")) {
            unsupportedReasons.add(STATIC_INITIALIZER);
        }

        MethodImplementation mimpl = m.getImplementation();

        if (mimpl == null) {
            unsupportedReasons.add(NO_IMPL);
            return new MethodAnalysisResult(false, unsupportedReasons, 0, 0);
        }

        int numInstructions = 0;
        int numBytes = 0;
        for (Instruction ins : mimpl.getInstructions()) {
            String reason = InstructionAnalysis.whyIsInstructionUnSupported(ins, parsedDex);
            if (reason != null) {
                unsupportedReasons.add(reason);
            }
            numInstructions++;
            numBytes += ins.getCodeUnits() * 2;
        }

        return new MethodAnalysisResult(unsupportedReasons.isEmpty(), unsupportedReasons, numInstructions, numBytes);
    }
}
