package de.marstra.dvo.obfuscator.analysis;

import de.marstra.dvo.obfuscator.util.GenericMethodReference;
import lombok.Getter;
import lombok.RequiredArgsConstructor;
import org.jf.dexlib2.Opcodes;
import org.jf.dexlib2.iface.ClassDef;
import org.jf.dexlib2.iface.DexFile;

import java.util.*;

@RequiredArgsConstructor
public class ParsedDex {
    private final DexPackage rootPackage;
    private final List<DexFile> dexFiles;
    private final Set<GenericMethodReference> callerSensitiveMethods;

    public ParsedDex(DexFile origDex) {
        this(Collections.singletonList(origDex));
    }

    public final <RetType, Param> RetType accept(final AnalysisVisitor<RetType, Param> av) {
        return rootPackage.accept(av, null);
    }

    public ParsedDex(List<DexFile> dexFiles) {
        if (dexFiles == null || dexFiles.size() < 1)
            throw new IllegalArgumentException();
        this.dexFiles = dexFiles;
        this.callerSensitiveMethods = CallerSensitiveAnalysis.getCallerSensitiveMethods(dexFiles);
        this.callerSensitiveMethods.addAll(CallerSensitiveAnalysis.getLibCallerSensitiveMethods());
        this.rootPackage = new DexPackage(null, null);
        for (DexFile dexFile : dexFiles) {
            dexFile.getClasses().forEach(rootPackage::addClass);
        }
    }

    public List<DexFile> toDexFiles() {
        ToDexFileVisitor dexFileVisitor = new ToDexFileVisitor(dexFiles.get(0).getOpcodes());
        this.rootPackage.accept(dexFileVisitor, null);
        dexFileVisitor.completeDexFile();
        return dexFileVisitor.getDexFiles();
    }

    public void addClass(ClassDef classDef) {
        rootPackage.addClass(classDef);
    }

    public boolean isCallerSensitive(GenericMethodReference gmr) {
        return callerSensitiveMethods.contains(gmr);
    }

    public String getRandomPackagePath() {
        DexPackage last = null;
        DexPackage cur = rootPackage;
        while (cur != null) {
            last = cur;
            cur = cur.getRandomChildPackage();
        }
        return last.getFullName();
    }

    public DexClass findClass(String name) {
        return rootPackage.findClass(name);
    }

    @RequiredArgsConstructor
    private static class ToDexFileVisitor extends AnalysisVisitorAdapter<Void, Void> {
        @Getter
        private List<DexFile> dexFiles = new LinkedList<>();
        private Set<ClassDef> classes = new HashSet<>();
        private int methods = 0, fields = 0;
        private final Opcodes opcodes;

        @Override
        public Void visit(DexClass dexClass, Void p) {
            int newMethods = dexClass.getMethods().size();
            int newFields = dexClass.getInstanceFields().size() + dexClass.getStaticFields().size();

            // TODO what is maximum amount? How much methods form runtime are referred?
            if ((methods + newMethods) > 0xA000 || (fields + newFields) > 0xA000)
                completeDexFile();

            classes.add(dexClass.toClassDef());
            methods += newMethods;
            fields += newFields;
            return null;
        }

        void completeDexFile() {
            if (classes.size() == 0) return;
            final Set<ClassDef> locClasses = classes;
            dexFiles.add(new DexFile() {
                // NOTE: refer local variable so a reference is passed to the new instance of the anonymous inner class
                // otherwhise $this pointer will be used to refer classes which would lead to all dexFiles containing
                // the same classes
                public Set<? extends ClassDef> getClasses() {
                    return locClasses;
                }

                public Opcodes getOpcodes() {
                    return opcodes;
                }
            });
            classes = new HashSet<>();
            methods = fields = 0;
        }
    }

    public DexMethod findMethod(GenericMethodReference gmr) {
        return rootPackage.findMethod(gmr);
    }
}
