package de.marstra.dvo.obfuscator.insn;

import com.squareup.javapoet.CodeBlock;
import lombok.Getter;

import java.util.function.BiConsumer;

public enum InstructionType {
    // move ops
    MOVE(5, Move::generateCase),
    MOVE_OBJECT(5, Move::generateCase),
    MOVE_RESULT(3, Move::generateCase, "lastResult"),
    MOVE_EXCEPTION(3, Move::generateCase, "lastException"),

    // return ops
    RETURN_VOID(1, Return::generateCase, true),
    RETURN(2, Return::generateCase, false),

    // const ops
    CONST(6, Const::generateCase),
    CONST_STRING(6, Const::generateCase),
    CONST_CLASS(4, Const::generateCase),
    CONST_WIDE(10, Const::generateCase),

    CHECK_CAST(4, CheckCast::generateCase),
    INSTANCE_OF(4, InstanceOf::generateCase),
    ARRAY_LENGTH(2, ArrayLength::generateCase),
    NEW_INSTANCE(4, NewInstance::generateCase),
    NEW_ARRAY(5, NewArray::generateCase),
    FILLED_NEW_ARRAY(6, FilledNewArray::generateCase),
    FILLED_NEW_ARRAY_RANGE(7, FilledNewArrayRange::generateCase),
    FILL_ARRAY_DATA(6, FillArrayData::generateCase),
    THROW(2, Throw::generateCase),
    GOTO(5, Goto::generateCase),
    SWITCH(6, Switch::generateCase),

    // compare operations
    CMPL_FLOAT(4, BinOp::generateCase, "float", "cmpl"),
    CMPG_FLOAT(4, BinOp::generateCase, "float", "cmpg"),
    CMPL_DOUBLE(4, BinOp::generateCase, "double", "cmpl"),
    CMPG_DOUBLE(4, BinOp::generateCase, "double", "cmpg"),
    CMP_LONG(4, BinOp::generateCase, "long", "cmp"),

    // branching operations
    IF_EQ(6, Branch::generateCase, "=="),
    IF_NE(6, Branch::generateCase, "!="),
    IF_LT(6, Branch::generateCase, "<"),
    IF_GE(6, Branch::generateCase, ">="),
    IF_GT(6, Branch::generateCase, ">"),
    IF_LE(6, Branch::generateCase, "<="),

    IF_EQZ(6, BranchZero::generateCase, "=="),
    IF_NEZ(6, BranchZero::generateCase, "!="),
    IF_LTZ(6, BranchZero::generateCase, "<"),
    IF_GEZ(6, BranchZero::generateCase, ">="),
    IF_GTZ(6, BranchZero::generateCase, ">"),
    IF_LEZ(6, BranchZero::generateCase, "<="),

    // load store operations
    AGET(4, ArrayLoadStore::generateCase, false),
    APUT(4, ArrayLoadStore::generateCase, true),
    IGET(5, InstanceLoadStore::generateCase, false),
    IPUT(5, InstanceLoadStore::generateCase, true),
    SGET(5, StaticLoadStore::generateCase, false),
    SPUT(5, StaticLoadStore::generateCase, true),

    // invoke ops
    INVOKE_VIRTUAL(6, Invoke::generateCase),
    INVOKE_DIRECT(6, Invoke::generateCase),
    INVOKE_STATIC(6, Invoke::generateCase),
    INVOKE_INTERFACE(6, Invoke::generateCase),

    INVOKE_VIRTUAL_RANGE(6, InvokeRange::generateCase),
    INVOKE_DIRECT_RANGE(6, InvokeRange::generateCase),
    INVOKE_STATIC_RANGE(6, InvokeRange::generateCase),
    INVOKE_INTERFACE_RANGE(6, InvokeRange::generateCase),

    // unary operations
    NEG_INT(3, UnOp::generateCase, "-$val.intValue()"),
    NOT_INT(3, UnOp::generateCase, "~$val.intValue()"),
    NEG_LONG(3, UnOp::generateCase, "-$val.longValue()"),
    NOT_LONG(3, UnOp::generateCase, "~$val.longValue()"),
    NEG_FLOAT(3, UnOp::generateCase, "-$val.floatValue()"),
    NEG_DOUBLE(3, UnOp::generateCase, "-$val.doubleValue()"),

    INT_TO_LONG(3, UnOp::generateCase, "$val.longValue()"),
    INT_TO_FLOAT(3, UnOp::generateCase, "(float) $val.intValue()"),
    INT_TO_DOUBLE(3, UnOp::generateCase, "(double) $val.intValue()"),

    LONG_TO_INT(3, UnOp::generateCase, "$val.intValue()"),
    LONG_TO_FLOAT(3, UnOp::generateCase, "(float) $val.longValue()"),
    LONG_TO_DOUBLE(3, UnOp::generateCase, "(double) $val.longValue()"),

    FLOAT_TO_INT(3, UnOp::generateCase, "(int) $val.floatValue()"),
    FLOAT_TO_LONG(3, UnOp::generateCase, "(long) $val.floatValue()"),
    FLOAT_TO_DOUBLE(3, UnOp::generateCase, "(double) $val.floatValue()"),

    DOUBLE_TO_INT(3, UnOp::generateCase, "(int) $val.doubleValue()"),
    DOUBLE_TO_LONG(3, UnOp::generateCase, "(long) $val.doubleValue()"),
    DOUBLE_TO_FLOAT(3, UnOp::generateCase, "(float) $val.doubleValue()"),

    INT_TO_BYTE(3, UnOp::generateCase, "$val.byteValue()"),
    INT_TO_CHAR(3, UnOp::generateCase, "$val.charValue()"),
    INT_TO_SHORT(3, UnOp::generateCase, "$val.shortValue()"),

    // binary operations
    ADD_INT(4, BinOp::generateCase, "int", "+"),
    SUB_INT(4, BinOp::generateCase, "int", "-"),
    MUL_INT(4, BinOp::generateCase, "int", "*"),
    DIV_INT(4, BinOp::generateCase, "int", "/"),
    REM_INT(4, BinOp::generateCase, "int", "%"),
    AND_INT(4, BinOp::generateCase, "int", "&"),
    OR_INT(4, BinOp::generateCase, "int", "|"),
    XOR_INT(4, BinOp::generateCase, "int", "^"),
    SHL_INT(4, BinOp::generateCase, "int", "<<"),
    SHR_INT(4, BinOp::generateCase, "int", ">>"),
    USHR_INT(4, BinOp::generateCase, "int", ">>>"),

    ADD_LONG(4, BinOp::generateCase, "long", "+"),
    SUB_LONG(4, BinOp::generateCase, "long", "-"),
    MUL_LONG(4, BinOp::generateCase, "long", "*"),
    DIV_LONG(4, BinOp::generateCase, "long", "/"),
    REM_LONG(4, BinOp::generateCase, "long", "%"),
    AND_LONG(4, BinOp::generateCase, "long", "&"),
    OR_LONG(4, BinOp::generateCase, "long", "|"),
    XOR_LONG(4, BinOp::generateCase, "long", "^"),
    SHL_LONG(4, BinOp::generateCase, "long", "<<"),
    SHR_LONG(4, BinOp::generateCase, "long", ">>"),
    USHR_LONG(4, BinOp::generateCase, "long", ">>>"),

    ADD_FLOAT(4, BinOp::generateCase, "float", "+"),
    SUB_FLOAT(4, BinOp::generateCase, "float", "-"),
    MUL_FLOAT(4, BinOp::generateCase, "float", "*"),
    DIV_FLOAT(4, BinOp::generateCase, "float", "/"),
    REM_FLOAT(4, BinOp::generateCase, "float", "%"),

    ADD_DOUBLE(4, BinOp::generateCase, "double", "+"),
    SUB_DOUBLE(4, BinOp::generateCase, "double", "-"),
    MUL_DOUBLE(4, BinOp::generateCase, "double", "*"),
    DIV_DOUBLE(4, BinOp::generateCase, "double", "/"),
    REM_DOUBLE(4, BinOp::generateCase, "double", "%"),

    // literal binary operations
    ADD_INT_LIT(5, BinOpLit::generateCase, "+"),
    RSUB_INT_LIT(5, BinOpLit::generateCase, "-", true),
    MUL_INT_LIT(5, BinOpLit::generateCase, "*"),
    DIV_INT_LIT(5, BinOpLit::generateCase, "/"),
    REM_INT_LIT(5, BinOpLit::generateCase, "%"),
    AND_INT_LIT(5, BinOpLit::generateCase, "&"),
    OR_INT_LIT(5, BinOpLit::generateCase, "|"),
    XOR_INT_LIT(5, BinOpLit::generateCase, "^"),
    SHL_INT_LIT(5, BinOpLit::generateCase, "<<"),
    SHR_INT_LIT(5, BinOpLit::generateCase, ">>"),
    USHR_INT_LIT(5, BinOpLit::generateCase, ">>>");

    @Getter
    private final int size;
    private final BiConsumer<CodeBlock.Builder, InstructionType> generateCase;
    @Getter
    private final Object[] generateArgs;

    InstructionType(int size, BiConsumer<CodeBlock.Builder, InstructionType> generateCase, Object... generateArgs) {
        this.size = size;
        this.generateCase = generateCase;
        this.generateArgs = generateArgs;
    }

    public void generateCase(CodeBlock.Builder builder) {
        generateCase.accept(builder, this);
    }
}
