package de.marstra.dvo.obfuscator.util;

import lombok.Data;
import lombok.extern.slf4j.Slf4j;

import java.io.File;
import java.io.IOException;
import java.io.InputStream;
import java.nio.charset.StandardCharsets;
import java.util.stream.Collectors;
import java.util.stream.Stream;

@Slf4j
public class CommandExecutor {
    @Data
    public static class CommandResult {
        private final int exitVal;
        private final String stdout;
        private final String stderr;

        private String indent(String orig) {
            return Stream.of(orig.split("\n"))
                    .map((s) -> "\t" + s)
                    .collect(Collectors.joining("\n"));
        }

        @Override
        public String toString() {
            return "exitVal: " + exitVal + "\n" +
                    "stdout: \n" + indent(stdout) + "\n" +
                    "stderr: \n" + indent(stderr) + "\n";
        }
    }

    public static CommandResult executeCommand(String command) throws IOException {
        return executeCommand(command, null, 120);
    }

    private static String strip(String in){
        int boundary = 10000;
        if(in.length() < boundary)
            return in;
        return in.substring(in.length() - boundary);
    }

    public static CommandResult executeCommand(String command, File workingDir, int timeout) throws IOException {
        log.debug("execute " + command);
        try {
            Process p = Runtime.getRuntime().exec(command, null, workingDir);
            StringBuilder[] out = new StringBuilder[]{new StringBuilder(), new StringBuilder()};
            InputStream[] streams = new InputStream[]{p.getInputStream(), p.getErrorStream()};
            byte[] buf = new byte[2048];

            // wait 120 seconds for process to finish and collect output, so the process can write to stdout/stderr
            // without blocking due to full buffers
            for (int sec = 0; sec < (timeout * 100) && p.isAlive(); sec++) {
                for (int stream = 0; stream <= 1; stream++) {
                    int avl = streams[stream].available();
                    if (avl > 0) {
                        int read = streams[stream].read(buf, 0, buf.length);
                        out[stream].append(new String(buf, 0, read, StandardCharsets.UTF_8));
                    }
                }
                Thread.sleep(10);
            }

            // collect remaining output
            for (int stream = 0; stream <= 1; stream++) {
                while(streams[stream].available() > 0) {
                    int read = streams[stream].read(buf, 0, buf.length);
                    out[stream].append(new String(buf, 0, read, StandardCharsets.UTF_8));
                }
            }

            if (p.isAlive()) {
                p.destroyForcibly();
                throw new RuntimeException("executing '" + command + "' timed out\n");
            }

            int exitVal = p.exitValue();
            CommandResult result = new CommandResult(exitVal, strip(out[0].toString()), strip(out[1].toString()));

            if (exitVal != 0) {
                String errorMsg = "executing '" + command + "' failed\n" + result;
                throw new IOException(errorMsg);
            }

            return result;
        } catch (InterruptedException e) {
            throw new RuntimeException(e);
        }
    }
}
