package de.marstra.dvo.obfuscator.util;

import com.android.tools.r8.CompilationFailedException;
import com.android.tools.r8.D8;
import com.android.tools.r8.D8Command;
import com.android.tools.r8.OutputMode;
import lombok.extern.slf4j.Slf4j;
import org.jf.dexlib2.DexFileFactory;
import org.jf.dexlib2.iface.DexFile;

import javax.tools.JavaCompiler;
import javax.tools.ToolProvider;
import java.io.ByteArrayOutputStream;
import java.io.IOException;
import java.io.PrintStream;
import java.nio.charset.StandardCharsets;
import java.nio.file.Files;
import java.nio.file.Path;
import java.nio.file.Paths;
import java.util.Arrays;
import java.util.LinkedList;
import java.util.List;
import java.util.stream.Collectors;
import java.util.stream.Stream;

@Slf4j
public class CompileUtil {

    public static void compileJava(Path srcFolder) {
        compileJava(srcFolder, null);
    }

    /**
     * Compiles all *.java files in srcFolder and its subfolders
     * creates a corresponding class file for each source file in the folder of the source file.
     * additionalFlags are passed to javac
     *
     * @param srcFolder
     * @param additionalFlags
     */
    public static void compileJava(Path srcFolder, List<String> additionalFlags) {
        final JavaCompiler compiler = ToolProvider.getSystemJavaCompiler();
        List<String> argsList = new LinkedList<>();

        // disable annotation preprocessing
        argsList.add("-proc:none");
        argsList.addAll(Arrays.asList("-cp", srcFolder.toString()));
        if (additionalFlags != null)
            argsList.addAll(additionalFlags);

        ByteArrayOutputStream bout = new ByteArrayOutputStream();
        ByteArrayOutputStream berr = new ByteArrayOutputStream();

        try (Stream<Path> fileStream = Files.walk(srcFolder);
             PrintStream pout = new PrintStream(bout, true, StandardCharsets.UTF_8.name());
             PrintStream perr = new PrintStream(berr, true, StandardCharsets.UTF_8.name())
        ) {
            fileStream.filter((f) -> Files.isRegularFile(f) && f.toString().endsWith(".java"))
                    .map(Path::toString)
                    .forEach(argsList::add);

            String[] args = argsList.toArray(new String[0]);

            int res = compiler.run(null, pout, perr, args);

            if (res != 0) {
                String stdOut = bout.toString(StandardCharsets.UTF_8.name());
                String stdErr = berr.toString(StandardCharsets.UTF_8.name());

                throw new RuntimeException("Failed compiling folder '" + srcFolder + "':\n" +
                        "output from javac:\n" +
                        "stdout:\n" +
                        stdOut + "\n" +
                        "stderr:\n" +
                        stdErr
                );
            }
        } catch (IOException e) {
            throw new RuntimeException(e);
        }
    }

    public static Path pathToAndroidJar() {
        String sdkDir = System.getenv("ANDROID_HOME");
        if (sdkDir == null)
            throw new RuntimeException("ANDROID_HOME not set");

        Path lib = Paths.get(sdkDir).resolve("platforms/android-28/android.jar");
        if (!Files.exists(lib))
            throw new RuntimeException("platforms/android-28/android.jar not found");
        return lib;
    }

    /**
     * converts all class files in dir to dalvik bytecode and puts them
     * into a dex file called classes.dex inside dir
     * the dexfile is read with dexlib2 and returend
     *
     * @param dir
     * @return resulting dexfile
     */
    public static DexFile dxClasses(Path dir) {
        Path outDex = dir.resolve("classes.dex");

        try (Stream<Path> fileStream = Files.walk(dir)) {
            List<Path> classFiles = fileStream.filter((f) -> Files.isRegularFile(f) && f.toString().endsWith(".class"))
                    .collect(Collectors.toList());

            D8Command.Builder commandBuilder = D8Command.builder()
                    .setMinApiLevel(19)
                    .setOutput(dir, OutputMode.DexIndexed)
                    .addProgramFiles(classFiles);

            commandBuilder.addLibraryFiles(pathToAndroidJar());

            D8.run(commandBuilder.build());
            return DexFileFactory.loadDexFile(outDex.toString(), null);
        } catch (IOException | CompilationFailedException e) {
            throw new RuntimeException(e);
        }
    }
}
