package de.marstra.dvo.obfuscator.util;

import lombok.extern.slf4j.Slf4j;
import org.apache.commons.io.FileUtils;

import java.io.Closeable;
import java.io.File;
import java.io.IOException;
import java.nio.file.Files;
import java.nio.file.Path;
import java.util.LinkedList;
import java.util.List;
import java.util.Random;
import java.util.function.Function;
import java.util.function.Supplier;

@Slf4j
public class Utils {
    @FunctionalInterface
    public interface ThrowingFunction<T, R> extends Function<T, R> {
        @Override
        default R apply(T t) {
            try {
                return applyThrows(t);
            } catch (Exception e) {
                throw new RuntimeException(e);
            }
        }

        R applyThrows(T t) throws Exception;
    }

    public static <T> T time(String description, Supplier<T> cb) {
        log.info("startTime: " + description);
        long start = System.currentTimeMillis();
        T result = cb.get();
        long duration = System.currentTimeMillis() - start;
        log.info("endTime:   " + description + " lasted " + duration + "ms");
        return result;
    }

    public static String[] parseType(String type) {
        if (type.charAt(0) != 'L' || type.charAt(type.length() - 1) != ';')
            throw new IllegalArgumentException("invalid class type: " + type);
        return type.substring(1, type.length() - 1).split("/");
    }

    public static String getClassName(String type) {
        String[] parts = parseType(type);
        return parts[parts.length - 1];
    }

    public static <T> List<T> iteratorToList(Iterable<T> it) {
        List<T> res = new LinkedList<>();
        it.forEach(res::add);
        return res;
    }

    public static <T> T doWithTmpFile(String prefix, ThrowingFunction<File, T> cb) {
        try {
            File tempFile = File.createTempFile(prefix, "");
            try (Closeable onClose = tempFile::delete) {
                return cb.apply(tempFile);
            }
        } catch (IOException e) {
            throw new RuntimeException(e);
        }
    }

    public static <T> T doInTmpDir(ThrowingFunction<Path, T> cb) {
        return doInTmpDir("", cb);
    }

    public static <T> T doInTmpDir(String prefix, ThrowingFunction<Path, T> cb) {
        try {
            Path tmpDir = Files.createTempDirectory(prefix);
            try (Closeable onClose = () -> FileUtils.deleteDirectory(tmpDir.toFile())) {
                return cb.apply(tmpDir);
            }
        } catch (IOException e) {
            throw new RuntimeException(e);
        }
    }

    public static String randomName() {
        String chars = "ABCDEFGHIJKLMNOPQRSTUVWXYabcdefghijklmnopqrstuvwxy_";
        Random random = new Random();

        int len = 4 + random.nextInt(16);
        char[] result = new char[len];
        for (int i = 0; i < result.length; i++) {
            result[i] = chars.charAt(random.nextInt(chars.length()));
        }
        return new String(result);
    }
}
