package interpreter;

import java.io.*;
import java.nio.ByteBuffer;

public final class VMODex {
    public final String[] strings;
    public final String[] types;
    public final FieldReference[] frefs;
    public final MethodReference[] mrefs;
    public final Method[] methods;

    public VMODex(FieldReference[] frefs, MethodReference[] mrefs, Method[] methods, String[] strings, String[] types) {
        this.frefs = frefs;
        this.mrefs = mrefs;
        this.methods = methods;
        this.strings = strings;
        this.types = types;
    }

    private static byte[] readWholeStream(InputStream is) {
        ByteArrayOutputStream bos = new ByteArrayOutputStream();

        try {
            int read;
            byte[] buf = new byte[1024];
            while ((read = is.read(buf, 0, buf.length)) != -1) {
                bos.write(buf, 0, read);
            }
            bos.flush();
            is.close();
        } catch (IOException e) {
            throw new RuntimeException(e);
        }

        return bos.toByteArray();
    }

    public static VMODex loadVMODexFile(InputStream is, byte[] key) {
        byte data[] = readWholeStream(is);
        return deserialize(data, key);
    }

    public static VMODex loadVMODexFile(String path, byte[] key) {
        try {
            return loadVMODexFile(new FileInputStream(path), key);
        } catch (FileNotFoundException e) {
            throw new RuntimeException(e);
        }
    }

    public void writeVMODexFile(String path, byte[] key) {
        try {
            FileOutputStream fout = new FileOutputStream(path);
            fout.write(serialize(key));
            fout.close();
        } catch (IOException e) {
            throw new RuntimeException(e);
        }
    }

    public byte[] serialize(byte[] key) {
        ByteArrayOutputStream baout = new ByteArrayOutputStream();

        try {
            baout.write(Utils.intToBytes(strings.length));
            for (String s : strings) {
                Utils.serializeString(baout, s);
            }

            baout.write(Utils.intToBytes(types.length));
            for (String s : types) {
                Utils.serializeString(baout, s);
            }

            baout.write(Utils.intToBytes(frefs.length));
            for (FieldReference f : frefs) {
                baout.write(f.serialize());
            }

            baout.write(Utils.intToBytes(mrefs.length));
            for (MethodReference m : mrefs) {
                baout.write(m.serialize());
            }

            baout.write(Utils.intToBytes(methods.length));
            for (Method m : methods) {
                baout.write(m.serialize());
            }
        } catch (IOException e) {
            throw new RuntimeException(e);
        }
        byte[] plain = baout.toByteArray();
        byte[] enc = new byte[plain.length];
        for (int i = 0; i < plain.length; i++) {
            enc[i] = (byte) (plain[i] ^ key[i % key.length]);
        }
        return enc;
    }

    public static VMODex deserialize(byte enc[], byte key[]) {
        byte[] plain = new byte[enc.length];
        for (int i = 0; i < plain.length; i++) {
            plain[i] = (byte) (enc[i] ^ key[i % key.length]);
        }

        ByteBuffer byteBuf = ByteBuffer.wrap(plain);
        String[] strings = new String[byteBuf.getInt()];
        for (int i = 0; i < strings.length; i++) {
            strings[i] = Utils.deserializeString(byteBuf);
        }

        String[] types = new String[byteBuf.getInt()];
        for (int i = 0; i < types.length; i++) {
            types[i] = Utils.deserializeString(byteBuf);
        }

        FieldReference[] frefs = new FieldReference[byteBuf.getInt()];
        for (int i = 0; i < frefs.length; i++) {
            frefs[i] = FieldReference.deserialize(byteBuf);
        }

        MethodReference[] mrefs = new MethodReference[byteBuf.getInt()];
        for (int i = 0; i < mrefs.length; i++) {
            mrefs[i] = MethodReference.deserialize(byteBuf);
        }

        Method[] methods = new Method[byteBuf.getInt()];
        for (int i = 0; i < methods.length; i++) {
            methods[i] = Method.deserialize(byteBuf);
        }
        return new VMODex(frefs, mrefs, methods, strings, types);
    }
}
