package interpreter;

import java.io.ByteArrayOutputStream;
import java.io.IOException;
import java.lang.reflect.Constructor;
import java.lang.reflect.Field;
import java.lang.reflect.Method;
import java.nio.ByteBuffer;
import java.nio.charset.StandardCharsets;
import java.util.Random;

public final class Utils {
    public static byte[] intToBytes(int val) {
        return ByteBuffer.allocate(Integer.BYTES).putInt(val).array();
    }

    public static byte[] longToBytes(long ins) {
        return ByteBuffer.allocate(Long.BYTES).putLong(ins).array();
    }

    public static boolean isWideType(String type) {
        char indicator = type.charAt(0);
        return indicator == 'J' || indicator == 'D';
    }

    public static byte[] getRandomBytes(int len) {
        Random rand = new Random();
        byte[] result = new byte[len];
        rand.nextBytes(result);
        return result;
    }

    public static void serializeString(ByteArrayOutputStream baout, String string) {
        try {
            if (string == null) {
                baout.write(Utils.intToBytes(-1));
                return;
            }
            byte[] data = string.getBytes(StandardCharsets.UTF_8);
            baout.write(Utils.intToBytes(data.length));
            baout.write(data);
        } catch (IOException e) {
            throw new RuntimeException(e);
        }
    }

    public static String deserializeString(ByteBuffer buf) {
        int len = buf.getInt();
        if(len == -1) return null;
        byte[] data = new byte[len];
        buf.get(data);
        return new String(data, StandardCharsets.UTF_8);
    }

    public static Class loadClassByName(String type, ClassLoader loader) {
        char indicator = type.charAt(0);
        switch (indicator) {
            case 'L': {
                String jType = type.substring(1, type.length() - 1);
                jType = jType.replaceAll("/", ".");
                try {
                    return Class.forName(jType, false, loader);
                } catch (ClassNotFoundException e) {
                    throw new RuntimeException(e);
                }
            }
            case 'V':
                return void.class;
            case 'Z':
                return boolean.class;
            case 'B':
                return byte.class;
            case 'S':
                return short.class;
            case 'C':
                return char.class;
            case 'I':
                return int.class;
            case 'J':
                return long.class;
            case 'F':
                return float.class;
            case 'D':
                return double.class;
            case '[':
                try {
                    return Class.forName(type.replaceAll("/", "."), false, loader);
                } catch (ClassNotFoundException e) {
                    throw new RuntimeException(e);
                }
            default:
                throw new IllegalArgumentException("Unknown Type:  " + type);
        }
    }

    public static String primitiveNameToType(String name) {
        switch (name) {
            case "boolean":
                return "Z";
            case "byte":
                return "B";
            case "char":
                return "C";
            case "short":
                return "S";
            case "int":
                return "I";
            case "float":
                return "F";
            case "long":
                return "J";
            case "double":
                return "D";
            default:
                throw new IllegalArgumentException("Unsupported primitive type: " + name);
        }
    }

    public static Object vmTypeToExternalType(String type, Object val) {
        char indicator = type.charAt(0);
        switch (indicator) {
            case '[':
            case 'L':
                // handle passing of null which gets created as "const/4 vX, 0x0"
                if (val instanceof InterpreterNumber) {
                    if (((InterpreterNumber) val).intValue() != 0) {
                        throw new IllegalArgumentException("Attempt to convert primitive type != 0 to object");
                    }
                    return null;
                }
                return val;
            case 'Z':
                return ((InterpreterNumber) val).booleanValue();
            case 'B':
                return ((InterpreterNumber) val).byteValue();
            case 'S':
                return ((InterpreterNumber) val).shortValue();
            case 'C':
                return ((InterpreterNumber) val).charValue();
            case 'I':
                return ((InterpreterNumber) val).intValue();
            case 'J':
                return ((InterpreterNumber) val).longValue();
            case 'F':
                return ((InterpreterNumber) val).floatValue();
            case 'D':
                return ((InterpreterNumber) val).doubleValue();
            default:
                throw new IllegalArgumentException("vmTypeToExternalType: Unsupported Type: " + type);
        }
    }

    public static Object externalTypeToVmType(String type, Object val) {
        char indicator = type.charAt(0);
        switch (indicator) {
            case '[':
            case 'V':
            case 'L':
                return val;
            case 'B':
                return new InterpreterNumber((Byte) val);
            case 'S':
                return new InterpreterNumber((Short) val);
            case 'I':
                return new InterpreterNumber((Integer) val);
            case 'J':
                return new InterpreterNumber((Long) val);
            case 'Z':
                return new InterpreterNumber((Boolean) val);
            case 'C':
                return new InterpreterNumber((Character) val);
            case 'F':
                return new InterpreterNumber((Float) val);
            case 'D':
                return new InterpreterNumber((Double) val);
            default:
                throw new IllegalArgumentException("externalTypeToVmType: Unsupported Type: " + type);
        }
    }

    public static Field getField(Class targetClass, String fieldName) {
        Field field = null;
        Class curClass = targetClass;
        while (field == null && curClass != null) {
            try {
                field = curClass.getDeclaredField(fieldName);
            } catch (NoSuchFieldException e) {
            }
            curClass = curClass.getSuperclass();
        }
        return field;
    }

    public static Method getMethod(Class targetClass, String methodName, Class[] pTypes) {
        Method method = null;
        Class curClass = targetClass;

        // get default methods from interface methods
        try {
            method = curClass.getMethod(methodName, pTypes);
        } catch (NoSuchMethodException e) {
        }

        while (method == null && curClass != null) {
            try {
                method = curClass.getDeclaredMethod(methodName, pTypes);
            } catch (NoSuchMethodException e) {
            }
            curClass = curClass.getSuperclass();
        }
        method.setAccessible(true);
        return method;
    }

    public static Constructor getConstructor(Class targetClass, Class[] pTypes) {
        Constructor constructor = null;
        Class curClass = targetClass;
        while (constructor == null && curClass != null) {
            try {
                constructor = curClass.getDeclaredConstructor(pTypes);
            } catch (NoSuchMethodException e) {
            }
            curClass = curClass.getSuperclass();
        }
        constructor.setAccessible(true);
        return constructor;
    }
}
