package de.marstra.dvo.obfuscator.obfuscator;

import org.intellij.lang.annotations.Language;
import org.junit.Test;

import java.util.Random;

public class ControlFlowTest {
    @Test
    public void testJumps() {
        @Language("java") String code = "" +
                "public class Test {" +
                "    public static int test(short val){\n" +
                "        if(val > 3) return 1;\n" +
                "        return 0;\n" +
                "    }\n" +
                "    public static Object test(Object[] args){\n" +
                "        return test((short) ((Integer) args[0]).intValue());\n" +
                "    }\n" +
                "}";
        new InterpreterTestCase(code, "LTest;->test(S)I")
                .addInputSample(4)
                .addInputSample(3)
                .addInputSample(2)
                .addInputSample(new Random().nextInt())
                .run();
    }

    @Test
    public void testCall() {
        @Language("java") String code = "" +
                "public class Test {\n" +
                "    public static int callee(int arg) {\n" +
                "       return arg * arg;\n" +
                "    }\n" +
                "    public static int test(int val){\n" +
                "        return callee(val);\n" +
                "    }\n" +
                "    public static Object test(Object[] args){\n" +
                "        return test((Integer) args[0]);\n" +
                "    }\n" +
                "}";

        new InterpreterTestCase(code, "LTest;->test(I)I", "LTest;->callee(I)I")
                .addInputSample(42)
                .run();
    }

    @Test
    public void testCallVarargs() {
        @Language("java") String code = "" +
                "public class Test {\n" +
                "    public static int callee(int a, int... b) {\n" +
                "       for (int i = 0; i < b.length; i++) a += b[i];\n" +
                "       return a;\n" +
                "    }\n" +
                "    public static int test(int val){\n" +
                "        return callee(val,2,3,4);\n" +
                "    }\n" +
                "    public static Object test(Object[] args){\n" +
                "        return test((Integer) args[0]);\n" +
                "    }\n" +
                "}";

        new InterpreterTestCase(code, "LTest;->test(I)I", "LTest;->callee(I[I)I")
                .addInputSample(1)
                .run();
    }

    @Test
    public void testCallRange() {
        @Language("java") String code = "" +
                "public class Test {\n" +
                "    public static long callee(int a, int b, long c, int d, int e, int f, long g, double h, long i, long j, long k) {\n" +
                "       return a + b + c + d + e + f;\n" +
                "    }\n" +
                "    public static long test(int val){\n" +
                "        return callee(val, 2 * val, 3 * val, 4 * val, 5 * val, 6 * val, 7 * val, 8 * val, 9 * val, 10 * val, 11 * val);\n" +
                "    }\n" +
                "    public static Object test(Object[] args){\n" +
                "        return test((Integer) args[0]);\n" +
                "    }\n" +
                "}";
        new InterpreterTestCase(code, "LTest;->test(I)J", "LTest;->callee(IIJIIIJDJJJ)J")
                .addInputSample(42)
                .run();
    }

    @Test
    public void testLambda() {
        @Language("java") String code = "" +
                "public class Test {\n" +
                "    public static int test() {\n" +
                "        return java.util.stream.Stream.of(1,2,3)\n" +
                "                .map((n) -> n*n)\n" +
                "                .reduce((sum, c) -> sum + c).get();\n" +
                "    }\n" +
                "    public static Object test(Object[] args){\n" +
                "        return test();\n" +
                "    }\n" +
                "}";
        // lambdas need to run in DVM execution environment
        // because they get converted to classes in dex which are not present in class files
        // NOTE might work in JVM executor when d8 is called with no-desugaring
        new InterpreterTestCase(code, InterpreterTestCase.DVM_EXEC, "LTest;->test()I").run();
    }

    @Test
    public void testMethodRef() {
        @Language("java") String code = "" +
                "import java.util.stream.IntStream;\n" +
                "public class Test {\n" +
                "    private static int square(int val) {\n" +
                "        return val * val;\n" +
                "    }\n" +
                "    public static int test() {\n" +
                "        return IntStream.of(1,2,3).map(Test::square).sum();\n" +
                "    }\n" +
                "    public static Object test(Object[] args){\n" +
                "        return test();\n" +
                "    }\n" +
                "}";
        new InterpreterTestCase(code, InterpreterTestCase.DVM_EXEC, "LTest;->test()I", "LTest;->square(I)I").run();
    }

//    TODO support invoke polymorphic
//    @Test
//    public void testInvokePolymorphic() {
//        @Language("java") String code = "" +
//                "import java.lang.invoke.MethodHandle;\n" +
//                "import java.lang.invoke.MethodHandles;\n" +
//                "import java.lang.invoke.MethodType;\n" +
//                "public class Test {\n" +
//                "    private static int square(int val) {\n" +
//                "        return val * val;\n" +
//                "    }\n" +
//                "    public static int test() throws Throwable {\n" +
//                "        MethodType type = MethodType.methodType(int.class, int.class);\n" +
//                "        MethodHandle handle = MethodHandles.lookup().findStatic(Test.class, \"square\", type);\n" +
//                "        return (Integer) handle.invoke(3);\n" +
//                "    }\n" +
//                "    public static Object test(Object[] args){\n" +
//                "        try{return test();}catch(Throwable t){return \"failed\";}\n" +
//                "    }\n" +
//                "}";
//        new InterpreterTestCase(code, InterpreterTestCase.JVM_EXEC, "LTest;->test()I").run();
//    }

    @Test
    public void testPackedSwitch() {
        @Language("java") String code = "" +
                "public class Test {\n" +
                "    public static int test(int dec){\n" +
                "        switch(dec) {\n" +
                "            case 1:\n" +
                "                return 3;\n" +
                "            case 2:\n" +
                "                return 42;\n" +
                "            case 3:\n" +
                "                return 18;\n" +
                "            default:\n" +
                "                return 38;\n" +
                "        }\n" +
                "    }\n" +
                "    public static Object test(Object[] args){\n" +
                "        return test((Integer) args[0]);\n" +
                "    }\n" +
                "}";
        new InterpreterTestCase(code, "LTest;->test(I)I")
                .addInputSample(1)
                .addInputSample(2)
                .addInputSample(3)
                .run();
    }

    @Test
    public void testSparseSwitch() {
        @Language("java") String code = "" +
                "public class Test {\n" +
                "    public static int test(int dec){\n" +
                "        int result = 0;\n" +
                "        switch(dec) {\n" +
                "            case 1:\n" +
                "                result += 1;\n" +
                "            case 2:\n" +
                "                result += 2;\n" +
                "                break;\n" +
                "            case 10:\n" +
                "                result += 10;\n" +
                "            default:\n" +
                "                result += 100;\n" +
                "        }\n" +
                "        result += 1000;\n" +
                "        return result;\n" +
                "    }\n" +
                "    public static Object test(Object[] args){\n" +
                "        return test((Integer) args[0]);\n" +
                "    }\n" +
                "}";
        new InterpreterTestCase(code, "LTest;->test(I)I")
                .addInputSample(1)
                .addInputSample(2)
                .addInputSample(42)
                .addInputSample(-1)
                .run();
    }

    // this testcase tests if multiple payloads of different type are handled correctly
    @Test
    public void testNestedSwitch() {
        @Language("java") String code = "" +
                "public class Test {\n" +
                "    public static int test(int a, int b) {\n" +
                "        int[] res = new int[]{1,2,3,4,5,6,7,8,9,10,11};\n" +
                "        switch (a) {\n" +
                "            case 1:\n" +
                "                switch (b) {\n" +
                "                    case 1: return res[1];\n" +
                "                    case 10: return res[2];\n" +
                "                    case 100: return res[3];\n" +
                "                }\n" +
                "            case 2:\n" +
                "                return 80;\n" +
                "            case 3:\n" +
                "                switch (b) {\n" +
                "                    case -10: return res[4];\n" +
                "                    case 0: return res[5];\n" +
                "                    case 10: return res[6];\n" +
                "                    default: return res[7];\n" +
                "                }\n" +
                "            case 4:\n" +
                "                switch (b) {\n" +
                "                    case 300: return res[8];\n" +
                "                    case 301: return res[9];\n" +
                "                    case 302: return res[10];\n" +
                "                }\n" +
                "        }\n" +
                "        return -1;\n" +
                "    }\n" +
                "    public static Object test(Object[] args){\n" +
                "        return test((Integer) args[0], (Integer) args[1]);\n" +
                "    }\n" +
                "}";
        new InterpreterTestCase(code, "LTest;->test(II)I")
                .addInputSample(1, 1)
                .addInputSample(1, 10)
                .addInputSample(1, 100)
                .addInputSample(1, 30)
                .addInputSample(2, 0)
                .addInputSample(3, -10)
                .addInputSample(3, 0)
                .addInputSample(3, 10)
                .addInputSample(3, 4)
                .addInputSample(4, 300)
                .addInputSample(4, 301)
                .addInputSample(4, 302)
                .addInputSample(4, 303)
                .addInputSample(7, 7)
                .run();
    }

    @Test
    public void testStringSwitch() {
        @Language("java") String code = "" +
                "public class Test {\n" +
                "    public static int test(String dec) {\n" +
                "        switch(dec) {\n" +
                "            case \"aaaa\":\n" +
                "                return 1;\n" +
                "            case \"bbbb\":\n" +
                "                return 2;\n" +
                "            default:\n" +
                "                return 3;\n" +
                "        }\n" +
                "    }\n" +
                "    public static Object test(Object[] args){\n" +
                "        return test((String) args[0]);\n" +
                "    }\n" +
                "}";

        new InterpreterTestCase(code, "LTest;->test(Ljava/lang/String;)I")
                .addInputSample("aaaa")
                .addInputSample("bbbb")
                .addInputSample("cccc")
                .run();
    }

    @Test
    public void testFilledNewArray() {
        @Language("java") String code = "" +
                "public class Test {\n" +
                "    public static int[][] test(int a, int b) {\n" +
                "        return new int[a][b];\n" +
                "    }\n" +
                "    public static Object test(Object[] args){\n" +
                "        return test((int) args[0], (int) args[1]);\n" +
                "    }\n" +
                "}";
        new InterpreterTestCase(code, "LTest;->test(II)[[I")
                .addInputSample(4,5)
                .run();
    }

    @Test
    public void testFilledNewArrayRange() {
        @Language("java") String code = "" +
                "public class Test {\n" +
                "    public static int[][][][][][] test(int a, int b, int c, int d, int e, int f) {\n" +
                "        return new int[a][b][c][d][e][f];\n" +
                "    }\n" +
                "    public static Object test(Object[] args){\n" +
                "        return test((int) args[0], (int) args[1], (int) args[2], (int) args[3], (int) args[4], (int) args[5]);\n" +
                "    }\n" +
                "}";
        new InterpreterTestCase(code, "LTest;->test(IIIIII)[[[[[[I")
                .addInputSample(4,5,3,5,1,2)
                .run();
    }
}
