package de.marstra.dvo.obfuscator.obfuscator;

import org.intellij.lang.annotations.Language;
import org.junit.Test;

import java.io.IOException;

public class ExceptionHandlingTest {
    @Test
    public void testThrow() {
        @Language("java") String code = "" +
                "public class Test {\n" +
                "    public static void test() throws Exception {\n" +
                "        throw new Exception(\"Exception Message\");\n" +
                "    }\n" +
                "    public static Object test(Object[] args){\n" +
                "        try {test();} catch (Exception e) {return e.getMessage();}" +
                "        return null;\n" +
                "    }\n" +
                "}";
        new InterpreterTestCase(code, "LTest;->test()V").run();
    }

    @Test
    public void testArithException() {
        @Language("java") String code = "" +
                "public class Test {\n" +
                "    public static int test(int a, int b) {\n" +
                "        return a / b;\n" +
                "    }\n" +
                "    public static Object test(Object[] args){\n" +
                "        try {return test((Integer) args[0], (Integer) args[1]);} catch (Exception e) {return e.getMessage();}" +
                "    }\n" +
                "}";
        new InterpreterTestCase(code, "LTest;->test(II)I", "LTest;->test([Ljava/lang/Object;)Ljava/lang/Object;")
                .addInputSample(42, 3)
                .addInputSample(42, 0)
                .run();
    }

    @Test
    public void testInvokeException() {
        @Language("java") String code = "" +
                "public class Test {\n" +
                "    public static void fail(){\n" +
                "        throw new IllegalStateException(\"Something went terribly wrong\");\n" +
                "    }\n" +
                "    public static void test() {\n" +
                "        fail();\n" +
                "    }" +
                "    public static Object test(Object[] args){\n" +
                "        try {test();} catch (IllegalStateException e) {return e.getMessage();}" +
                "        return null;\n" +
                "    }\n" +
                "}";
        new InterpreterTestCase(code, "LTest;->test()V", "LTest;->fail()V").run();
    }

    @Test
    public void testTryCatch() {
        @Language("java") String code = "" +
                "public class Test {\n" +
                "    public static String test() {\n" +
                "        try {\n" +
                "            throw new Exception(\"Exception Message\");\n" +
                "        }\n" +
                "        catch(Exception e){\n" +
                "            return e.getMessage();\n" +
                "        }\n" +
                "    }\n" +
                "    public static Object test(Object[] args){\n" +
                "        return test();\n" +
                "    }\n" +
                "}";

        new InterpreterTestCase(code, "LTest;->test()Ljava/lang/String;").run();
    }

    @Test
    public void testTryCatchFinally() {
        @Language("java") String code = "" +
                "public class Test {\n" +
                "    public static int test(int div1, int div2) {\n" +
                "        int res = 1024;\n" +
                // divide by variable so exception might is theoretically possible (DivByZero)
                // => additional exceptionhandler for catch clause is created
                "        try {\n" +
                "            res /= div1;\n" +
                "        }\n" +
                "        catch(Exception e) {\n" +
                "            res /= div2;\n" +
                "            return -1;\n" +
                "        }\n" +
                "        finally {\n" +
                "            res += 0x42;\n" +
                "        }\n" +
                "        return res;\n" +
                "    }\n" +
                "    public static Object test(Object[] args){\n" +
                "        try {return test((Integer) args[0], (Integer) args[1]);} catch (Exception e) {return e.getMessage();}" +
                "    }\n" +
                "}";
        new InterpreterTestCase(code, "LTest;->test(II)I")
                .addInputSample(0, 0)
                .addInputSample(2, 2)
                .addInputSample(0, 2)
                .addInputSample(2, 0)
                .run();
    }

    @Test
    public void testNestedTrys() {
        @Language("java") String code = "" +
                "public class Test {\n" +
                "    public static int test(Exception exp) {\n" +
                "        try {\n" +
                "            try {\n" +
                "                throw exp;\n" +
                "            } catch (IllegalArgumentException e) {\n" +
                "                return 1;\n" +
                "            } catch (RuntimeException e) {\n" +
                "                return 2;\n" +
                "            }\n" +
                "        } catch (Exception e) {\n" +
                "            return 3;\n" +
                "        }\n" +
                "    }\n" +
                "    public static Object test(Object[] args){\n" +
                "        String result = \"\";\n" +
                "        try {result += test(new IllegalArgumentException());} catch (Exception e) {result += e.getMessage();}\n" +
                "        try {result += test(new RuntimeException());} catch (Exception e) {result += e.getMessage();}\n" +
                "        try {result += test(new ArithmeticException());} catch (Exception e) {result += e.getMessage();}\n" +
                "        try {result += test(new java.io.IOException());} catch (Exception e) {result += e.getMessage();}\n" +
                "        try {result += test((Exception) null);} catch (Exception e) {result += e.getMessage();}\n" +
                "        return result;" +
                "    }\n" +
                "\n" +
                "}";
        new InterpreterTestCase(code, "LTest;->test(Ljava/lang/Exception;)I").run();
    }

    @Test
    public void tryWithResource() {
        @Language("java") String code = "" +
                "class Resource implements AutoCloseable {\n" +
                "    @Override\n" +
                "    public void close() throws Exception {\n" +
                "        throw new Exception(\"failed to close\");\n" +
                "    }\n" +
                "}\n" +
                "\n" +
                "public class Test {\n" +
                "    public static void test() {\n" +
                "        try (Resource r = new Resource()) {\n" +
                "            throw new Exception(\"exception in body\");\n" +
                "        } catch (Exception e) {\n" +
                "            if (!e.getSuppressed()[0].getMessage().equals(\"failed to close\"))\n" +
                "                throw new IllegalStateException(\"suppressed exception missing\");\n" +
                "            if (!e.getMessage().equals(\"exception in body\"))\n" +
                "                throw new IllegalStateException(\"catched wrong exception\");\n" +
                "        }\n" +
                "    }\n" +
                "    public static Object test(Object[] args){\n" +
                "        test();\n" +
                "        return null;\n" +
                "    }\n" +
                "}\n";
        new InterpreterTestCase(code, "LTest;->test()V").run();
    }

    @Test
    public void testFailingConstructor(){
        @Language("java") String code = "" +
                "public class Test {\n" +
                "    final boolean value;\n" +
                "    Test(int num){\n" +
                "        value = (500 / num) > 10;\n" +
                "    }\n" +
                "    public static Boolean test(int a) {\n" +
                "        try {\n" +
                "             return new Test(a).value;\n" +
                "        } catch (ArithmeticException e) {\n" +
                "             return null;\n" +
                "        }\n" +
                "    }\n" +
                "    public static Object test(Object[] args){\n" +
                "        return test((int) args[0]);\n" +
                "    }\n" +
                "}\n";
        new InterpreterTestCase(code, "LTest;->test(I)Ljava/lang/Boolean;")
                .addInputSample(10)
                .addInputSample(100)
                .addInputSample(0)
                .run();
    }

    @Test
    public void testFailingSuperCall(){
        @Language("java") String code = "" +
                "class Base {\n" +
                "    Boolean method(int num) {return (500 / num) > 10;}\n" +
                "}\n" +
                "public class Test extends Base {\n" +
                "    Boolean method(int num) {return false;}\n" +
                "    Boolean test(int num) {\n" +
                "        try{\n" +
                "             return super.method(num);\n" +
                "        } catch (ArithmeticException e) {\n" +
                "             return null;" +
                "        }\n" +
                "    }\n" +
                "    public static Object test(Object[] args){\n" +
                "        return new Test().test((int) args[0]);\n" +
                "    }\n" +
                "}";
        new InterpreterTestCase(code, "LTest;->test(I)Ljava/lang/Boolean;", "LTest;->method(I)Ljava/lang/Boolean;", "LBase;->method(I)Ljava/lang/Boolean;")
                .addInputSample(10)
                .addInputSample(100)
                .addInputSample(0)
                .run();
    }
}
