package de.marstra.dvo.obfuscator.obfuscator;

import com.squareup.javapoet.*;
import de.marstra.dvo.obfuscator.ByteCodeMapping;
import de.marstra.dvo.obfuscator.InterpreterGenerator;
import de.marstra.dvo.obfuscator.StubInjector;
import de.marstra.dvo.obfuscator.VMODexCompiler;
import de.marstra.dvo.obfuscator.analysis.ParsedDex;
import de.marstra.dvo.obfuscator.insn.InstructionType;
import de.marstra.dvo.obfuscator.util.CommandExecutor;
import de.marstra.dvo.obfuscator.util.CompileUtil;
import de.marstra.dvo.obfuscator.util.GenericMethodReference;
import interpreter.Utils;
import interpreter.VMODex;
import org.apache.commons.io.FileUtils;
import org.jf.dexlib2.DexFileFactory;
import org.jf.dexlib2.iface.DexFile;

import javax.lang.model.element.Modifier;
import java.io.File;
import java.io.IOException;
import java.lang.reflect.InvocationTargetException;
import java.lang.reflect.Method;
import java.net.URL;
import java.net.URLClassLoader;
import java.nio.charset.StandardCharsets;
import java.nio.file.Files;
import java.nio.file.Path;
import java.util.*;
import java.util.stream.Collectors;
import java.util.stream.Stream;

import static de.marstra.dvo.obfuscator.InterpreterGenerator.*;

abstract class TestExecutor {
    protected Path tmpDir;
    protected ByteCodeMapping byteCodeMapping;
    protected Path dexDir;
    protected Path dexFilePath;
    protected ParsedDex parsedDex;
    protected static byte[] vmodexKey = Utils.getRandomBytes(16);

    public TestExecutor(ByteCodeMapping byteCodeMapping) {
        this.byteCodeMapping = byteCodeMapping;
    }

    private static String patchClassRef(String ref) {
        // only patch refs to classes in default package
        if (ref.charAt(0) != 'L' || ref.contains("/"))
            return ref;
        return "Lpatched/" + ref.substring(1);
    }

    /**
     * compile testCase with code source in tempDir
     * tmpDir/dexDir/orig: contains unpatched source and class files
     * tmpDir/dexDir/patched: contains unpatched source and class files
     * tmpDir/dexDir/<additional>: might contain additional sources such as main class which runs test classes
     * tmpDir/dexDir/classes.dex contains 2 times the classes to test and
     *
     * @param code
     * @param additionalSources
     */
    protected void compileTestCase(String code, List<JavaFile> additionalSources) {
        this.dexDir = tmpDir.resolve("dexDir");
        this.dexFilePath = dexDir.resolve("classes.dex");
        try {
            Files.createDirectory(dexDir);

            for (String version : new String[]{"orig", "patched"}) {
                Path pckDir = tmpDir.resolve(version);
                Files.createDirectory(pckDir);
                String pckCode = "package " + version + ";\n" + code;
                Files.write(pckDir.resolve("Test.java"), pckCode.getBytes(StandardCharsets.UTF_8));
                FileUtils.copyDirectoryToDirectory(pckDir.toFile(), dexDir.toFile());
            }

            for (JavaFile jf : additionalSources) {
                jf.writeTo(dexDir);
            }

            // compile java sources to *.class
            CompileUtil.compileJava(dexDir);
            // compile *.class to classes.dex
            this.parsedDex = new ParsedDex(CompileUtil.dxClasses(dexDir));
        } catch (IOException e) {
            throw new RuntimeException(e);
        }
    }

    public void runTest(String code, GenericMethodReference entryPoint, Set<GenericMethodReference> toPatch, List<Object[]> inputs) {
        try {
            // adjust toPatch to refer correct package
            toPatch = toPatch.stream().map((gmr) ->
                    new GenericMethodReference(
                            patchClassRef(gmr.getDefiningClass()),
                            gmr.getName(),
                            patchClassRef(gmr.getReturnType()),
                            Stream.of(gmr.getParameterTypes()).map(TestExecutor::patchClassRef).toArray(String[]::new)
                    )
            ).collect(Collectors.toSet());

            tmpDir = Files.createTempDirectory("interpTest");
            execute(code, entryPoint, toPatch, inputs);
        } catch (IOException e) {
            throw new RuntimeException(e);
        } finally {
            if (tmpDir != null) {
                try {
                    FileUtils.deleteDirectory(tmpDir.toFile());
                } catch (IOException e) {
                    throw new RuntimeException(e);
                }
            }
        }
    }

    protected abstract void execute(String code, GenericMethodReference entryPoint, Set<GenericMethodReference> toPatch, List<Object[]> inputs);
}

class JVMExecutor extends TestExecutor {
    private static final boolean DEBUG = true;
    protected Path interpreterIdeaPath;

    public JVMExecutor(ByteCodeMapping byteCodeMapping) {
        super(byteCodeMapping);
    }

    private Method findMethod(ClassLoader classLoader, Class clazz, GenericMethodReference ref) throws NoSuchMethodException {
        Class[] params = Stream.of(ref.getParameterTypes()).map((s) -> {
            return Utils.loadClassByName(s, classLoader);
        }).toArray(Class[]::new);
        return clazz.getMethod(ref.getName(), params);
    }

    @Override
    protected void execute(String code, GenericMethodReference entryPoint, Set<GenericMethodReference> toPatch, List<Object[]> inputs) {
        compileTestCase(code, Collections.emptyList());
        VMODexCompiler.CompilationResult result = VMODexCompiler.compile(parsedDex, toPatch, this.byteCodeMapping);

        try {
            // generate and load interpreter and set vmodex
            InterpreterGenerator interpGen = new InterpreterGenerator(byteCodeMapping, new HashSet<>(Arrays.asList(InstructionType.values())),
                    Collections.emptySet(), vmodexKey, true);
//            InterpreterGenerator interpGen = new InterpreterGenerator(byteCodeMapping, new HashSet<>(Arrays.asList(InstructionType.values())), new HashSet<>(Arrays.asList(VerbosityLevel.EVERY_INSTRUCTION)), true);
            interpGen.compile(dexDir);

            if (DEBUG) {
                // Hack to allow source level debugging of generated interpreter using jetbrains idea:
                // Copy generated interpreter source to sources recognized by idea
                Path interpreterSource = dexDir.resolve(INTERPRETER_PACKAGE_NAME).resolve(INTERPRETER_CLASS_NAME + ".java");
                Path codeLoc = new File(de.marstra.dvo.obfuscator.obfuscator.InterpreterTestCase.class.getProtectionDomain().getCodeSource().getLocation().getPath()).toPath();
                interpreterIdeaPath = codeLoc.resolve("../../src/main/java/" + INTERPRETER_PACKAGE_NAME + "/" + INTERPRETER_CLASS_NAME + ".java");
                FileUtils.copyFile(interpreterSource.toFile(), interpreterIdeaPath.toFile());
            }

            URLClassLoader classLoader = new URLClassLoader(new URL[]{
                    dexDir.toFile().toURL()
            });
            Class interpreterClass = classLoader.loadClass(INTERPRETER_PACKAGE_NAME + "." + INTERPRETER_CLASS_NAME);
            Method interpretMethod = interpreterClass.getMethod(INTERPRETER_METHOD_NAME, int.class, Object[].class);
            Method setVmoDexMethod = interpreterClass.getMethod("setVmoDex", VMODex.class);
            setVmoDexMethod.invoke(null, result.getVmoDex());

            // patch class files
            Path[] classFilesToPatch = Arrays.stream(dexDir.resolve("patched").toFile().listFiles())
                    .filter((f) -> f.getName().endsWith("class"))
                    .map(File::toPath)
                    // sort classes so they get loaded in the right order
                    .sorted()
                    .toArray(Path[]::new);
            StubInjector.injectInterpreterClassFiles(result.getConvertedMethodReferences(), classFilesToPatch);

            Class origClass = classLoader.loadClass("orig.Test");
            Class patchedClass = classLoader.loadClass("patched.Test");
            Method patchedMethod = findMethod(classLoader, patchedClass, entryPoint);
            patchedMethod.setAccessible(true);
            Method origMethod = findMethod(classLoader, origClass, entryPoint);
            origMethod.setAccessible(true);

            for (Object[] args : inputs) {
                try {
                    System.out.println("use inputs: " + Arrays.toString(args));
                    long start = System.currentTimeMillis();
                    Object expected = origMethod.invoke(null, new Object[]{args});
                    long origTime = System.currentTimeMillis() - start;

                    start = System.currentTimeMillis();
                    Object actual = patchedMethod.invoke(null, new Object[]{args});
                    long patchedTime = System.currentTimeMillis() - start;

                    String factor = String.format("%.2f", ((double) patchedTime / (double) origTime));
                    System.out.println("original time: " + origTime + "ms, patched time: " + patchedTime + "ms => factor: " + factor);

                    if (expected != actual && !expected.equals(actual)) {
                        if (!expected.getClass().isArray() || !actual.getClass().isArray()
                                || !Arrays.deepEquals(new Object[]{expected}, new Object[]{actual})) {
                            throw new AssertionError("expected:<" + Arrays.deepToString(new Object[]{expected}) + "> but was:<" + Arrays.deepToString(new Object[]{actual}) + ">");
                        }
                    }
                } catch (IllegalAccessException | InvocationTargetException e) {
                    throw new RuntimeException(e);
                }
            }
        } catch (IOException | ClassNotFoundException | NoSuchMethodException | IllegalAccessException | InvocationTargetException e) {
            throw new RuntimeException(e);
        } finally {
            if (interpreterIdeaPath != null && interpreterIdeaPath.toFile().exists()) {
                try {
                    FileUtils.forceDelete(interpreterIdeaPath.toFile());
                } catch (IOException e) {
                    throw new RuntimeException(e);
                }
            }
        }

    }

    @Override
    public String toString() {
        return "JVMExecutor";
    }
}

class DVMExecutor extends TestExecutor {
    private static final String TESTER_CLASS = "Tester";
    private static DexFile interpreterDex;

    public DVMExecutor(ByteCodeMapping byteCodeMapping) {
        super(byteCodeMapping);
    }

    // generate interpreter dex file only once (compiling it lasts very long)
    public DexFile getInterpreterDex() {
        if (interpreterDex == null) {
            InterpreterGenerator interpGen = new InterpreterGenerator(byteCodeMapping, new HashSet<>(Arrays.asList(InstructionType.values())),
                    Collections.emptySet(), vmodexKey, true);
            interpreterDex = interpGen.generateDex();
        }
        return interpreterDex;
    }

    private JavaFile generateTesterClass(GenericMethodReference entryPoint, List<Object[]> inputs) {
        CodeBlock.Builder builder = CodeBlock.builder();

        for (Object[] input : inputs) {
            String inputVals = Stream.of(input).map((i) -> {
                if (i == null)
                    return "null";
                if (i instanceof Byte)
                    return "(byte)" + i;
                if (i instanceof Character)
                    return "'" + i + "'";
                if (i instanceof Short)
                    return "(short)" + i;
                if (i instanceof Long)
                    return i + "L";
                if (i instanceof Float)
                    return i + "f";
                if (i instanceof String)
                    return "\"" + i + "\"";

                return i.toString();
            }).collect(Collectors.joining(", "));
            String intputType = input.getClass().getName().replaceAll("(^\\[L?|;$)", "");

            String testMethod = entryPoint.getDefiningClass().substring(1, entryPoint.getDefiningClass().length() - 1) +
                    "." + entryPoint.getName();

            builder.add("{\n")
                    .indent()
                    .addStatement("Object[] margs = new " + intputType + "[]{" + inputVals + "}")
                    .addStatement("System.out.println(\"use inputs: \" + java.util.Arrays.toString(margs))")
                    .addStatement("long start = System.currentTimeMillis()")
                    .addStatement("Object expected = orig." + testMethod + "(margs)")
                    .addStatement("long origTime = System.currentTimeMillis() - start")

                    .addStatement("start = System.currentTimeMillis()")
                    .addStatement("Object actual = patched." + testMethod + "(margs)")
                    .addStatement("long patchedTime = System.currentTimeMillis() - start")

                    .addStatement("String factor = String.format(\"%.2f\", ((double) patchedTime / (double) origTime))")
                    .addStatement("System.out.println(\"original time: \" + origTime + \"ms, patched time: \" + patchedTime +\"ms => factor: \" + factor)")

                    // compare results
                    .add("" +
                            "if (expected != actual && !expected.equals(actual)) {\n" +
                            "    if (!expected.getClass().isArray() || !actual.getClass().isArray()\n" +
                            "            || !java.util.Arrays.deepEquals(new Object[]{expected}, new Object[]{actual})) {\n" +
                            "        throw new AssertionError(\"expected:<\" + expected + \"> but was:<\" + actual + \">\");\n" +
                            "    }\n" +
                            "}\n")
                    .unindent()
                    .add("}\n");
        }

        TypeSpec.Builder classBuilder = TypeSpec.classBuilder(TESTER_CLASS)
                .addModifiers(Modifier.PUBLIC, Modifier.FINAL)
                .addMethod(MethodSpec.methodBuilder("main")
                        .addModifiers(Modifier.PUBLIC, Modifier.STATIC)
                        .addParameter(ParameterSpec.builder(String[].class, "args").build())
                        .returns(void.class)
                        .addCode(builder.build())
                        .build());

        TypeSpec interpreterType = classBuilder.build();
        return JavaFile.builder("", interpreterType).build();
    }

    @Override
    protected void execute(String code, GenericMethodReference entryPoint, Set<GenericMethodReference> toPatch, List<Object[]> inputs) {
        compileTestCase(code, Collections.singletonList(generateTesterClass(entryPoint, inputs)));

        try {
            VMODexCompiler.CompilationResult compResult = VMODexCompiler.compile(parsedDex, toPatch, this.byteCodeMapping);
            Path vmodexFile = dexDir.resolve(VMODEX_FILE);
            Files.write(vmodexFile, compResult.getVmoDex().serialize(vmodexKey));

            StubInjector.injectInterpreter(parsedDex, compResult.getConvertedMethodReferences(), getInterpreterDex(), true);
            List<DexFile> patchedDexs = parsedDex.toDexFiles();
            if (patchedDexs.size() != 1)
                throw new RuntimeException("more than one dex file in testcase");

            DexFileFactory.writeDexFile(dexFilePath.toString(), patchedDexs.get(0));

            final String testFolder = "/data/local/tmp/interpTest";
//            CommandExecutor.executeCommand("adb shell rm -rf " + testFolder);
            CommandExecutor.executeCommand("adb shell mkdir -p " + testFolder);
            CommandExecutor.executeCommand("adb push " + vmodexFile + " " + testFolder);
            CommandExecutor.executeCommand("adb push " + dexFilePath + " " + testFolder);
            CommandExecutor.CommandResult result = CommandExecutor.executeCommand(
                    "adb shell dalvikvm -cp " + testFolder + "/classes.dex:" + testFolder + "  " + TESTER_CLASS);
            System.out.println(result.getStdout());
        } catch (IOException e) {
            throw new RuntimeException(e);
        }
    }

    @Override
    public String toString() {
        return "DVMExecutor";
    }
}

public class InterpreterTestCase {
    public static final TestExecutor JVM_EXEC = new JVMExecutor(ByteCodeMapping.getDefaultMapping());
    public static final TestExecutor DVM_EXEC = new DVMExecutor(ByteCodeMapping.getDefaultMapping());
    public static final List<TestExecutor> AVAILABLE_EXECUTORS = Arrays.asList(
//            DVM_EXEC,
            JVM_EXEC
    );

    public static final String TEST_CLASS = "Test";
    public static final GenericMethodReference TEST_METHOD = new GenericMethodReference(
            "L" + TEST_CLASS + ";",
            "test",
            "Ljava/lang/Object;",
            "[Ljava/lang/Object;"
    );

    private final String code;
    private final Set<GenericMethodReference> toPatch;
    private final List<TestExecutor> executors;
    private final List<Object[]> inputs;

    public InterpreterTestCase(String code, List<TestExecutor> executors, GenericMethodReference... toPatch) {
        this.code = code;
        this.executors = executors;
        this.toPatch = new HashSet<>(Arrays.asList(toPatch));
        this.inputs = new LinkedList<>();
    }

    public InterpreterTestCase(String code, List<TestExecutor> executors, String... toPatch) {
        this(code, executors, Stream.of(toPatch).map(GenericMethodReference::new).toArray(GenericMethodReference[]::new));
    }

    public InterpreterTestCase(String code, TestExecutor executor, String... toPatch) {
        this(code, Collections.singletonList(executor), toPatch);
    }

    public InterpreterTestCase(String code, String... toPatch) {
        this(code, AVAILABLE_EXECUTORS, toPatch);
    }

    public InterpreterTestCase addInputSample(Object... sample) {
        inputs.add(sample.clone());
        return this;
    }

    public void run() {
        // if no input data was given execute only once without arguments
        if (inputs.size() == 0) {
            inputs.add(new Object[0]);
        }

        String methodName = Thread.currentThread().getStackTrace()[2].getMethodName();

        for (TestExecutor executor : executors) {
            if (AVAILABLE_EXECUTORS.contains(executor)) {
                System.out.println("----------------------------------------------------");
                System.out.println("  run test case '" + methodName + "' using '" + executor + "'");
                System.out.println("----------------------------------------------------");
                executor.runTest(code, TEST_METHOD, toPatch, inputs);
            } else {
                System.out.println("----------------------------------------------------");
                System.out.println("  skip test case '" + methodName + "' (needs unavailable executor: '" + executor + "')");
                System.out.println("----------------------------------------------------");
            }
        }
    }
}
