import "../css/style.scss";

import jQuery from 'jquery';
import 'jstree';
import saveAs from 'file-saver';
import pako from 'pako';
import _ from 'lodash';
import Chart from 'chart.js';
import convert from 'color-convert';

window.$ = jQuery;
window.jQuery = jQuery;
window._ = _;

window.config = {
    interpreterVerbosityLevels: [],
    signDebug: false,
    protect: false,
    debugProtection: false,
    renameInterpreter: false,
    methodsToObfuscate: [],
};
window.apk = null;
window.bigUpdate = false;

function doBigUpdate(cb) {
    window.bigUpdate = true;
    cb();
    window.bigUpdate = false;
}

function escapeHtml(str) {
    var div = document.createElement('div');
    div.appendChild(document.createTextNode(str));
    return div.innerHTML;
}

function updateSelection() {
    window.config.methodsToObfuscate = [];
    $('#apktree').jstree("get_selected", true).forEach((node) => {
        if (node.type === "method" && node.original.analysisResult.supported) {
            window.config.methodsToObfuscate.push(node.original.gmr);
        }
    });
    $("#num_selected").text("#Selected Methods: " + window.config.methodsToObfuscate.length);
    $("#configview").text(JSON.stringify(window.config, null, 2));
}

window.onload = () => {
    $('#apktree').jstree({
        "core": {
            "animation": 0,
            // "data": {
            //     "url": "/sample_tree.json",
            //     "data": "json",
            // },
        },
        "types": {
            "package": {},
            "class": {
                "icon": "/img/class.svg",
            },
            "method": {
                "icon": "/img/method.svg",
            }
        },
        "plugins": [
            "search",
            "types",
            "sort",
            "checkbox",
            // "state",
        ]
    })
        .on("changed.jstree", (e, data) => {
            if (window.bigUpdate) return;
            updateSelection();
        })
        .on("refresh.jstree", (e, data) => {
            let treeModel = data.instance._model.data;
            const highlight = (text) => `<span style="color: red">${text}</span>`;
            const nodeText = (node) => `${escapeHtml(node.name)} (${highlight(node.numUnsupportedMethods)}/${node.numMethods})`;

            for (let [id, node] of Object.entries(treeModel)) {
                let origNode = node.original;
                if (origNode) {
                    if (origNode.type === "method") {
                        if (!origNode.analysisResult.supported) {
                            // disable all unsupported nodes
                            $('#apktree').jstree("disable_node", id);
                            // add reason as tooltip
                            node.li_attr.title = origNode.analysisResult.unsupportedReasons.join("\n");
                        }
                        node.text = `${escapeHtml(origNode.gmr.name + gmrSignature(origNode.gmr))} (${origNode.analysisResult.numInstructions})`;
                    }
                    else if (origNode.type === "class" || origNode.type === "package") {
                        node.text = nodeText(origNode)
                    }
                }
            }
            // Hack to add number of methods to root
            let rootDomEle = $("#j1_1_anchor")[0];
            let rootNode = apk.analysisResult.parsedApk;
            rootDomEle.innerHTML = rootDomEle.innerHTML + nodeText(rootNode);
        });

    function drawErrorDistribution(errorDistribution) {
        let ctx = document.getElementById('errorDistribution').getContext('2d');
        let colors = [];
        let reasons = [];
        let frequencies = [];

        let it = 0;
        const nextColor = () => {
            var h = it % 360;
            var s = 50 + ((it / 360) % 50);
            var v = 70;
            let [r, g, b] = convert.hsv.rgb(h, s, v);
            it += 20;
            return "rgb(" + r + "," + g + "," + b + ")";
        };

        for (let [key, value] of Object.entries(errorDistribution)) {
            colors.push(nextColor());
            reasons.push(key);
            frequencies.push(value);
        }

        let chart = new Chart(ctx, {
            type: 'doughnut',
            data: {
                labels: reasons,
                datasets: [{
                    borderColor: "#cbcbcb",
                    data: frequencies,
                    backgroundColor: colors,
                }]
            },

            // Configuration options go here
            options: {}
        });
    }

    function droppedApk(file) {
        window.apk = {
            name: file.name,
            size: file.size,
            fileContent: file.b64content,
        };

        $.post("/api/analyseapk", window.apk.fileContent)
            .done((data) => {
                window.apk.analysisResult = data;
                $('#apktree').jstree(true).settings.core.data = data.parsedApk;
                $('#apktree').jstree(true).refresh();
                drawErrorDistribution(data.errorDistribution);
            })
            .fail((err) => {
                throw new Error("post /api/analyseapk failed, error:" + err);
            });
    }

    function gmrSignature(gmr) {
        return "(" + gmr.parameterTypes.join("") + ")" + gmr.returnType;
    }

    function gmrToString(gmr) {
        return gmr.definingClass + "->" + gmr.name + gmrSignature(gmr);
    }

    function uploadedConfig(config) {
        if (!window.apk || !window.apk.analysisResult) {
            throw new Error("no analysed apk to which config could be applied");
        }
        if (!config.methodsToObfuscate) {
            throw new Error("invalid obfuscation configuation file was provided");
        }
        const treeData = $('#apktree').jstree()._model.data;

        const gmrToId = {};
        for (let [id, node] of Object.entries(treeData)) {
            if (!node.original || !node.original.gmr)
                continue;
            const hash = gmrToString(node.original.gmr);
            gmrToId[hash] = id;
        }

        const nodeIds = [];
        config.methodsToObfuscate.forEach(gmr => {
            const nodeId = gmrToId[gmrToString(gmr)];
            if (!nodeId) {
                throw new Error("Uploaded config contains method reference not found in loaded apk-file: " + JSON.stringify(gmr));
            }
            nodeIds.push(nodeId);
        });

        doBigUpdate(() => {
            $('#apktree').jstree("deselect_all");
            $('#apktree').jstree("check_node", nodeIds)
        });
        ["signDebug", "protect", "debugProtection", "renameInterpreter", "interpreterVerbosityLevels"].forEach((property) => {
            if(config[property] !== undefined)
                window.config[property] = config[property];
        });
        updateSelection();
    }

    $("#dropzone")
        .bind("dragenter", (event) => {
            event.preventDefault();
            event.stopPropagation();
            $(event.currentTarget).addClass('drag-over');
        })
        .bind("dragover", (event) => {
            event.preventDefault();
            event.stopPropagation();
            $(event.currentTarget).addClass('drag-over');
        })
        .bind("dragleave", (event) => {
            event.preventDefault();
            event.stopPropagation();
            $(event.currentTarget).removeClass('drag-over');
        })
        .on("drop", (event) => {
            event.preventDefault();
            event.stopPropagation();

            const items = event.originalEvent.dataTransfer.items;

            if (items.length !== 1 || items[0].kind !== "file") {
                throw new Error("invalid content dropped");
            }

            let file = items[0].getAsFile();

            function readFile(handler) {
                let reader = new FileReader();
                reader.onload = (fileEvent) => {
                    let content = fileEvent.target.result;
                    content = content.substr(content.indexOf("base64,") + "base64,".length);
                    file.b64content = content;
                    handler(file);
                };
                reader.readAsDataURL(file);
            }

            switch (file.type) {
                case "application/vnd.android.package-archive":
                    readFile(droppedApk);
                    break;
                case "application/json":
                    readFile((file) => {
                        let data = atob(file.b64content);
                        data = JSON.parse(data);
                        uploadedConfig(data);
                    });
                    break;
                case "application/gzip":
                    readFile((file) => {
                        // see https://stackoverflow.com/questions/14620769/decompress-gzip-and-zlib-string-in-javascript
                        // base 64 to binary
                        let data = atob(file.b64content);
                        // binary to string and string to character-number array
                        data = new Uint8Array(data.split('').map(function (x) {
                            return x.charCodeAt(0);
                        }));
                        // decompression
                        data = pako.inflate(data);
                        // convert gunzipped byteArray back to ascii string
                        data = new TextDecoder("utf-8").decode(data);
                        data = JSON.parse(data);
                        uploadedConfig(data);
                    });
                    break;
                default:
                    throw new Error("wrong file type droppen: " + file.type);
            }

            event.originalEvent.dataTransfer.items.clear();
            $(event.currentTarget).removeClass('drag-over');
        });

    $('#searchfield').keyup(() => {
        if (window.searchTimeout) {
            clearTimeout(window.searchTimeout);
        }
        window.searchTimeout = setTimeout(function () {
            var v = $('#searchfield').val();
            $('#apktree').jstree(true).search(v);
        }, 250);
    });

    $('#obfuscateBtn').click(() => {
        if (!window.apk) {
            throw new Error("No apk to obfuscate provided.");
        }
        if (!window.config.methodsToObfuscate.length) {
            throw new Error("No Methods to obfuscate selected.");
        }

        // need to use plain XMLHttpRequest, because jquery fails to retreive binary data
        var request = new XMLHttpRequest();
        request.open("POST", "/api/obfuscate", true);
        request.setRequestHeader('Content-Type', 'application/json; charset=UTF-8');
        request.responseType = "arraybuffer";

        request.onload = function (oEvent) {
            var arrayBuffer = request.response; // Note: not oReq.responseText
            if (request.status === 200) {
                var blob = new Blob([arrayBuffer], {type: "application/vnd.android.package-archive"});
                const newName = window.apk.name.split(".").slice(0, -1).join(".") + "_patched.apk";
                saveAs(blob, newName);
            }
            else {
                throw new Error(`Obfuscation Failed (${request.status}): ${request.statusText}`);
            }
        };

        request.send(JSON.stringify({config: window.config, apkFile: window.apk.fileContent}));
    });

    $('#downloadConfigBtn').click(() => {
        var blob = new Blob([JSON.stringify(window.config, null, 2)], {type: "text/plain;charset=utf-8"});
        saveAs(blob, "obfuscation_config.json");
    });

    $('.nav .nav-link').click((event) => {
        let tab = $(event.target).attr("id");
        $(".nav-link").removeClass("active");
        $(event.target).addClass("active");

        $(".tabContainer").hide();
        $('#' + tab + 'C').fadeIn('slow');
        console.log(tab);
    });

    ["signDebug", "protect", "debugProtection", "renameInterpreter"].forEach((property) => {
        $('#' + property).click((event) => {
            window.config[property] = !window.config[property];
            $("#configview").text(JSON.stringify(window.config, null, 2));
        });
    });

    ["INTERPRET_START", "EVERY_INSTRUCTION"].forEach((property) => {
        $('#' + property).click((event) => {
            if(window.config.interpreterVerbosityLevels.includes(property)) {
                _.remove(window.config.interpreterVerbosityLevels, (x) => x === property);
            } else {
                window.config.interpreterVerbosityLevels.push(property);
            }
            $("#configview").text(JSON.stringify(window.config, null, 2));
        });
    });

};