package de.marstra.dvo.obfuscator.util;

import com.fasterxml.jackson.annotation.JsonCreator;
import com.fasterxml.jackson.annotation.JsonIgnore;
import com.fasterxml.jackson.annotation.JsonProperty;
import lombok.EqualsAndHashCode;
import lombok.Getter;
import org.jf.dexlib2.immutable.reference.ImmutableMethodReference;

import java.util.Arrays;
import java.util.LinkedList;
import java.util.List;

// TODO document format of typestrings
@EqualsAndHashCode
public class GenericMethodReference {
    @Getter
    private final String definingClass;
    @Getter
    private final String name;
    @Getter
    private final String[] parameterTypes;
    @Getter
    private final String returnType;

    public GenericMethodReference(String definingClass, String name, String returnType, String... parameterTypes) {
        this.definingClass = definingClass;
        this.name = name;
        this.parameterTypes = parameterTypes;
        this.returnType = returnType;
    }

    @JsonCreator
    public GenericMethodReference(@JsonProperty("definingClass") String definingClass, @JsonProperty("name") String name,
                                  @JsonProperty("returnType") String returnType, @JsonProperty("parameterTypes") List<? extends CharSequence> parameterTypes) {
        // parameterTypes.toArray(new String[0]) fails with ArrayStoreException
        String[] array = new String[parameterTypes.size()];
        for (int i = 0; i < array.length; i++) {
            array[i] = parameterTypes.get(i).toString();
        }
        this.definingClass = definingClass;
        this.name = name;
        this.parameterTypes = array;
        this.returnType = returnType;
    }

    public GenericMethodReference(String definingClass, String name, String signature) {
        this(definingClass, name, returnTypeFromSignature(signature), parameterTypesFromSignature(signature));
    }

    public GenericMethodReference(org.jf.dexlib2.iface.reference.MethodReference methodReference) {
        this(methodReference.getDefiningClass(), methodReference.getName(), methodReference.getReturnType(), methodReference.getParameterTypes());
    }

    public GenericMethodReference(interpreter.MethodReference methodReference) {
        this(methodReference.definingClass, methodReference.name, methodReference.returnType, methodReference.parameterTypes);
    }

    /**
     * parse Method Reference from String of form: CLASS_NAME->METHOD_NAME(ARG_TYPE_1ARG_TYPE_2...)RETURN_TYPE_1
     * examples:
     * Ljava/lang/String;->toLowerCase()V
     * Ljava/lang/String;->concat(Ljava/lang/String;)Ljava/lang/String;
     * Ljava/lang/String;->length()I
     * Ljava/lang/String;->charAt(I)C
     *
     * @param ref method reference string
     */
    public GenericMethodReference(String ref) {
        String[] parts = ref.split("->");
        if (parts.length != 2)
            throw new IllegalArgumentException("invalid ref given: " + ref);

        this.definingClass = parts[0];
        int idx = parts[1].indexOf('(');
        this.name = parts[1].substring(0, idx);
        String signature = parts[1].substring(idx);
        this.parameterTypes = parameterTypesFromSignature(signature);
        this.returnType = returnTypeFromSignature(signature);
    }

    public static String returnTypeFromSignature(String signature) {
        int idx = signature.lastIndexOf(')');
        String returnType = signature.substring(idx + 1).trim();
        if (returnType.isEmpty())
            throw new IllegalArgumentException("Could not parse returntype from signature " + signature);
        return returnType;
    }

    // TODO rewrite so only valid signatures are parsed
    // TODO write tests
    // see https://source.android.com/devices/tech/dalvik/dex-format#typedescriptor
    public static String[] parameterTypesFromSignature(String signature) {
        int idx = signature.lastIndexOf(')');
        String paramString = signature.substring(1, idx);

        if (paramString.isEmpty()) {
            return new String[0];
        }

        List<String> params = new LinkedList<>();
        final String typeStarters = "ZBSCIJFDL[";
        StringBuilder curType = new StringBuilder();
        boolean classDef = false;

        for (int i = 0; i < paramString.length(); i++) {
            char curChar = paramString.charAt(i);

            if (!classDef && !typeStarters.contains("" + curChar))
                throw new IllegalArgumentException("invalid signature given: " + signature);

            if (!classDef && typeStarters.contains("" + curChar)) {
                if (curType.length() > 0) {
                    params.add(curType.toString());
                    curType = new StringBuilder();
                }
            }

            if (curChar == 'L' || curChar == '[') {
                classDef = true;
            }
            if (curChar == ';') {
                classDef = false;
            }

            curType.append(curChar);
        }
        assert curType.length() > 0;
        params.add(curType.toString());
        return params.toArray(new String[0]);
    }

    public org.jf.dexlib2.iface.reference.MethodReference toDexLibMethodReference() {
        return new ImmutableMethodReference(definingClass, name, Arrays.asList(parameterTypes), returnType);
    }

    public interpreter.MethodReference toVmoMethodReference() {
        return new interpreter.MethodReference(definingClass, name, parameterTypes, returnType);
    }

    @JsonIgnore
    public String getSignature() {
        return "(" + String.join("", parameterTypes) + ")" + returnType;
    }

    @Override
    public String toString() {
        return definingClass + "->" + name + getSignature();
    }

    // TODO replace / with .?
    public static String convertDexTypeToJVMType(String type) {
        if (type.charAt(0) == '[')
            return type;

        if (type.charAt(0) != 'L' && type.charAt(type.length() - 1) != ';')
            throw new IllegalArgumentException("invalid dex type: " + type);

        return type.substring(1, type.length() - 1);
    }
}
